-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HammerOfJustas.lua
-- DESC: Main file containing:
-- * frame handlers
-- * update loop
-- * command-line parsing
-- * addon messaging
-- * profiling functionality
-- * saved variable management
-- * game event handling

-- TODO general:
-- - track incoming damage spells / skills. If these are significant damage dealers, then use them
--   to predict mitigations like Shield of the Righteous and Shield Block when the spell is being 
--   cast.

-- TODO BFA:
-- base initial TTL off of session DPS when available
-- crowd-control cooldowns as rules (to get out of CCs)
-- Show second-best rotation by running rules a second time before forcing the first choice to fail

-- X profiling -- add a list of functions, get the time for each function at the end of frame, print the top functions
-- - add health and mana bars so we can see them without turning on personal buffs?
-- X need to figure out how to fix the unhide in combat issue
--   - instead of hide and unhide we just have to enable / disable mouse
--   - This fixes issues with hidden bars usurping mouse input, but doesn't fix issues with buttons locked out during combat.
-- X incoming damage can be used to predict when the player will die and put up a LoH or DS icon.
-- - if we have any usable trinkets they can be added to the skill or buff bars
-- X show cooldown on rotation
-- X consider selecting rotation skills less frequently than every frame so things feel more stable.
-- - Ret
--   X don't show blessings in a raid or party unless we can check if anyone in the party has our buff
--   - change inquisition timer to a bar?
--   - Should Wake of Ashes show if you will only waste 1 hopo?
--     - if Avenging Wrath / Crusade is on, yes
--     X if Wake is available and we have a choice between skills that will take us to 3 HoPo or 3+ HoPo,
--       choose the skill that takes us to exactly 3 HoPo
--   - Show Inquisition on cooldowns.
--   X test Hammer of Wrath in rotation.
--     X BUG: Hammer of Wrath showing up when target has > 20% health.
--   X Add Shield of Vengeance to cooldowns.
--   - Add Flash of Light to cooldowns (with Selfless Healer)
--   X BUG: when Inquisition is not a selected talent, inquisition timer always shows
--   X Word of Glory support
--   X Justicar's Vengeance support
--   x Hammer of Wrath when Crusade / Avenging Wrath is active
--   - Inquisition timer text sometimes flickers to full alpha
-- - Protection:
--   X have yet to try pro with this addon in BFA
-- - notify when cleanse is necessary (non-raid only)
-- X BUG: when loading the addon as Prot, Buff 1 frame text shows.
-- X BUG: Wake of Ashes and Blade of Justice aren't showing in rotation
--   - bad variable name used when added Prot range checking... Lua honey-badgers unknown variables.
-- X disable addon when in a vehicle or controlling another character
-- X BUG: showing cooldowns frame toggle didn't save. Misspelled HoJ._showCooldownsFrame when writing saved variables
-- TODO WoD:
-- * instead of change skills and buffs, change to rotation and skills
--   - rotation would show current best skill and next available skill,
--     including situational skills like Rebuke and LoH.
--   - skills would show currently available skills, and skills on CD
-- X Show Hammer of Wrath on skills bar ... geez.
-- X Show Divine Storm and TV/FV on skills bar
-- X Sort skills not on CD by priority instead of cooldown
-- * Supplication!
-- X Show Divine Storm when Divine Purpose procs and Empowered Divine Storm procs
-- X Show Templar's Verdict / Final Verdict when Divine Purpose procs
--
-- TODO Mists:
-- * show hammer of wrath in skills bar
-- * show Divine Crusader in skills bar
-- * option for show global cooldowns on skills
-- X countdown timer on inquisition
-- * separate buffs window in to a buffs and skills window
-- * rebuke icon when target is casting an interruptible spell
-- TODO Cataclysm:
-- X inquisition timer
-- X dim when mounted
-- TODO:
-- * warn when Ret has Righteous Fury on in a raid or instance. Righteous Fury is
--   useful in some PvE situations like holding aggro on Kilnmasters on Timeless Isle.
-- * if the player's target doesn't have aggro, show the icon for Hand of Reckoning
--   - this must be configurable and shouldn't default to on in instance and raids
-- * customize interface more
--   - allow icons to be clickable!
--   - allow several different setups for buttons or make the button windows movable
--     - just make a specific window for each button?  this would speed up checking
--       for which buffs are being shown.
-- * add hammer of wrath icon when target is below 20% health and HoW is not on cooldown
-- * allow user to choose the time before an icon is off cooldown for it to appear
--   - i.e. show Hammer of Wrath 0.25 seconds before the cooldown is up
--   - Cooldown Early Warning
-- X add Art of War
-- * add Divine Plea
-- * when in combat, check that the paladin has his preferred buffs such as Blessing
--   of Might, and a seal.
-- * show Flash Heal and Excorcism when Art of War is up.
--   - it's only necessary to show these if they're clickable, otherwise a single
--     art of war icon lets us know we can cast either.
--     - problem with this is that exorcism could be on cooldown
-- * option to show Turn Evil
-- * option to show Holy Wrath
-- * show Righteous Fury if ret points in Improved Righteous Fury
--   - for ret this should turn off in raids and instances
-- * trinkets
-- * show Hand of Freedom when movement impaired.
-- * generalize...
--   - choose a skill
--   - choose rules for when to show the skill
--     - mounted / not mounted
--     - in combat
--     - hostile target
--     - another buff is active
--     - choose early warning or not
--     - target health is below some %
-- -----------------------------------------------------------------------------

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HammerOfJustas.lua" )

-- -----------------------------------------------------------------------------
-- Slash command handling
-- -----------------------------------------------------------------------------
SLASH_HAMMEROFJUSTAS1, SLASH_HAMMEROFJUSTAS2 = '/hoj', '/hammerofjustas'

function SlashCmdList.HAMMEROFJUSTAS( msg, editbox )
	local command, rest = msg:match( "^(%S*)%s*(.-)$" );
	local name = rest:match( "^(%S*)%s*(%a-)" );

	-- HoJ:Print( "command = '" .. command .. "', rest = '" .. rest .. "'." );
	-- HoJ:Print( "name = '" .. name .. "'." );

	if ( command == "options" or command == "opt" ) then
		InterfaceOptionsFrame_OpenToCategory(HoJ._text["Hammer of Justas"])
		InterfaceOptionsFrame_OpenToCategory(HoJ._text["Hammer of Justas"])
	elseif ( command == "ignore" or command == "add" ) then
		HoJ._blockedNames[ name ] = name;
		HoJ:Msg( "Added " .. name .. " to whisper ignore list." );
	elseif ( command == "unignore" or command == "remove" ) then
		HoJ._blockedNames[ name ] = nil;
		HoJ:Msg( "Removed " .. name .. " from whisper ignore list." );
	elseif ( command == "list" ) then
		local count = 1;
		HoJ:Msg( "Ignore List:" );
		for key, value in pairs( HoJ._blockedNames ) do
			HoJ:Print( HoJ.STR_WHITE .. "  " .. count .. ": " .. value );
			count = count + 1;
		end
	elseif ( command == "silent" ) then
		if ( HoJ._silentIgnore == true ) then
			HoJ._silentIgnore = false;
			HoJ:Msg( "Silent Ignore is OFF." );
		else
			HoJ._silentIgnore = true;
			HoJ:Msg( "Silent Ignore is ON." );
		end
	elseif ( command == "stealth" ) then
		if ( HoJ._stealthIgnore == true ) then
			HoJ._stealthIgnore = false;
			HoJ:Msg( "Stealth Ignore is OFF." );
		else
			HoJ._stealthIgnore = true;
			HoJ:Msg( "Stealth Ignore is ON." );
		end
	elseif ( command == "history" ) then
		HoJ:Msg( "Ignore History:" );
		for key, value in pairs( HoJ._ignoreHistory ) do
			HoJ:Print( HoJ.STR_WHITE .. "From " .. key .. ": " );
			for i, historyMsg in ipairs( HoJ._ignoreHistory[ key ] ) do
				HoJ:Print( HoJ.STR_WHISPER .. "  " .. historyMsg );
			end
		end
	elseif ( command == "clear" ) then
		HoJ._ignoreHistory = { };
		HoJ:Print( "History cleared." );
	elseif ( command == "prof" or command == "profile" ) then
		if ( HoJ._enableProfiling ) then
			HoJ._enableProfiling = false;
			HoJ:Msg( "Profiling turned OFF." );
		else
			HoJ._enableProfiling = true;
			HoJ:Msg( "Profiling turned ON." );
		end
	elseif ( command == "thresh" or command == "threshold" ) then
		HoJ._profileThreshold = tonumber( name )
		if ( HoJ._profileThreshold == nil ) then
			HoJ._profileThreshold = 0.3
		end
		HoJ:Msg( "Profiling threshold set to " .. HoJ._profileThreshold );
	elseif ( command == "func" ) then
		HoJ._profileFunction = name
	elseif ( command == "disable" ) then
		HoJ._enabled = false;
		HoJ._needsReset = true;
		HoJ:Msg( "Update disabled." );
	elseif ( command == "enable" ) then
		HoJ._enabled = true;
		HoJ:Msg( "Update enabled." );
	elseif ( command == "vercheck" ) then
		HoJ._versionChecksEnabled = not HoJ._versionChecksEnabled
		if ( HoJ._versionChecksEnabled ) then
			HoJ:Msg( "Version checks enabled." )
		else
			HoJ:Msg( "Version checks disabled." )
		end
	elseif ( command == "verbroadcast" ) then
		HoJ._versionBroadcastEnabled = not HoJ._versionBroadcastEnabled
		if ( HoJ._versionBroadcastEnabled ) then
			HoJ:Msg( "Version broadcast enabled." )
		else
			HoJ:Msg( "Version broadcast disabled." )
		end
	elseif ( command == "save" ) then
		HoJ:CreateSavedVariablesFromHoJ( "Writing saved variables." )
	elseif ( command == "debugRange" ) then
		HoJ._showRange = not HoJ._showRange
		if ( HoJ._showRange ) then
			HoJ:Msg( "Showing range to target." )
		else
			HoJ:Msg( "Disabled showing range to target." )
		end
	elseif ( command == "debugBlessings" ) then
		HoJ._debugBlessings = not HoJ._debugBlessings
		HoJ:Msg( "_debugBlessings = " .. tostring( HoJ._debugBlessings ) )
	elseif ( command == "debugRules" ) then
		if ( name == nil or name == '' ) then
			if ( HoJ._debugRulesLevel == 0 ) then
				HoJ._debugRulesLevel = 2
			else
				HoJ._debugRulesLevel = 0
			end
		else
			HoJ._debugRulesLevel = tonumber( name )
		end
		HoJ:Msg( "_debugRulesLevel = " .. tostring( HoJ._debugRulesLevel ) )
	elseif ( command == "debugNextSkill" ) then
		if ( name == nil or name == '' ) then
			if ( HoJ._debugNextSkillLevel == 0 ) then
				HoJ._debugNextSkillLevel = 2
			else
				HoJ._debugNextSkillLevel = 0
			end
		else
			HoJ._debugNextSkillLevel = tonumber( name )
		end
		HoJ:Msg( "_debugNextSkillLevel = " .. tostring( HoJ._debugNextSkillLevel ) )
	elseif ( command == "debugPower" ) then
		HoJ._debugPower = not HoJ._debugPower
		HoJ:Msg( "_debugPower = " .. tostring( HoJ._debugPower ) )
	elseif ( command == "debugHealth" ) then
		HoJ._debugHealth = not HoJ._debugHealth
		HoJ:Msg( "_debugHealth = " .. tostring( HoJ._debugHealth ) )
	elseif ( command == "debugAlphas" ) then
		HoJ:DebugMsg( format( "_globalAlpha = %.2f", HoJ._globalAlpha ) )
		HoJ:DebugMsg( format( "_rotationAlpha = %.2f", HoJ._rotationAlpha ) )
		HoJ:DebugMsg( format( "_rotationSecondaryAlpha = %.2f", HoJ._rotationSecondaryAlpha ) )
		HoJ:DebugMsg( format( "_cooldownsAlpha = %.2f", HoJ._cooldownsAlpha ) )
		HoJ:DebugMsg( format( "_buffsAlpha = %.2f", HoJ._buffsAlpha ) )
		HoJ:DebugMsg( format( "_holyPowerAlpha = %.2f", HoJ._holyPowerAlpha ) )
		HoJ:DebugMsg( format( "_debugAlpha = %.2f", HoJ._debugAlpha ) )
		local hfra = HoJFrame_Rotation:GetAlpha()
		HoJ:DebugMsg( format( "HoJ_Rotation:GetAlpha() = %2.f", hfra ) )
		local hrfa = HoJ_RotationFrame:GetAlpha()
		HoJ:DebugMsg( format( "HoJ_RotationFrame:GetAlpha() = %2.f", hrfa ) )

		for i, v in ipairs( HoJ._rotationFrames) do
			HoJ:DebugMsg( format( "%i: frame alpha = %.2f", i, v.frame:GetAlpha() ) )
			HoJ:DebugMsg( format( "%i: texture alpha = %.2f", i, v.texture:GetAlpha() ) )
			HoJ:DebugMsg( format( "%i: cooldownFrame alpha = %.2f", i, v._cooldownFrame:GetAlpha() ) )
		end
	elseif ( command == "test" ) then
		HoJ:SaveVariables()
	else
		HoJ:Msg( "slash commands:" );
		HoJ:Print( HoJ.STR_WHITE .. "  ignore <name> - ignores whispers from the specified player." );
		HoJ:Print( HoJ.STR_WHITE .. "  add <name> - same as ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  unignore <name> - stops ignoring whispers from the specified player." );
		HoJ:Print( HoJ.STR_WHITE .. "  remove <name> - same as unignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  list - lists all currently ignored players." );
		HoJ:Print( HoJ.STR_WHITE .. "  silent - toggle silent ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  stealth - toggle stealth ignore." );
		HoJ:Print( HoJ.STR_WHITE .. "  profile - toggle addon profiling." );
		HoJ:Print( HoJ.STR_WHITE .. "  disable - disable the addon's update." );
		HoJ:Print( HoJ.STR_WHITE .. "  enable - enable the addon's update." );
		HoJ:Print( HoJ.STR_WHITE .. "  vercheck - toggle checking for updated version." );
		HoJ:Print( HoJ.STR_WHITE .. "  verbroadcast - toggle broadcasting version number to addon channel." );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:EnterCombat
-- -----------------------------------------------------------------------------
function HoJ:EnterCombat( ... )
	--HoJ:DebugMsg( "HoJ:EnterCombat" );
	HoJ._inCombat = true

	local curTime = HoJ._curTime
	local dpsStats = HoJ._combatDpsStats

	HoJ:ResetDpsStats( dpsStats, curTime )

	HoJ._combatStartTime = curTime

	HoJ._lastPower = UnitPower( "player", HoJ._classSpecInfo._powerType )
	HoJ._powerPerSec = HoJ._curSpecInfo._defaultPowerPerSec
	HoJ._totalPowerGain = 0
	HoJ._powerGainInterval = 3.0
	HoJ._lastPowerGainTime = curTime
	HoJ._nextPowerGainTime = curTime + HoJ._powerGainInterval
	HoJ._powerGainStep = HoJ._curSpecInfo._powerGainStep
end

-- -----------------------------------------------------------------------------
-- HoJ:LeaveCombat
-- -----------------------------------------------------------------------------
function HoJ:LeaveCombat( ... )
	--HoJ:DebugMsg( "HoJ:LeaveCombat" );
	HoJ._inCombat = false

	local predictionErrorSeconds = HoJ._curTime - HoJ._combatDpsStats._firstPredictedTargetDeathTime
	-- HoJ:DebugMsg( string.format( "Combat first TTL prediction error = " .. HoJ._curTime .. " - " .. HoJ._combatDpsStats._firstPredictedTargetDeathTime .. " = %.2f", predictionErrorSeconds ) )
	predictionErrorSeconds = HoJ._curTime - HoJ._combatDpsStats._predictedTargetDeathTime
	-- HoJ:DebugMsg( string.format( "Combat TTL prediction error = " .. HoJ._curTime .. " - " .. HoJ._combatDpsStats._predictedTargetDeathTime .. " = %.2f", predictionErrorSeconds ) )

	HoJ:AddSessionDps( HoJ._curTime, HoJ._combatDpsStats )

	HoJ:ResetDpsStats( HoJ._combatDpsStats, 0.0 )

	local combatDuration = GetTime() - HoJ._combatStartTime
	HoJ._powerPerSec = HoJ._totalPowerGain / combatDuration
end

-- -----------------------------------------------------------------------------
-- HoJ:DoDPSRotation
-- checks the DPS rotation twice, once for the current time, and if that fails
-- once for a future time
-- -----------------------------------------------------------------------------
function HoJ:DoDPSRotation( skillsList, rotationSelectionList )
	--HoJ:DumpRotationRules( "retRules", rotationSelectionList )
	if ( rotationSelectionList == nil ) then
		return
	end

	local bestSkill, secondBestSkill = self:SelectSkill( rotationSelectionList )
	-- if bestSkill is nil, then we can find the next skill off cooldown and use that
	if ( bestSkill == nil ) then
		if ( secondBestSkill == nil ) then
			bestSkill, secondBestSkill = HoJ._curSpecInfo._nextSkillSelectFn( self, rotationSelectionList )
			if ( HoJ._debugRulesLevel > 0 ) then
				if ( bestSkill ~= nil ) then
					HoJ:DebugMsg( "Selected next skill " .. bestSkill.name )
				else
					HoJ:DebugMsg( "No skill selected!" )
				end
			end
		else
			bestSkill = secondBestSkill
			secondBestSkill = nil
		end
	elseif ( HoJ._debugRulesLevel > 0 ) then
		if ( secondBestSkill ~= nil ) then
			HoJ:DebugMsg( "Selected skills " .. tostring( bestSkill.name ) .. " and " .. tostring( secondBestSkill.name ) )
		else
			HoJ:DebugMsg( "Selected skills " .. tostring( bestSkill.name ) .. " and nil" )
		end
	end

	HoJ._powerBarMax = -1.0

	if ( bestSkill ~= nil ) then
		self:AddIcon( skillsList, bestSkill )

		-- if the _powerBarMax is set to anything > 0.0, then a power bar will be show next to the rotation skill
		if ( bestSkill._powerGen < 0 and HoJ._curSpecInfo._autoPowerGen == true ) then
			HoJ._powerBarMax = bestSkill._powerGen * -1.0
		end

		if ( secondBestSkill ~= nil and HoJ._showRotationSecondaryFrame ) then
			self:AddIcon( skillsList, secondBestSkill )
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckDPSRotation
-- checks the DPS rotation twice, once for the current time, and if that fails
-- once for a future time
-- -----------------------------------------------------------------------------
function HoJ:CheckDPSRotation( skillsList, rotationRules )
	if ( not HoJ._showRotationFrame ) then
		return false;
	end

	if ( not HoJ._inCombat and not HoJ._targetCanAttack ) then
		-- HoJ:DebugMsg( "not in combat or target can't attack" );
		return false;
	end
	self:DoDPSRotation( skillsList, rotationRules )
end

-- -----------------------------------------------------------------------------
-- HoJ._events:CHAT_MSG_ADDON
-- -----------------------------------------------------------------------------
function HoJ._events:CHAT_MSG_ADDON( prefix, text, channel, sender, target, zoneChannelID, localID, name, instanceID )
	if ( prefix == HoJ._addonPrefix ) then
--[[
		local playerRealm = GetRealmName()
		local playerName = UnitName( "player" )
		local playerFullName = playerName .. "-" .. playerRealm
		if ( playerFullName ~= sender ) then
			HoJ:Msg( HoJ.MSG_PREFIX .. "addon message from " .. sender )
		end
--]]
		local t = HoJ:SplitString( text, " " )
		if ( t[1] == "version" ) then
			HoJ:HandleVersionBroadcast( t )
		else
			HoJ:ChannelError( "Unknown addon message" )
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:ADDON_LOADED( ... )
-- -----------------------------------------------------------------------------
function HoJ._events:ADDON_LOADED( arg1 )
	if ( arg1 == "HammerOfJustas" ) then
		HoJ_OnAddonLoaded( self );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LOGOUT( ... )
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_LOGOUT( arg1 )
	HoJ:CreateSavedVariablesFromHoJ()
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LOGIN
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_LOGIN( ... )
	--HoJ:DebugMsg( "PLAYER_LOGIN" )

	HoJ:Finalize( "PLAYER_LOGIN" )

	HoJ._worldState = HoJ.WORLD_STATE_LOGGED_IN
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_ENTERING_WORLD
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_ENTERING_WORLD( ... )
	--HoJ:DebugMsg( "PLAYER_ENTERING_WORLD" )

	-- TODO: multiple greetings for the heck of it
	--HoJ:Print( "Welcome, " .. HoJ._playerName .. ". May the light favor your hammer." );

	-- HoJ:DebugMsg( "Entering world: level = " .. UnitLevel( "player" ) )

	-- HoJ:Finalize( "PLAYER_ENTERING_WORLD" )

	HoJ._worldState = HoJ.WORLD_STATE_ENTERED
	HoJ._enteredWorldUpdateFrameNum = HoJ._updateFrameNum
	if ( HoJ._curSpecInfo ~= nil ) then
		HoJ:ResetBlessings( HoJ._curSpecInfo._activeBlessings )
	end

	-- tell other players running HOJ what our version is
	HoJ:BroadcastVersion()
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LEAVING_WORLD
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_LEAVING_WORLD( ... )
	--HoJ:DebugMsg( "Leaving world: level = " .. UnitLevel( "player" ) )
	HoJ._worldState = HoJ.WORLD_STATE_LEFT
	HoJ._needsReset = true -- reset icons and such
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_THREAT_LIST_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_THREAT_LIST_UPDATE( ... )
	--HoJ:DebugMsg( "UNIT_THREAT_LIST_UPDATE" );
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_ENTER_COMBAT
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_REGEN_DISABLED( ... )
	-- HoJ:DebugMsg( "PLAYER_REGEN_DISABLED" )
	HoJ:EnterCombat( ... )
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_LEAVE_COMBAT
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_REGEN_ENABLED( ... )
	-- HoJ:DebugMsg( "PLAYER_REGEN_ENABLED" )
	HoJ:LeaveCombat( ... )
end


-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_CONTROL_LOST
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_CONTROL_LOST( ... )
	-- HoJ:DebugMsg( "PLAYER_CONTROL_LOST" )
	HoJ._controlLost = true
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_CONTROL_GAINED
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_CONTROL_GAINED( ... )
	-- HoJ:DebugMsg( "PLAYER_CONTROL_GAINED" )
	HoJ._controlLost = false
end

--[[
-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_DYNAMIC_FLAGS
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_DYNAMIC_FLAGS( unitId, flags, ... )
	HoJ:DebugMsg( "UNIT_DYNAMIC_FLAGs: Unit = " .. tostring( unitId ) .. ", flags = " .. tostring( flags ) )
end
--]]

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_POWER_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_POWER_UPDATE( unitTarget, powerType )
	if ( unitTarget == "player" and HoJ._inCombat ) then
		local curTime = GetTime()
		local combatDuration = curTime - HoJ._combatStartTime

		if ( powerType == "RAGE" or powerType == "HOLY_POWER" ) then
			local curPower = UnitPower( "player", HoJ._classSpecInfo._powerType )
			local powerGain = curPower - HoJ._lastPower
			HoJ._lastPower = curPower
			if ( powerGain > 0 ) then
				-- always update the last gain time if > 0
				local lastPowerGainTime = HoJ._lastPowerGainTime
				HoJ._lastPowerGainTime = curTime
				local wasCapped = curPower >= HoJ._curSpecInfo._powerMax
				if ( HoJ._debugPower ) then
					HoJ:DebugMsg( "UNIT_POWER_UPDATE: CP = " .. string.format( "%.1f", curPower ) ..
						", PM = " .. string.format( "%.1f", HoJ._curSpecInfo._powerMax ) .. 
						", PG = " .. string.format( "%.1f", powerGain ) ..
						", WC = " .. tostring( wasCapped ) )
				end
				-- only update the step if the spec uses power gain and power didn't cap cap
				if ( not wasCapped and HoJ._curSpecInfo._autoPowerGen == true ) then
					HoJ._powerGainInterval = curTime - lastPowerGainTime
					HoJ._nextPowerGainTime = curTime + HoJ._powerGainInterval
					if ( powerGain > HoJ._powerGainStep ) then
						HoJ._powerGainStep = powerGain
					end
				end
				HoJ._totalPowerGain = HoJ._totalPowerGain + powerGain
				-- use _defaultPowerPerSec for the start of the fight
				if ( combatDuration > 5.0 ) then
					HoJ._powerPerSec = HoJ._totalPowerGain / combatDuration
--[[					
					HoJ:DebugMsg( "PowerGain = " .. string.format( "%.2f", powerGain ) .. 
						", tpg = " .. string.format( "%.2f", HoJ._totalPowerGain ) .. 
						", cd = " .. string.format( "%2.f", combatDuration ) .. 
						", pps = " .. string.format( "%.2f", HoJ._powerPerSec ) )
--]]						
				end
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:PLAYER_TALENT_UPDATE
-- -----------------------------------------------------------------------------
function HoJ._events:PLAYER_TALENT_UPDATE( ... )
	-- HoJ:DebugMsg( "PLAYER_TALENT_UPDATE" );
	-- HoJ._needsReset = true;
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateRecentAttackersList
-- -----------------------------------------------------------------------------
function HoJ:UpdateRecentAttackersList( attackersList )
	local removedItem = false
	local numAttackers = 0
	for key, value in pairs( attackersList ) do
		local timeSinceUpdate = HoJ._curTime - value._updateTime
		if ( timeSinceUpdate > 2.0 ) then
			attackersList[key] = nil
			removedItem = true
		else
			numAttackers = numAttackers + 1
		end
	end

--[[		
		local i = 1	
		for key, value in pairs( attackersList ) do
			HoJ:DebugMsg( "attacker[" .. i .. "] = " .. key .. ", unitName = " .. value._unitName )
			i = i + 1		
		end
--]]
	return numAttackers
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateRecentMeleeAttackers
-- -----------------------------------------------------------------------------
function HoJ:UpdateRecentMeleeAttackers()
	if ( self._updateFrameNum < self._nextRecentMeleeAttackerUpdateFrame ) then
		return 
	end

	self._nextRecentMeleeAttackerUpdateFrame = self._updateFrameNum + self.UPDATES_PER_SEC

	self._numMeleeAttackers = HoJ:UpdateRecentAttackersList( self._recentMeleeAttackers )
end

-- -----------------------------------------------------------------------------
-- HoJ:AddRecentAttacker
-- -----------------------------------------------------------------------------
function HoJ:AddRecentAttacker( attackersList, sourceGUID, sourceName, spellId, spellName )
	--if ( sourceGUID == self._playerGUID ) then
	--	HoJ:DebugMsg( "Self attack by spell " .. tostring( spellName ) .. ", id = " .. tostring( spellId ) )
	--end
	if ( attackersList[sourceGUID] == nil ) then
		attackersList[sourceGUID] = {
			_updateTime = HoJ._curTime,
			_unitName = sourceName
		}
		-- HoJ:DebugMsg( "Added attacker '" .. sourceName .. "'" )
	else 
		-- already exists, so update the time
		attackersList[sourceGUID]._updateTime = self._curTime
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:AddAttackerIfInMultitargetRange
-- -----------------------------------------------------------------------------
function HoJ:AddAttackerIfInMultitargetRange( attackerList, attackerGUID, attackerName, spellId, spellName, maxRange )
	if ( self._curSpecInfo ~= nil ) then
		if ( self._curSpecInfo._multiTargetRange ~= nil ) then
			if ( maxRange < self._curSpecInfo._multiTargetRange ) then
				HoJ:AddRecentAttacker( attackerList, attackerGUID, attackerName, spellId, spellName )
			elseif ( self._curSpecInfo._meleeAoESpells ~= nil and self._curSpecInfo._meleeAoESpells[spellId] ~= nil ) then
				-- consider it a hit if delivered by player AoE damage
				HoJ:AddRecentAttacker( attackerList, attackerGUID, attackerName, spellId, spellName )
			end
		else
			HoJ:Error( "curSpec " .. self._curSpecInfo._name .. " has no multitarget range set." )
		end
	end			
end

-- -----------------------------------------------------------------------------
-- HoJ:HandleCombatLogEvent
-- -----------------------------------------------------------------------------
function HoJ:HandleCombatLogEvent( ... )	
	local timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags = select( 1, ... )


	local playerIsSource = sourceGUID == self._playerGUID
	local playerIsTarget = destGUID == self._playerGUID
	local isDamage = string.find( event, "_DAMAGE" )
	local isMiss = string.find( event, "_MISSED" )
	local isHeal = string.find( event, "_HEAL" )

	-- HoJ:DebugMsg( "LOG EVENT: event = " .. tostring( event ) )

	if ( isDamage or isMiss or isHeal ) then
		local isSpell = string.find( event, "SPELL_" ) == 1				
		local commonSelectStart
		if ( isSpell or isHeal ) then
			commonSelectStart = 15
		else
			commonSelectStart = 12
		end
		if ( isDamage or isMiss ) then

			local amount, overkill, school, resisted, blocked, absorbed, critical, glancing, 
					crushing, isOffHand = select( commonSelectStart, ... )

			local spellId = 0
			local spellName = "swing"
			local spellSchool = 1
			local name = "melee"
			local rank = 1
			local icon = 0
			local castTime = 0
			local minRange = 0
			local maxRange = self._curSpecInfo._multiTargetRange

			if ( isSpell ) then
				spellId, spellName, spellSchool = select( 12, ... )
				name, rank, icon, castTime, minRange, maxRange = GetSpellInfo( spellId )
			end

			local actualDamageDone = 0
			if ( isDamage ) then 
				if ( type( amount ) ~= "string" ) then
					actualDamageDone = amount
					if ( overkill ~= nil and type( overkill ) ~= "string" ) then
						actualDamageDone = actualDamageDone - overkill
					end
					if ( resisted ~= nil and type( resisted ) ~= "string" ) then
						actualDamageDone = actualDamageDone - resisted
					end
					if ( blocked ~= nil and type( blocked ) ~= "string" ) then
						actualDamageDone = actualDamageDone - blocked
					end
					if ( absorbed ~= nil and type( absorbed ) ~= "string" ) then
						actualDamageDone = actualDamageDone - absorbed
					end
				end

				-- should we track this as a major damage skill?
				if ( isSpell and playerIsTarget ) then
					-- use total damage, not actual damage, since the player may have mitigated it.
					local damageToHealthRatio = amount / HoJ._playerMaxHealth
					if ( damageToHealthRatio > 0.03 ) then
						if ( HoJ._majorDamageSpells[spellId] == nil ) then
--[[
							HoJ:DebugMsg( "Added major damage spell " .. spellName .. 
									"(" .. spellId .. "): DTHR = " .. 
									string.format( "%.2f", damageToHealthRatio ) )
--]]
						end
						HoJ._majorDamageSpells[spellId] = { _damageToHealthRatio = damageToHealthRatio }
					end
				end
			end
			--[[
			if ( playerIsSource or playerIsTarget ) then
				HoJ:DebugMsg( event .. ": src = " .. sourceName ..
					", target = " .. destName ..
					", spell = " .. spellName ..
					", school = " .. spellSchool .. 
					", actualDamageDone = " .. actualDamageDone ..
					", amount = " .. tostring( amount ) ..
					", overkill = " .. tostring( overkill ) .. 
					", resisted = " .. tostring( resisted ) ..
					", blocked = " .. tostring( blocked ) .. 
					", absorbed = " .. tostring( absorbed ) )
			end
			--]]

			-- add attackers if the player is the source or the target of a damage spell
			if ( playerIsSource ) then
				HoJ:AddRecentAttacker( self._recentMeleeAttackers, destGUID, destName, spellId, spellName )
				if ( isDamage ) then
					HoJ:AddDpsDamage( HoJ._combatDpsStats, actualDamageDone, 1 )
				end
			elseif ( playerIsTarget ) then
				if ( not isSpell ) then
					HoJ:AddRecentAttacker( self._recentMeleeAttackers, sourceGUID, sourceName, spellId, spellName )
				else
					HoJ:AddAttackerIfInMultitargetRange( self._recentMeleeAttackers, sourceGUID, sourceName, spellId, spellName, maxRange )
					if ( spellSchool ~= 1 ) then
						HoJ._lastSpellDamageTime = self._curTime
					end
				end

				if ( isDamage ) then
					HoJ._incomingDamageBuffer:add( { _time = HoJ._curTime, _health = UnitHealth( "player" ), _amount = actualDamageDone} )
				end
			end
		elseif ( isHeal and playerIsTarget ) then
			local amount, overHeal, absorbed, critical = select( commonSelectStart, ... )
--[[
			HoJ:DebugMsg( "isHeal: amount = " .. tostring( amount ) .. 
					", overHeal = " .. tostring( overHeal ) .. 
					", absorbed == " .. tostring( absorbed ) .. 
					", crit = " .. tostring( critical ) )
--]]
			local actualAmountHealed = amount - ( overHeal - absorbed )
			HoJ._incomingHealsBuffer:add( { _time = HoJ._curTime, _health = UnitHealth( "player" ), _amount = actualAmountHealed} )
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:COMBAT_LOG_EVENT_UNFILTERED
-- -----------------------------------------------------------------------------
function HoJ._events:COMBAT_LOG_EVENT_UNFILTERED( timestamp_, hideCaster_, guid_, flags_, raidFlags_ )
	HoJ:HandleCombatLogEvent( CombatLogGetCurrentEventInfo() )	
end

-- -----------------------------------------------------------------------------
-- HoJ:_events:UNIT_SPELLCAST_SUCCEEDED
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_SPELLCAST_SUCCEEDED( unitID, castGUID, spellID ) -- spell, rank, lineID, spellID )
	local spellName = HoJ:GetSpellName( spellID )
	-- local debug = spellID == 26573
	-- HoJ:Log( debug, "UNIT_SPELLCAST_SUCCEEDED: " .. spellName .. " by " .. unitID )
	if ( unitID == "player" ) then

		-- check for an aoe spell
		local aoeSpellTable = HoJ._curSpecInfo._aoeSpellTable
		local aoeBuffTable = HoJ._curSpecInfo._aoeBuffTable
		if ( aoeSpellTable ~= nil ) then
			local aoeSpell = aoeSpellTable[spellID]
			if ( aoeSpell ~= nil ) then
				if ( aoeSpell._buffId ~= nil ) then
					local aoeBuff = aoeBuffTable[aoeSpell._buffId]
					if ( aoeBuff ) then
						local spellInfo = aoeBuff._spellInfo
						aoeBuff._startTime = GetTime()
						aoeBuff._duration = spellInfo._duration
						aoeBuff._expiration = aoeBuff._startTime + aoeBuff._duration
						-- HoJ:Log( debug, "Logged start cast of AoE " .. aoeBuff._spellInfo.name )
					end
				end
			end
		end

		local spellTarget = "target"
		if ( not UnitIsPlayer( "target" ) ) then
			spellTarget = "player"
		end
		--HoJ:DebugMsg( "spellTarget = " .. spellTarget )

		local name, realm = UnitName( spellTarget )
--[[
		if ( realm ~= nil ) then
			HoJ:DebugMsg( spellName .. " cast on player " .. name .. "-" .. realm )
		else
			HoJ:DebugMsg( spellName .. " cast on player " .. name )
		end
--]]
		local activeBlessings = HoJ._curSpecInfo._activeBlessings
		if ( activeBlessings ~= nil ) then
			local blessing = activeBlessings[spellID]
			if ( blessing ~= nil ) then
				blessing.targetName = name
				blessing.start = HoJ._curTime
				blessing.duration = blessing.spellInfo.buffDurationSeconds
				blessing.found = false
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_ENTERED_VEHICLE
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_ENTERED_VEHICLE( unitId )
	--HoJ:DebugMsg( "UNIT_ENTERED_VEHICLE" )
	if ( unitId == "player" ) then
		for i, frame in ipairs( HoJ._windowFrames ) do
			frame:Hide()
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ._events:UNIT_EXITED_VEHICLE
-- -----------------------------------------------------------------------------
function HoJ._events:UNIT_EXITED_VEHICLE( unitId )
	--HoJ:DebugMsg( "UNIT_EXITED_VEHICLE" )
	if ( unitId == "player" ) then
		for i, frame in ipairs( HoJ._windowFrames ) do
			frame:Show()
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:MouseOver
-- -----------------------------------------------------------------------------
function HoJ:MouseOver()
	-- At some point in Cata or Mists, EnableMouse became protected in combat.
	--if ( HoJ._inCombat ) then
	--	return;
	--end

	local mouseFocus = GetMouseFocus();
	for i, v in pairs( HoJ._windowFrames ) do
		if ( MouseIsOver( v ) ) then
			if ( IsControlKeyDown() ) then
				-- HoJ:DebugMsg( "Enabling mouse for frame '" .. v:GetName() .. "'" );
				v:EnableMouse( true );

				if ( IsMouseButtonDown( "LeftButton" ) ) then
					-- HoJ:DebugMsg( "Start moving frame '" .. v:GetName() .. "'" );
					v:StartMoving();
				end
			end
			if ( not IsMouseButtonDown( "LeftButton" ) ) then
				v:StopMovingOrSizing();
			end
		else
			--HoJ:DebugMsg( "Disabling mouse for frame '" .. v:GetName() .. "'" );
			v:EnableMouse( false );
			v:StopMovingOrSizing();
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:EnableFrame
-- -----------------------------------------------------------------------------
function HoJ:EnableFrame( frame )
	table.insert( HoJ._windowFrames, frame )
	frame:Show()
	frame:SetMovable( true )
	frame:EnableMouse( true )
	frame:SetPoint( "CENTER", UIParent )	
--[[
	if ( HoJ._isActive ) then
		table.insert( HoJ._windowFrames, self );
		self:Show();
		self:SetMovable( true );
		self:EnableMouse( true );
		self:SetPoint( "CENTER", UIParent );
	else
		self:Hide();
	end
--]]
end

-- -----------------------------------------------------------------------------
-- HoJ_OnEvent
-- -----------------------------------------------------------------------------
function HoJ_OnEvent( self, event, ... )
	--HoJ:DebugMsg( "event = " .. event );

	-- do nothing for non-paladins
	if ( HoJ._finalized == true and HoJ._isActive == false ) then
		return;
	end
	-- call event functions
	HoJ._events[ event ](self, ... );
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetFrames
-- -----------------------------------------------------------------------------

function HoJ:ResetFrames( frameList )
	-- clear all icon frames
	for key, value in pairs( frameList ) do
		value.frame:Hide()
--[[
		if ( value.texture ~= nil ) then
			value.texture:SetTexture( nil )
			value.texture:Hide()
		end
		if ( value.cooldown ~= nil ) then
			value.cooldown:Hide()
		end
		if ( value.charges ~= nil ) then
			value.charges:Hide()
		end
		value.spellInfo = nil
		--value.text:Hide();
]]
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Reset
-- -----------------------------------------------------------------------------
function HoJ:Reset()
	-- HoJ:DebugMsg( "Reset" );

	self._level = UnitLevel( "player" ) -- handle party sync

	HoJ_InitSpellInfo()

	HoJ:ResetFrames( self._buffFrames )
	HoJ:ResetFrames( self._cooldownFrames )
	HoJ:ResetFrames( self._rotationFrames )

	HoJ:ClearIconList( self._buffs )
	HoJ:ClearIconList( self._cooldowns )
	HoJ:ClearIconList( self._rotationSkills )

	if ( self._curSpecInfo ~= nil and self._curSpecInfo._resetFn ~= nil ) then
		self._curSpecInfo._resetFn( self )
	end

	HoJ:ResetHealth()

	HoJ:ResetDpsStats( HoJ._combatDpsStats, 0.0 )
	HoJ:ResetSessionDps()

	self._needsReset = false
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcDps
-- returns
-- currentDps
-- maxDps
-- isValid
-- -----------------------------------------------------------------------------
function HoJ:CalcDps( curTime, totalTime, damageDone, attacksHit, attacksHitThreshold, timeThreshold, curMaxDps )
	local currentDps = 0.0
	local isValid = false
	local newMaxDps = curMaxDps

	if ( totalTime > 0.0 ) then
		currentDps = damageDone / totalTime
		isValid = ( attacksHit > attacksHitThreshold ) and ( totalTime > timeThreshold )
		if ( currentDps > curMaxDps ) then
			newMaxDps = currentDps
		end
	end

	return currentDps, newMaxDps, isValid
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcDpsForStats
-- updates recalculates dps based on the stats in a dps stats table
-- -----------------------------------------------------------------------------
function HoJ:CalcDpsForStats( curTime, dpsStats )
	if ( dpsStats == nil ) then
		HoJ:Error( "dpsStats is nil!" )
		return
	end
	local totalTime = dpsStats._totalTime -- only session stats have this
	if ( totalTime == nil ) then
		totalTime = curTime - dpsStats._startTime
	end
	if ( totalTime == nil ) then
		HoJ:Error( "totalTime is nil!" )
	end
	local currentDps, maxDps, isValid = HoJ:CalcDps( 
			curTime, 
			totalTime, 
			dpsStats._damageDone, 
			dpsStats._attacksHit, 
			dpsStats._attacksHitThreshold, 
			dpsStats._timeThreshold, 
			dpsStats._maxDps )
	dpsStats._dps = currentDps
	dpsStats._maxDps = maxDps
	dpsStats._isValid = isValid
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetDpsStats
-- -----------------------------------------------------------------------------
function HoJ:ResetDpsStats( dpsStats, startTime )
	if ( dpsStats == nil ) then
		HoJ:Error( "dpsStats is nil!" )
		return
	end
	dpsStats._isValid = false
	dpsStats._startTime = startTime
	dpsStats._damageDone = 0
	dpsStats._attacksHit = 0
	dpsStats._dps = 0
	dpsStats._maxDps = 0
	dpsStats._targetTimeToLive = -1.0
	dpsStats._firstPredictedTargetDeathTime = -1.0
	dpsStats._predictedTargetDeathTime = -1.0
end

-- -----------------------------------------------------------------------------
-- HoJ:PauseDpsStats
-- -----------------------------------------------------------------------------
function HoJ:PauseDpsStats( dpsStats, curTime )
	if ( dpsStats == nil ) then
		HoJ:Error( "dpsStats is nil!" )
		return
	end

	dpsStats._isValid = false
	dpsStats._startTime = startTime
	dpsStats._damageDone = 0
	dpsStats._attacksHit = 0
	dpsStats._dps = 0
	dpsStats._maxDps = 0
	dpsStats._targetTimeToLive = -1.0
	dpsStats._firstPredictedTargetDeathTime = -1.0
	dpsStats._predictedTargetDeathTime = -1.0
end

-- -----------------------------------------------------------------------------
-- HoJ:AddDpsDamage
-- -----------------------------------------------------------------------------
function HoJ:AddDpsDamage( dpsStats, damage, hits )
	if ( dpsStats == nil ) then
		HoJ:Error( "dpsStats is nil!" )
		return
	end
	dpsStats._damageDone = dpsStats._damageDone + damage
	dpsStats._attacksHit = dpsStats._attacksHit + hits
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcTimeToLive
-- returns the time it will take currentHealth to reach 0 if taking damage at
-- the specified dps.
-- If useAverage == true, then the ttl is calculated once with current dps and
-- once with max dps, and the average of the two times is returned
-- -----------------------------------------------------------------------------
function HoJ:CalcTimeToLive( dpsStats, currentHealth, useAverage )
	if ( dpsStats == nil or not dpsStats._isValid ) then
		return -1.0
	end
	local ttl = currentHealth / dpsStats._dps
	if ( useAverage ) then
		ttl2 = currentHealth / dpsStats._maxDps
		return ( ttl + ttl2 ) * 0.5
	end
	return ttl
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateDpsStats
-- returns:
--   targetTimeToLive
-- -----------------------------------------------------------------------------
function HoJ:UpdateDpsStats( curTime, dpsStats, targetExists, targetHealth )
	local wasValid = dpsStats._isValid

	HoJ:CalcDpsForStats( curTime, dpsStats )

	local timeToLive = -1.0
	if ( targetExists ) then
		-- use combat DPS stats first, if valid
		timeToLive = HoJ:CalcTimeToLive( HoJ._combatDpsStats, targetHealth, true )
		if ( timeToLive < 0.0 ) then
			-- if combat DPS stats aren't valid, use session dps stats
			timeToLive = HoJ:CalcTimeToLive( HoJ._sessionDpsStats, targetHealth, true )
		end
	end

	if ( timeToLive >= 0.0 ) then
		dpsStats._targetTimeToLive = timeToLive
		dpsStats._predictedTargetDeathTime = curTime + timeToLive
		if ( dpsStats._firstPredictedTargetDeathTime < 0 ) then 
			dpsStats._firstPredictedTargetDeathTime = curTime + timeToLive
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcSessionDps
-- returns
--   sessionDps - averaged dps across all tracked combat sessions
--   totalDuration - total time for all tracked combat sessions
--   totalAttacksHit - number of attacks hit across all tracked combat sessions
-- -----------------------------------------------------------------------------
function HoJ:CalcSessionDps( sessionDps )
	if ( HoJ._dpsBuffer == nil ) then
		-- HoJ:DebugMsg( "No session dps" )
		HoJ:ResetDpsStats( sessionDps, 0.0 )
		return false
	end

	local dpsBuffer = HoJ._dpsBuffer

	local totalDamageDone = 0
	local totalDuration = 0.0
	local totalAttacksHit = 0
	local maxDps = 0.0

	local size = dpsBuffer.maxSize
	local count = dpsBuffer:count()

	for i = 0, ( count - 1 ) do
		local index = ( dpsBuffer.tail + i ) % size
		local dpsInfo = dpsBuffer[index]
		if ( dpsInfo == nil ) then
			HoJ:Error( "dpsInfo is nil!" )
		else
			-- HoJ:DebugMsg( "dpsBuffer[" .. index .. "] is valid" )
			totalDamageDone = totalDamageDone + dpsInfo._damageDone
			totalDuration = totalDuration + dpsInfo._duration
			totalAttacksHit = totalAttacksHit + dpsInfo._attacksHit
			if ( totalDuration > 0.0 ) then
				local dps = totalDamageDone / totalDuration
				if ( dps > maxDps ) then
					maxDps = dps
				end
			end
		end
	end
	
	sessionDps._totalTime = totalDuration
	sessionDps._damageDone = totalDamageDone
	sessionDps._attacksHit = totalAttacksHit

	-- HoJ:DebugMsg( "Dps sessions = " .. count .. ", damage = " .. totalDamageDone .. ", time =" .. totalDuration )
	if ( totalDuration == 0.0 ) then
		HoJ:ResetDpsStats( sessionDps, 0.0 )
	else
		HoJ:CalcDpsForStats(  HoJ._curTime, sessionDps )
	end

	return sessionDps._isValid
end

-- -----------------------------------------------------------------------------
-- HoJ:AddSessionDps
-- -----------------------------------------------------------------------------
function HoJ:AddSessionDps( curTime, dpsStats )
	-- HoJ:DebugMsg( "AddSessionDps" )
	if ( HoJ._dpsBuffer == nil ) then
		-- save the last 10 combat sessions dps to keep a running dps average for the session
		HoJ._dpsBuffer = HoJ.CircularBuffer:new( 20 )
		--HoJ:DebugMsg( "Created circular buffer." )
	elseif ( HoJ._dpsBuffer:isFull() ) then
		-- we have to free or the next add will fail
		HoJ._dpsBuffer:remove()
	end

	-- if the dps stats aren't valid, don't record them
	if ( not dpsStats._isValid ) then
		return
	end

	local dpsInfo = {
		_damageDone = dpsStats._damageDone,
		_duration = curTime - dpsStats._startTime,
		_attacksHit = dpsStats._attacksHit,
	}
	if ( dpsInfo._damageDone == nil ) then
		HoJ:DebugMsg( "_damageDone is nil!" )
	elseif ( dpsInfo._duration == nil ) then
		HoJ:Error( "_duration is nil!" )
	elseif ( dpsInfo._attacksHit == nil ) then
		HoJ:Error( "_attacksHit is nil!" )
	end

	HoJ._dpsBuffer:add( dpsInfo )
	
	--HoJ:DebugMsg( "dps buffer has " .. HoJ._dpsBuffer:count() .. " entries." )
	
	HoJ:CalcSessionDps( HoJ._sessionDpsStats )
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetSessionDps
-- -----------------------------------------------------------------------------
function HoJ:ResetSessionDps()
	if ( HoJ._dpsBuffer ~= nil ) then
		HoJ._dpsBuffer:empty()
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcTargetToughness
-- -----------------------------------------------------------------------------
function HoJ:CalcTargetToughness( targetExists, 
		targetIsDead, 
		targetIsBoss, 
		targetIsTrivial,
		targetClassification, 
		targetLevel, 
		targetHealth, 
		targetMaxHealth, 
		dpsStats )

	local targetIsTough = false

	if ( targetExists and not targetIsDead ) then
		local healthRatio = ( targetHealth / HoJ._playerHealth )
		local maxHealthRatio = ( targetMaxHealth / HoJ._playerMaxHealth )
--[[
		HoJ:DebugMsg( "targetIsBoss = " .. tostring( targetIsBoss ) ..
				", targetClass = " .. targetClassification ..
				", targetLevel = " .. targetLevel ..
				string.format( ", healthRatio = %.1f", healthRatio ) ..
				string.format( ", maxHealthRatio = %.1f", maxHealthRatio ) )
--]]
		targetIsTough = not targetIsTrivial and ( targetIsBoss
						or targetClassification == "rareelite"
						or targetClassification == "elite"
						or targetLevel > HoJ._level + 3
						or ( healthRatio > 0.8 )
						or ( maxHealthRatio > 1.0 ) )

		if ( not targetIsTough and dpsStats._isValid ) then
			-- HoJ:DebugMsg( "Target TTL = " .. dpsStats._targetTimeToLive )
			if ( dpsStats._targetTimeToLive > 30.0 ) then
				targetIsTough = true
			end
		end
		-- HoJ:DebugMsg( "Target is tough(2) = " .. tostring( targetIsTough ) )
	end

	return targetIsTough
end

-- -----------------------------------------------------------------------------
-- HoJ:IsEnabled
-- -----------------------------------------------------------------------------
function HoJ:IsEnabled( levelCheck )
	if ( levelCheck and self._level < 10 ) then
		return false
	end
	if ( self._enabled == false or self._curSpecInfo == nil ) then
		return false
	end
	return self._curSpecInfo._enabled
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateTarget
-- -----------------------------------------------------------------------------
function HoJ:UpdateTarget( curTime, inVehicle )
	local target = "target";
	local targetIsDead = UnitIsDead( target )
	local targetCanAttack = false
	local targetCanBeInterrupted = false
	local targetHealth = 0;
	local targetMaxHealth = 0
	local targetHealthRatio = 0.0;
	local targetLevel = 1;
	local targetClassification = "";
	local targetIsBoss = false
	local targetIsTrivial = true
	local isTankingTarget = UnitDetailedThreatSituation("player", "target")
	if ( isTankingTarget == nil ) then
		isTankingTarget = false
	end
	local targetExists = UnitExists( target );

	-- cache target stuff if we have a valid target
	if ( targetExists ) then
		targetIsTrivial = UnitIsTrivial( target ) 

		if ( not targetIsDead and UnitCanAttack( "player", target ) ) then
			targetCanAttack = true;
		else
			targetCanAttack = false;
		end

		if ( targetCanAttack ) then
			local channelName, _, _, _, _, _, channelNotInterruptible, channelSpellId = UnitChannelInfo( target )
			local castName, _, _, _, _, _, _, castNotInterruptible, castSpellId = UnitCastingInfo( target )

			local spellId = nil
			local spellName = ""

			if ( channelName ~= nil ) then
				targetCanBeInterrupted = not channelNotInterruptible
				spellId = channelSpellId
				spellName = channelName
			elseif ( castName ~= nil ) then
				targetCanBeInterrupted = not castNotInterruptible
				spellId = castSpellId
				spellName = castName
			end
--[[
			HoJ:DebugMsg( "isTankingTarget = " .. tostring( isTankingTarget ) ..
					", spellId = " .. tostring( spellId ) ..
					", inVehicle = " .. tostring( inVehicle ) )
--]]
			if ( isTankingTarget and spellId ~= nil and not inVehicle ) then
				local majorDamageSpell = HoJ._majorDamageSpells[spellId]
				if ( majorDamageSpell ~= nil ) then
					if ( HoJ._incomingMajorDamage == nil or HoJ._incomingMajorDamage._spellId ~= spellId ) then
						HoJ._incomingMajorDamage = { _name = spellName, _spellId = spellId, _damageToHealthRatio = majorDamageSpell._damageToHealthRatio }
						-- HoJ:DebugMsg( "Incoming major damage: " .. spellName .. ", DTHR: " .. string.format( "%.2f", majorDamageSpell._damageToHealthRatio ) )
					end
				end
			else
				HoJ._incomingMajorDamage = nil
			end
		end

		targetHealth = UnitHealth( target )
		targetMaxHealth = UnitHealthMax( target )
		targetHealthRatio = targetHealth / targetMaxHealth
		targetLevel = UnitLevel( target )
		targetClassification = UnitClassification( target )

		HoJ:FindRanges( "target", HoJ._curSpecInfo )

		-- target toughtness rating
		targetIsBoss = self:IsBossUnit( target, targetClassification )
	end

	HoJ._targetCanAttack = targetCanAttack
	HoJ._targetCanBeInterrupted = targetCanBeInterrupted
	HoJ._targetHealth = targetHealth
	HoJ._targetMaxHealth = targetMaxHealth
	HoJ._targetHealthRatio = targetHealthRatio
	HoJ._targetLevel = targetLevel
	HoJ._targetClassification = targetClassification
	HoJ._targetIsBoss = targetIsBoss

	HoJ:UpdateDpsStats( curTime, HoJ._combatDpsStats, targetExists, HoJ._targetHealth )

	HoJ._targetIsTough = HoJ:CalcTargetToughness( targetExists, targetIsDead, targetIsBoss,
			targetIsTrivial, targetClassification, targetLevel, targetHealth, targetMaxHealth,
			HoJ._combatDpsStats )
end

-- -----------------------------------------------------------------------------
-- HoJ:Update
-- -----------------------------------------------------------------------------
function HoJ:Update()
	self._updateFrameNum = self._updateFrameNum + 1

	HoJ:BroadcastVersion()

	if ( not HoJ:IsEnabled( true ) ) then
		HoJ:ResetFrames( HoJ._buffFrames )
		HoJ:ResetFrames( HoJ._cooldownFrames )
		HoJ:ResetFrames( HoJ._rotationFrames )
		return
	end

	--HoJ:DebugMsg( "HoJ_Update" );
	-- update holy power amount
	HoJ._power = UnitPower( "player", HoJ._classSpecInfo._powerType )
	local usesComboPoints = HoJ._classSpecInfo._usesComboPoints 
	if ( usesComboPoints ~= nil and usesComboPoints == true ) then
		HoJ._comboPoints = UnitPower( "player", 4 )
	else
		HoJ._comboPoints = 0
	end
	if ( IsStealthed() == true ) then
		HoJ._isStealthed = true
	else
		HoJ._isStealthed = false
	end

	-- we don't display our own cooldown text at some points when OmniCC is loaded
	if ( HoJ._omniCC == false ) then
		if ( IsAddOnLoaded( "OmniCC" ) ) then
			HoJ._omniCC = true
			--HoJ:DebugMsg( "OmniCC is loaded. HammerOfJustas will hide OmniCC's inquisition timer." );
		else
			HoJ._omniCC = false
		end
	end

	HoJ:UpdateHealth2( HoJ._curTime )

	if ( HoJ._showDebugFrame ) then
		HoJ_DebugFrame:Show()
		HoJ_DebugTexture:Show()
		HoJ_DebugFrame:SetAlpha( HoJ._debugAlpha * HoJ._globalAlpha )
		HoJ_DebugTexture:SetAlpha( HoJ._debugAlpha * HoJ._globalAlpha )
		HoJ_DebugText:SetFont( "Fonts\\ARIALN.ttf", 12, "OUTLINE" )
		
		local sessionTotalTime = 0.0
		local sessionAttacksHit = 0
		local sessionDps = 0.0
		if ( HoJ._sessionDpsStats._isValid ) then
			sessionTotalTime = HoJ._sessionDpsStats._totalTime
			sessionDps = HoJ._sessionDpsStats._dps
			sessionAttacksHit = HoJ._sessionDpsStats._attacksHit
		end
--[[
		HoJ_DebugText:SetText( "Attackers: " .. HoJ._numMeleeAttackers .. 
				" SDPS: " .. string.format( "%.1f", sessionDps ) ..
				" SDPS Time: " .. string.format( "%.0f", sessionTotalTime ) ..
				" SDPS Hits: " .. sessionAttacksHit ..
				" DPS: " .. string.format( "%.0f", HoJ._combatDpsStats._dps ) ..
				" TTTL: " .. string.format( "%.2f", HoJ._combatDpsStats._targetTimeToLive ) ..
				" SPD: " .. string.format( "%.1f", ( GetUnitSpeed("player") / 0.07 ) ) ..
				" PPS: " .. string.format( "%.1f", HoJ._powerPerSec ) ..
				" PGI: " .. string.format( "%.1f", HoJ._powerGainInterval ) ..
				" PGS: " .. string.format( "%.1f", HoJ._powerGainStep ) ..
				" NPGT: " .. string.format( "%1.f", HoJ._nextPowerGainTime - HoJ._curTime ) )
--]]
		HoJ_DebugText:SetText( "Attackers: " .. HoJ._numMeleeAttackers .. 
			" DPS: " .. string.format( "%.0f", HoJ._combatDpsStats._dps ) ..
			" SPD: " .. string.format( "%.1f", ( GetUnitSpeed("player") / 0.07 ) ) ..
			"\n PHR: " .. string.format( "%.2f", HoJ._playerHealthRatio ) ..
			" IDPSR: " .. string.format( "%.2f", HoJ._incomingDpsRatio ) ..
			" NIDPSR: " .. string.format( "%.2f", ( HoJ._netIncomingDps / HoJ._playerMaxHealth ) ) ..
			" NIDPSCHR: " .. string.format( "%.2f", ( HoJ._netIncomingDps / HoJ._playerHealth ) ) ..
			"\n PTTL: " .. string.format( "%.1f", HoJ._playerTimeToLive ) ..
			" PTTND: " .. string.format( "%.1f", HoJ._playerTimeToNearDeath ) )
	else
		HoJ_DebugFrame:Hide()
	end		

	HoJ.ProfileFunction( "UpdateRecentMeleeAttackers", HoJ.UpdateRecentMeleeAttackers, self )
	
	HoJ.ProfileFunction( "UpdateTarget", HoJ.UpdateTarget, self, HoJ._curTime, HoJ._playerInVehicle )

	if ( UnitIsDeadOrGhost( "player" ) ) then
		--HoJ:DebugMsg( "Player is dead" )
		HoJ:ClearIconList( HoJ._buffs )
		HoJ:ClearIconList( HoJ._rotationSkills )
		HoJ:ClearIconList( HoJ._cooldowns )
		HoJ.ProfileFunction( "UpdateIcons:Rotation:", HoJ.UpdateIcons, self, HoJ._rotationSkills, HoJ._rotationFrames, false );
		HoJ.ProfileFunction( "UpdateIcons:Cooldowns:", HoJ.UpdateIcons, self, HoJ._cooldowns, HoJ._cooldownFrames, true )
		HoJ.ProfileFunction( "UpdateIcons:Buffs:", HoJ.UpdateIcons, self, HoJ._buffs, HoJ._buffFrames, true )
		return
	end

	local updateCooldowns = HoJ._updateType == HoJ.UPDATE_COOLDOWNS
	local updateBuffs = HoJ._updateType == HoJ.UPDATE_BUFFS
	local updateRotation = HoJ._updateType == HoJ.UPDATE_ROTATION

	-- rotate through the update types
	HoJ._updateType = HoJ._updateType + 1
	if ( HoJ._updateType > HoJ.UPDATE_MAX ) then
		HoJ._updateType = HoJ.UPDATE_START
	end

	-- HoJ:DebugMsg( "Updating spec " .. HoJ._curSpecInfo._name )
	if ( updateCooldowns or updateRotation or updateBuffs ) then
		HoJ:FindActiveBuffs( "player", HoJ._curSpecInfo._activeBuffList )
	end

	HoJ._curSpecInfo._updateFn( self, HoJ._curSpecInfo, updateCooldowns, updateBuffs, updateRotation )
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateGlobalCooldown
-- -----------------------------------------------------------------------------
function HoJ:UpdateGlobalCooldown()
	local start, duration, enabled = GetSpellCooldown( HoJ.SpellInfo.globalCooldown.id );
	if ( start > 0 and duration > 0 ) then
		HoJ._gcdDuration = duration;
		HoJ._gcdRemaining = ( start + duration ) - HoJ._curTime;
	else
		HoJ._gcdRemaining = 0;
	end

	--HoJ:DebugMsg( "Global Cooldown: duration = " .. HoJ._gcdDuration .. ", remaining = " .. HoJ._gcdRemaining );
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateRender
-- -----------------------------------------------------------------------------
function HoJ:UpdateRender( frame, elapsed )
	self._renderFrameNum = self._renderFrameNum + 1

	local energyFrame = self._rotationEnergyFrame

	if ( HoJ:IsEnabled( true ) ) then
		if ( self._curSpecInfo._updateRenderFn ~= nil ) then
			self._curSpecInfo._updateRenderFn( self, frame, elapsed )
		end

		if ( self._powerBarMax > 0.0 and self._classSpecInfo ~= nil ) then
			local powerGen = self._powerBarMax
			-- local power = self._power
			local power = UnitPower( "player", self._classSpecInfo._powerType )
			local maxPower = powerGen
			local curPower = math.min( power, powerGen )
			local ratio = curPower / maxPower
			local top = -1.0 + ratio

			--HoJ:DebugMsg( format( "powerGen = %.2f, power = %.2f, ratio = %.2f", powerGen, power, ratio ) )

			energyFrame.barTexture:Show()
			energyFrame.overlayTexture:Show()
			energyFrame.barTexture:SetTexCoord( 0, 1.0, top, top + 1.0 )
		else
			energyFrame.barTexture:Hide()
			energyFrame.overlayTexture:Hide()
		end
	else
		for i, v in ipairs( HoJ._HoPo )do
			--HoJ:DebugMsg( "Setting alpha " .. i )
			v:SetAlpha( 0.0 );
			HoJ._HoPoGlow[i]:SetAlpha( 0.0 )
        end
        
        HoJ_RotationFrame:SetAlpha( 0.0 )
        
		-- hide tv frame
		HoJFrame.HPMaxed:SetAlpha( 0.0 )
		HoJFrame.HPMaxedGlow:SetAlpha( 0.0 )

		energyFrame.barTexture:Hide()
		energyFrame.overlayTexture:Hide()
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnUpdate
-- -----------------------------------------------------------------------------
function HoJ_OnUpdate( frame, elapsed )
	HoJ._profileFrameNum = HoJ._profileFrameNum + 1
	HoJ._frameNum = HoJ._frameNum + 1;

	-- this prevents weird stuff from happening between the time when the player
	-- leaves the world and returns (for instance, taking a portal)
	if ( HoJ._worldState ~= HoJ.WORLD_STATE_ENTERED ) then
		return
	end

	if ( not HoJ:IsEnabled( true ) ) then
--[[		
		if ( HoJ._needsReset ) then
			HoJ:Reset();
		end
--]]
		return
	end

	if ( HoJ._enableProfiling ) then
		ResetCPUUsage();
	end

	if ( HoJ._playerRaceId == 0 ) then
		_, _, HoJ._playerRaceId = UnitRace( "player" )
	end

	local lastRole = HoJ._role;
	local curSpecInfo = HoJ:UpdateSpec()
	if ( HoJ._role ~= lastRole ) then
		if ( curSpecInfo == nil ) then
			HoJ:Error( HoJ._text["SPEC_UNKNOWN"] )
		else
			HoJ:Msg( HoJ._text["SPEC_CHANGED"] .. HoJ._curSpecInfo._name .. "." )
			if ( not curSpecInfo._enabled ) then
				HoJ:Msg( curSpecInfo._name .. HoJ._text["SPEC_NOT_SUPPORTED"] )
			end
		end
		HoJ._needsReset = true
	end

	local level = UnitLevel( "player" )
	if ( HoJ._level ~= level ) then
		HoJ:Msg( HoJ._text["LEVEL_CHANGED"] .. HoJ._level .. " to " .. level .. "." )
		HoJ._level = level
		HoJ._needsReset = true
	end

	if ( HoJ._needsReset ) then
		HoJ:Reset();
	end

	--HoJ:DebugMsg( "HoJ_OnUpdate" );
	-- do nothing for non-paladins
	local playerInVehicle = UnitInVehicle( "player" )
	if ( playerInVehicle == nil ) then
		HoJ:DebugMsg( "playerInVehicle = " .. tostring( playerInVehicle ) )
		playerInVehicle = false
	end

	HoJ._playerInVehicle = playerInVehicle
	if ( HoJ._isActive == false or playerInVehicle ) then
		return;
	end

	-- update time
	HoJ._curTime = GetTime();
	HoJ:UpdateGlobalCooldown();
	HoJ:MouseOver();

	-- update general logic
	if ( HoJ._curTime >= HoJ._nextUpdateTime ) then
		HoJ._nextUpdateTime = HoJ._curTime + HoJ.UPDATE_INTERVAL

		HoJ.ProfileFunction( "HoJ:Update", HoJ.Update, HoJ )
	end

	-- update render logic
	if ( HoJ._curTime >= HoJ._nextRenderTime ) then
		HoJ._nextRenderTime = HoJ._curTime + HoJ.RENDER_INTERVAL;

		HoJ.ProfileFunction( "HoJ:UpdateRender", HoJ.UpdateRender, HoJ, frame, elapsed )
	end

	if ( HoJ._enableProfiling ) then
		UpdateAddOnCPUUsage();
		local addonMs = GetAddOnCPUUsage( "HammerOfJustas" )
		local updateMs = GetFunctionCPUUsage( HoJ.Update, true )
		local renderMs = GetFunctionCPUUsage( HoJ.UpdateRender, true )
		local eventMs = GetEventCPUUsage()
		--if ( addonMs ~= nil and addonMs > HoJ._profileThreshold ) then
			local metThreshold = HoJ._profileFunction == nil
			for key, value in pairs( HoJ._profileTable ) do
				if ( value ~= nil and value.name ~= nil and value.timeMs >= HoJ._profileThreshold and value.frameNum == HoJ._profileFrameNum and
					( HoJ._profileFunction == nil or ( string.find( value.name, HoJ._profileFunction ) ~= nil ) ) ) then
					HoJ:Msg( "func " .. value.name .. "( " .. value.count .. " ): " .. value.timeMs .. " ms" )
					metThreshold = true
				end
			end
			if ( metThreshold ) then
				HoJ:Msg( "Addon CPU: " .. addonMs .. "ms" )
				HoJ:Msg( "Update: " .. updateMs .. "ms" )
				HoJ:Msg( "Render: " .. renderMs .. "ms" )
			end
		--end
		if ( eventMs ~= nil and eventMs > HoJ._profileThreshold ) then
			HoJ:Msg( "Events: " .. eventMs .. "ms" )
		end
	end
end

function HoJ:FindChild( frame, substr )
	--HoJ:DebugMsg( "FindChild( " .. frame:GetName() .. " )" );
	local children = frame:GetChildren();
	for i, child in ipairs( children ) do
		HoJ:DebugMsg( "Child " .. i .. " is " .. child.GetName() );
		if ( string.find( child:GetName(), substr ) ~= nil ) then
			return child;
		end
	end
	return nil;
end

function HoJ:FindRegion( frame, substr )
	--HoJ:DebugMsg( "FindRegion( " .. frame:GetName() .. " )" );
	local regions = { frame:GetRegions() }
	for i, region in ipairs( regions ) do
		-- HoJ:DebugMsg( "Region " .. i .. " is " .. region.GetName() );
		-- if ( string.find( region:GetName(), substr ) ~= nil ) then
		if ( region and region:GetObjectType() == "FontString" ) then
			return region;
		end
	end
	return nil;
end

function HoJ:SetTitleTextColor( frame, colorR, colorG, colorB, colorA )
	--HoJ:DebugMsg( "SetTitleTextColor( " .. frame:GetName() .. " )" );
	local child = HoJ:FindChild( frame, "TitleText" );
	if ( child ~= nil ) then
		child:SetTextColor( colorR, colorG, colorB, colorA );
	end
end

function HoJ:ShowTitle( frame )
	local region = HoJ:FindRegion( frame, "TitleText", frame:GetRegions() );
	if ( region ~= nil ) then
		region:Show();
	end
end

function HoJ:HideTitle( frame )
	local region = HoJ:FindRegion( frame, "TitleText", frame:GetRegions() );
	if ( region ~= nil ) then
		region:Hide();
	end
end

function HoJ:IsHoJFrame( frame )
	for i, v in ipairs( HoJ._windowFrames ) do
		if ( v == frame ) then
			return true;
		end
	end
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ_OnEnter
-- -----------------------------------------------------------------------------
function HoJ_OnEnter( frame )
	--HoJ:DebugMsg( "HoJ_OnEnter( " .. frame:GetName() .. " )" );
	if ( not HoJ:IsEnabled( false ) ) then
		return
	end

	if ( IsControlKeyDown() ) then
		-- if this isn't one of the parent windows, recurse until we get to one
		-- then, turn on background alpha for all the windows.
		if ( not HoJ:IsHoJFrame( frame ) ) then
			HoJ_OnEnter( frame:GetParent() );
		else
			for i, wf in ipairs( HoJ._windowFrames ) do
				wf:SetBackdropColor( 0.5, 0.5, 0.5, 0.5 );
				--HoJ:SetTitleTextColor( wf, 0.5, 0.5, 0.5, 0.5 );
				HoJ:ShowTitle( wf );
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLeave
-- -----------------------------------------------------------------------------
function HoJ_OnLeave( frame )
	--HoJ:DebugMsg( "HoJ_OnLeave( " .. frame:GetName() .. " )" );
	if ( not HoJ:IsEnabled( false ) ) then
		return;
	end

	-- if this isn't one of the parent windows, recurse until we get to one
	-- then, turn on background alpha for all the windows.
--	if ( not IsControlKeyDown() ) then
	if ( not HoJ:IsHoJFrame( frame ) ) then
		HoJ_OnLeave( frame:GetParent() );
	else
		for i, wf in ipairs( HoJ._windowFrames ) do
			wf:SetBackdropColor( 0.5, 0.5, 0.5, 0.0 );
			--HoJ:SetTitleTextColor( wf, 0.5, 0.5, 0.5, 0.0 );
			HoJ:HideTitle( wf );
		end
	end
--	end
	
	-- frame:StopMovingOrSizing();
end

-- -----------------------------------------------------------------------------
-- HoJ_OnDragStart
-- -----------------------------------------------------------------------------
function HoJ_OnDragStart( frame, button )
	-- HoJ:DebugMsg( "HoJ_OnDragStart down for '" .. frame:GetName() .. "'" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnDragStop
-- -----------------------------------------------------------------------------
function HoJ_OnDragStop( frame )
	-- HoJ:DebugMsg( "HoJ_OnDragStop for '" .. frame:GetName() .. "'" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnMouseDown
-- -----------------------------------------------------------------------------
function HoJ_OnMouseDown( frame, button )
	-- HoJ:DebugMsg( "HoJ_OnMouseDown for '" .. frame:GetName() .. "'" )
	--if ( IsControlKeyDown() and not frame.isLocked ) then
    -- 	frame:StartMoving()
   	--end
end

-- -----------------------------------------------------------------------------
-- HoJ_OnMouseUp
-- -----------------------------------------------------------------------------
function HoJ_OnMouseUp( frame, button )
	-- HoJ:DebugMsg( "HoJ_OnMouseUp for '" .. frame:GetName() .. "'" )
	--frame:StopMovingOrSizing();
end

-- -----------------------------------------------------------------------------
-- HoJ:Initialize
-- -----------------------------------------------------------------------------
function HoJ:Initialize()
	-- HoJ:DebugMsg( "Initialize" )

	HoJ._curTime = GetTime()

	HoJ:InitAddonMessaging()
	
	HoJ._healthBuffer = HoJ.CircularBuffer:new( HoJ.HEALTH_SAMPLES_PER_SEC * HoJ.HEALTH_SAMPLE_SECONDS )

	-- keep track of the last n incoming hits
	HoJ._incomingDamageBuffer = HoJ.CircularBuffer:new( 50 )
	-- keep track of the last n incoming heals
	HoJ._incomingHealsBuffer = HoJ.CircularBuffer:new( 50 )

	-- hook the chat handler
	HoJ:AddMessageFilter()
end

-- -----------------------------------------------------------------------------
-- HoJ:Finalize
-- -----------------------------------------------------------------------------
function HoJ:Finalize( caller )
	if ( HoJ._finalized ) then
		return 
	end

	local specId = GetSpecialization()
	if ( specId == nil ) then
		-- FIXME: this probably fails for newbie toons
		HoJ:DebugMsg( "ERROR: class spec is not set!" )
		return
	end

	local curSpecInfo = nil
	local _, className = UnitClass( "player" )
	local updateClass = HoJ._className ~= className
	if ( updateClass ) then
		-- loading class for the first time
		HoJ:InitClass( className )
		curSpecInfo = HoJ:UpdateSpec()

		local tbl = HoJ._classFunctionTable[className]
		if ( tbl ~= nil ) then
			-- HoJ:DebugMsg( "Calling load function for " .. className )
			tbl._loadFn( self )
		end
	end	

	-- HoJ:DebugMsg( "Finalize from " .. caller .. " succeeded!" )
	HoJ._level = UnitLevel( "player" )
	HoJ._playerName = UnitName( "player" )
	HoJ._playerGUID = UnitGUID( "player" )

	if ( HoJ._debugSpellName ~= nil and HoJ._debugSpellName ~= '' ) then
		HoJ:DebugMsg( "_debugSpellName = '" .. HoJ._debugSpellName .. "'" )
	end

	-- hide frames
	local enabled = HoJ:IsEnabled( true )
	for i, frame in ipairs( HoJ._windowFrames ) do
		--HoJ:DebugMsg( "Checking frame " .. frame:GetName() )
		if ( not enabled ) then
			--HoJ:DebugMsg( "Hiding frame " .. frame:GetName() )
			frame:Hide()
		end
	end

	if ( not HoJ._isActive or curSpecInfo == nil ) then
		HoJ:Msg( HoJ._text["CLASS_NOT_SUPPORTED"] )
		return 
	else
		if ( curSpecInfo._enabled ) then
			HoJ:Msg( HoJ._text["SPECIALIZATION_IS"] .. curSpecInfo._name .. "." )
		else
			HoJ:Msg( curSpecInfo._name .. HoJ._text["SPEC_NOT_SUPPORTED"] )
		end
	end

	HoJ:LoadSavedVars()

	HoJ:SetupAddOnOptions()

	local globalAlpha = HoJ._globalAlpha
	HoJFrame_Cooldowns:SetAlpha( HoJ._cooldownsAlpha * globalAlpha )
	HoJFrame_Rotation:SetAlpha( HoJ._rotationAlpha * globalAlpha )
	HoJ_RotationSecondaryFrame:SetAlpha( HoJ._rotationSecondaryAlpha * globalAlpha )
	HoJFrame_Buffs:SetAlpha( HoJ._buffsAlpha * globalAlpha )
	HoJFrame_Debug:SetAlpha( HoJ._debugAlpha * globalAlpha )

	HoJ:Print( HoJ.ADDON_NAME .. HoJ._text["VERSION"] .. HoJ:GetVersionString() .. HoJ._text["LOADED_FOR"] .. HoJ._text[HoJ._classSpecInfo._className .. "S"] .. "." )
	HoJ:Print( HoJ._text["GREETING_FIRST"] .. HoJ._playerName .. HoJ._text["GREETING_SECOND"] )

	-- one finalized is true we can handle all events
	HoJ._finalized = true

	-- save variables
	HoJFrame:RegisterEvent( "PLAYER_LOGOUT" )

	-- handle player leaving the world
	HoJFrame:RegisterEvent( "PLAYER_LEAVING_WORLD" )

	-- check for entering combat
	HoJFrame:RegisterEvent( "PLAYER_REGEN_DISABLED" )

	-- check for leaving combat
	HoJFrame:RegisterEvent( "PLAYER_REGEN_ENABLED" )

	-- check for successful spellcasts
	HoJFrame:RegisterEvent( "UNIT_SPELLCAST_SUCCEEDED" )
	
	-- monitor damage to determine number of melee attackers 
	-- HoJFrame:RegisterEvent( "UNIT_HEALTH" )
	HoJFrame:RegisterEvent( "COMBAT_LOG_EVENT_UNFILTERED" )

	HoJFrame:RegisterEvent( "UNIT_ENTERED_VEHICLE" )
	HoJFrame:RegisterEvent( "UNIT_EXITED_VEHICLE" )

	HoJFrame:RegisterEvent( "UNIT_POWER_UPDATE" )

	HoJFrame:RegisterEvent( "PLAYER_CONTROL_LOST" )
	HoJFrame:RegisterEvent( "PLAYER_CONTROL_GAINED" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad
-- -----------------------------------------------------------------------------
function HoJ_OnLoad( self )
	-- HoJ:DebugMsg( self:GetName() .. ":OnLoad()" )

	HoJ:EnableFrame( self );

	-- variables loaded
	HoJFrame:RegisterEvent( "ADDON_LOADED" );

	-- player has logged in
	HoJFrame:RegisterEvent( "PLAYER_LOGIN" )

	-- handle player entering the world
	HoJFrame:RegisterEvent( "PLAYER_ENTERING_WORLD" )

	HoJ:Initialize( "HoJ_OnLoad" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnAddonLoaded
-- -----------------------------------------------------------------------------
function HoJ_OnAddonLoaded( self )
	-- HoJ:DebugMsg( "OnAddonLoaded()" )

	-- HoJ:Finalize( "HoJ_OnAddonLoaded()" )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_HoPo
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_HoPo( self )
	-- HoJ:DebugMsg( self:GetName() .. ":OnLoad()" )

	HoJ:EnableFrame( self );

	HoJ._HoPo = { self.HP1, self.HP2, self.HP3, self.HP4, self.HP5 }
	HoJ._HoPoGlow = { self.HP1Glow, self.HP2Glow, self.HP3Glow, self.HP4Glow, self.HP5Glow }
--[[	
	HoJ._HoPo[0] = self.HP1;
	HoJ._HoPoGlow[0] = self.HP1Glow;
	HoJ._HoPo[1] = self.HP2;
	HoJ._HoPoGlow[1] = self.HP2Glow;
	HoJ._HoPo[2] = self.HP3;
	HoJ._HoPoGlow[2] = self.HP3Glow;
	HoJ._HoPo[3] = self.HP4;
	HoJ._HoPoGlow[3] = self.HP4Glow;
	HoJ._HoPo[4] = self.HP5;
	HoJ._HoPoGlow[4] = self.HP5Glow;
--]]	
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Buffs
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Buffs( self )
	-- HoJ:DebugMsg( self:GetName() .. ":OnLoad()" )

	HoJ:EnableFrame( self );

	HoJ._buffFrames = {
		{
			frame = HoJ_Buff1Frame,
			texture = HoJ_Buff1Texture,
			spellInfo = nil,
			text = HoJ_Buff1Text,
			_timeLeftFontString = HoJ_Buff1TimeLeft,
			_stacksFontString = HoJ_Buff1Stacks,
		},
		{
			frame = HoJ_Buff2Frame,
			texture = HoJ_Buff2Texture,
			spellInfo = nil,
			text = HoJ_Buff2Text,
			_timeLeftFontString = HoJ_Buff2TimeLeft,
			_stacksFontString = HoJ_Buff2Stacks,
		},
		{
			frame = HoJ_Buff3Frame,
			texture = HoJ_Buff3Texture,
			spellInfo = nil,
			text = HoJ_Buff3Text,
			_timeLeftFontString = HoJ_Buff3TimeLeft,
			_stacksFontString = HoJ_Buff3Stacks,
		},
		{
			frame = HoJ_Buff4Frame,
			texture = HoJ_Buff4Texture,
			spellInfo = nil,
			text = HoJ_Buff4Text,
			_timeLeftFontString = HoJ_Buff4TimeLeft,
			_stacksFontString = HoJ_Buff4Stacks,
		},
		{
			frame = HoJ_Buff5Frame,
			texture = HoJ_Buff5Texture,
			spellInfo = nil,
			text = HoJ_Buff5Text,
			_timeLeftFontString = HoJ_Buff5TimeLeft,
			_stacksFontString = HoJ_Buff5Stacks,
		}
	}
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Cooldowns
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Cooldowns( self )
	-- HoJ:DebugMsg( self:GetName() .. ":OnLoad()" )

	HoJ:EnableFrame( self );

	HoJ._cooldownFrames = {
		{
			frame = HoJ_Cooldown1Frame,
			texture = HoJ_Cooldown1Texture,
			spellInfo = nil,
			text = HoJ_Cooldown1Text
		},
		{
			frame = HoJ_Cooldown2Frame,
			texture = HoJ_Cooldown2Texture,
			spellInfo = nil,
			text = HoJ_Cooldown2Text
		},
		{
			frame = HoJ_Cooldown3Frame,
			texture = HoJ_Cooldown3Texture,
			spellInfo = nil,
			text = HoJ_Cooldown3Text
		},
		{
			frame = HoJ_Cooldown4Frame,
			texture = HoJ_Cooldown4Texture,
			spellInfo = nil,
			text = HoJ_Cooldown4Text
		},
		{
			frame = HoJ_Cooldown5Frame,
			texture = HoJ_Cooldown5Texture,
			spellInfo = nil,
			text = HoJ_Cooldown5Text
		}
	};
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Rotation
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Rotation( self )
	-- HoJ:DebugMsg( self:GetName() .. ":OnLoad()" )

	HoJ:EnableFrame( self );

	HoJ._rotationFrames = {
		{
			frame = HoJ_RotationFrame,
			texture = HoJ_RotationTexture,
			spellInfo = nil,
			text = HoJ_RotationText,
			_cooldownFrame = HoJ_RotationCooldown,
			_chargesFontString = HoJ_RotationCharges
		},
		{
			frame = HoJ_RotationSecondaryFrame,
			texture = HoJ_RotationSecondaryTexture,
			spellInfo = nil,
			text = HoJ_RotationSecondaryText,
			_cooldownFrame = HoJ_RotationSecondaryCooldown,
			_chargesFontString = HoJ_RotationSecondaryCharges
		}
	}

	HoJ._rotationEnergyFrame = {
		barTexture = HoJ_EnergyBarBase,
		overlayTexture = HoJ_EnergyBarOverlay
	}

	HoJ._rotationEnergyFrame.barTexture:SetTexture( "Interface\\AddOns\\HammerOfJustas\\Images\\energy_bar_8x64.tga", "CLAMPTOBLACKADDITIVE", "CLAMPTOBLACKADDITIVE" )
	--HoJ._rotationEnergyFrame.barTexture:SetTexture( "Interface\\AddOns\\HammerOfJustas\\Images\\square_128.tga" )
	--HoJ._rotationEnergyFrame.barTexture:SetPoint( "BOTTOMLEFT" )
	--HoJ._rotationEnergyFrame.barTexture:SetVertTile( true )
	--HoJ._rotationEnergyFrame.barTexture:SetVertTile( true )
end

-- -----------------------------------------------------------------------------
-- HoJ_OnLoad_Debug
-- this is pre-WoD stuff
-- FIXME: remove the call to this from window template
-- -----------------------------------------------------------------------------
function HoJ_OnLoad_Debug( self )
	-- HoJ:DebugMsg( self:GetName() .. ":OnLoad()" )

	HoJ:EnableFrame( self );

	local fontName, fontHeight, fontFlags = HoJ_DebugText:GetFont();
	HoJ_DebugText:SetFont( fontName, fontHeight, "THICKOUTLINE" );
	HoJ_DebugText:SetTextColor( 1.0, 1.0, 1.0 );
end
