-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Utils.lua
-- DESC: Addon utility functions.

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Utils.lua" )

-- ============================================================================

-- ============================================================================
-- Profiling

HoJ._profileTable = { }
HoJ._profileFrameNum = 0
HoJ._enableProfiling = false
HoJ._profileThreshold = 0.3 -- threshold in ms before showing CPU time
HoJ._profileFunction = nil

-- ----------------------------------------------------------------------------
-- HoJ.ProfileBegin
-- ----------------------------------------------------------------------------

function HoJ.ProfileBegin( name_, key )
	if ( HoJ._profileTable[key] == nil ) then
		HoJ._profileTable[key] = { name = name_, startMs = 0.0, endMs = 0.0, timeMs = 0.0, count = 0 }
	end
	HoJ._profileTable[key].startMs = debugprofilestop()
end

-- ----------------------------------------------------------------------------
-- HoJ.ProfileBegin2
-- ----------------------------------------------------------------------------
function HoJ.ProfileBegin2( name )
	HoJ.ProfileBegin( name, name )
end

-- ----------------------------------------------------------------------------
-- HoJ.ProfileEnd
-- ----------------------------------------------------------------------------
function HoJ.ProfileEnd( key, startTime, spam )
	local endMs = debugprofilestop()

	local entry = HoJ._profileTable[key]
	entry.endMs = endMs
	entry.timeMs = entry.endMs - entry.startMs
	entry.count = entry.count + 1
	entry.frameNum = HoJ._profileFrameNum
	if ( spam ~= nil ) then
		DEFAULT_CHAT_FRAME:AddMessage( "ProfileEnd( " .. HoJ._profileFrameNum .. " ): " .. key .. " = " .. entry.timeMs .. " ms" );
	end
end

-- ----------------------------------------------------------------------------
-- HoJ.ProfileFunction
-- ----------------------------------------------------------------------------

function HoJ.ProfileFunction( name, func, ... )
	local startTime = HoJ.ProfileBegin( name, name )
	func( ... )
	HoJ.ProfileEnd( name, startTime )
end

-- ----------------------------------------------------------------------------
-- HoJ.ProfileFunction
-- ----------------------------------------------------------------------------
function HoJ.ProfileClear( name )
	local entry = HoJ._profileTable[name]
	entry.timeMs = 0
end

-- -----------------------------------------------------------------------------
-- ToStr
-- -----------------------------------------------------------------------------
function ToStr( value )
	if ( value == nil ) then
		return "nil"
	end
	return tostring( value )
end

-- -----------------------------------------------------------------------------
-- TableMemberToStr
-- -----------------------------------------------------------------------------
function TableMemberToStr( tbl, member )
	if ( tbl == nil ) then
		return "nil table"
	elseif ( tbl[member] == nil ) then
		return "nil"
	end
	return tostring( tbl[member] )
end

-- -----------------------------------------------------------------------------
-- HoJ:SplitString
-- -----------------------------------------------------------------------------
function HoJ:SplitString( inStr, separator )
	if type( inStr ) ~= "string" then
		HoJ:Error( "parsed parameter is not a table." )
		return
	end

	if ( separator == nil ) then
		separator = "%s"
	end

	local out = { }
	for str in string.gmatch( inStr, "([^" .. separator .. "]+)" ) do
		table.insert( out, str )
	end
	return out
end

-- -----------------------------------------------------------------------------
-- HoJ:GetVersionString
-- -----------------------------------------------------------------------------
function HoJ:GetVersionString()
	local version = "" .. HoJ.MAJOR_VERSION .. "." .. HoJ.MINOR_VERSION .. "." .. HoJ.BUILD_VERSION .. "-" .. HoJ.BUILD_TYPE
	return version
end

-- -----------------------------------------------------------------------------
-- HoJ:Log
-- -----------------------------------------------------------------------------
function HoJ:Log( show, text )
	if ( show and DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ(" .. HoJ.STR_GREEN .. string.format( "%.2f", HoJ._curTime ) .. HoJ.STR_YELLOW .. "): " .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:FTLog
-- -----------------------------------------------------------------------------
function HoJ:FTLog( numFrames, show, text )
	if ( ( HoJ._frameNum % numFrames ) ~= 0 ) then
		return
	end
	HoJ:Log( show, text )
end

-- -----------------------------------------------------------------------------
-- HoJ:DebugMsg
-- -----------------------------------------------------------------------------
function HoJ:DebugMsg( text )
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ(" .. HoJ.STR_GREEN .. string.format( "%.2f", HoJ._curTime ) .. HoJ.STR_YELLOW .. "): " .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ChannelInfo
-- -----------------------------------------------------------------------------
function HoJ:ChannelInfo( text )
--[[
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_GREEN .. "HoJ." .. HoJ.STR_WHITE .. text )
	end
--]]
end

-- -----------------------------------------------------------------------------
-- HoJ:ChannelError
-- -----------------------------------------------------------------------------
function HoJ:ChannelError( text )
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_RED .. "HoJ." .. HoJ.STR_WHITE .. text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Error
-- -----------------------------------------------------------------------------
function HoJ:Error( text )
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_RED .. "HoJ." .. HoJ.STR_WHITE .. text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Error
-- -----------------------------------------------------------------------------
function HoJ:FrameThrottleError( numFrames, text )
	if ( ( HoJ._frameNum % numFrames ) ~= 0 ) then
		return
	end
	if ( DEFAULT_CHAT_FRAME ) then
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_RED .. "HoJ." .. HoJ.STR_WHITE .. text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ThrottleMsg
-- -----------------------------------------------------------------------------
function HoJ:ThrottleMsg( interval, text )
	if ( DEFAULT_CHAT_FRAME ) then
--[[
		local msg = HoJ._throttledMessages[text]
		if ( msg == nil ) then
			msg = {
				lastTime = self._curTime
			}
			HoJ._throttledMessages[text] = msg
		else
			local timeSincePrinted = self._curTime - msg.lastTime
			if ( timeSincePrinted < interval ) then
				return
			end
		end

		msg.lastTime = self._curTime
--]]
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:FrameThrottleMsg
-- -----------------------------------------------------------------------------
function HoJ:FrameThrottleMsg( numFrames, text )
	if ( DEFAULT_CHAT_FRAME ) then
--[[
		local msg = HoJ._throttledMessages[text]
		if ( msg == nil ) then
			msg = {
				lastTime = self._curTime
			}
			HoJ._throttledMessages[text] = msg
		else
			local timeSincePrinted = self._curTime - msg.lastTime
			if ( timeSincePrinted < interval ) then
				return
			end
		end

		msg.lastTime = self._curTime
--]]
		if ( ( HoJ._frameNum % numFrames ) ~= 0 ) then
			return
		end
		-- without frame number spam
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text )
		-- to enable frame number spam
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Warn
-- -----------------------------------------------------------------------------
function HoJ:Warn( text )
	-- should be commented out for releases
	if ( DEFAULT_CHAT_FRAME ) then
		--DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_PURPLE .. "(" ..HoJ._frameNum .. "):" .. HoJ.STR_YELLOW .. "HoJ." .. HoJ.STR_WHITE .. text );
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.STR_RED .. "HoJ." .. HoJ.STR_WHITE .. text );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Print
-- -----------------------------------------------------------------------------
function HoJ:Print( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Msg
-- -----------------------------------------------------------------------------
function HoJ:Msg( text )
	if ( DEFAULT_CHAT_FRAME ) then
		DEFAULT_CHAT_FRAME:AddMessage( HoJ.MSG_PREFIX .. text )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:PrintTable
-- -----------------------------------------------------------------------------
function HoJ:PrintTable( t, name, depth )
    if ( t == nil ) then
        HoJ:DebugMsg( "PrintTable: table is nil!" )
        return
    end
    
    local prevIndentStr = string.rep( "  ", depth * 2 )
    HoJ:DebugMsg( prevIndentStr .. "[" .. name .. "] = {" )

    local indentStr = string.rep("  ", ( depth + 1 ) * 2)

    for key, value in pairs( t ) do
        local valueType = type( value )
        if ( valueType == "table" ) then
            HoJ:PrintTable( value, key, depth + 1 )
        elseif ( valueType == "string" ) then
            HoJ:DebugMsg( indentStr .. "[" .. key .. "] = '" .. value .. "'" )
        elseif ( valueType == "boolean" ) then
			HoJ:DebugMsg( indentStr .. "[" .. key .. "] = " .. tostring( value ) )
		elseif ( valueType == "function" ) then
            HoJ:DebugMsg( indentStr .. "[" .. key .. "] = " .. key )
        else
            HoJ:DebugMsg( indentStr .. "[" .. key .. "] = " .. value )
        end
    end

    HoJ:DebugMsg( prevIndentStr .. "}" )
end

--[[
function DeepCopyTable( sourceTable )
    local originalType = type( sourceTable )
    local destTable
    if originalType == 'table' then
        -- copy tables
        destTable = {}
        for originalKey, orginalValue in next, sourceTable, nil do
            destTable[DeepCopyTable( originalKey )] = DeepCopyTable( orginalValue )
        end
        setmetatable( destTable, DeepCopyTable( getmetatable( sourceTable ) ) )
    else
        -- copy everything else
        destTable = sourceTable
    end
    return destTable
end
]]

-- -----------------------------------------------------------------------------
-- HoJ:CopyTable
-- simple recursive copy that doesn't copy metatables
-- -----------------------------------------------------------------------------
function HoJ:CopyTable( source )
	local sourceType = type( source )
	local destTable
	if ( sourceType == "table" ) then
		destTable = { }
		for srcKey, srcValue in pairs( source ) do
			destTable[srcKey] = HoJ:CopyTable( srcValue )
		end
	else
		destTable = source
	end
	return destTable
end

-- -----------------------------------------------------------------------------
-- HoJ:MergeTable
-- simple recursive merge that doesn't copy metatables
-- sub-tables are merged recursively.
-- if a non-table key in dest matches a key in a source, then the dest key is
-- overwritten. 
-- -----------------------------------------------------------------------------
function HoJ:MergeTable( source, dest )
	local sourceType = type( source )
	local destType = type( dest )
	if ( sourceType ~= "table" or destType ~= "table" ) then
		HoJ:Error( "Attempted to merge non-tables! Merge failed!" )
		return false
	end

	for srcKey, srcValue in pairs( source ) do
		local srcValueType = type( srcValue )
		if ( srcValueType == "table" ) then
			if ( dest[srcKey] == nil ) then
				dest[srcKey] = { }
			end
			if ( not HoJ:MergeTable( srcValue, dest[srcKey] ) ) then
				return false
			end
		else
			dest[srcKey] = srcValue
		end
	end

	return true
end

-- -----------------------------------------------------------------------------
-- HoJ:InitClass
-- -----------------------------------------------------------------------------
function HoJ:InitClass( className )
	self._className = className
	self._classId = HoJ._classNameToClassId[className]
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateSpec
-- -----------------------------------------------------------------------------
function HoJ:UpdateSpec()

	self._classSpecInfo = HoJ[self._className]
	if ( self._classSpecInfo ~= nil ) then
		local specId = GetSpecialization()	
		if ( specId == nil ) then
			specId = self._classSpecInfo._defaultSpecId
			-- this should never really happen, but is probably survivable
			-- I thnk I've seen GetSpecialization return nil after entering world?
			HoJ:Error( "No specialization selected. Defaulting to " .. specName .. "." )
		end

		local classSpecIdTable = HoJ._specIdToSpecName[self._className]
		if ( classSpecIdTable ~= nil ) then
			-- we should always have a spec id table for a supported class
			local specName = classSpecIdTable[specId]
			if ( specName ~= nil ) then
				self._curSpecInfo = self._classSpecInfo[specName]
				self._role = self._curSpecInfo._role
				self._isActive = true
			
				return self._curSpecInfo
			else
				HoJ:Error( "classSpecIdTable has nil entry for " .. specName )
			end
		else
			HoJ:Error( "classSpecIdTable is nil for " .. self._className .. "!" )
		end
	end
	self._curSpecInfo = nil
	self._role = nil
	self._isActive = false
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellName
-- -----------------------------------------------------------------------------
function HoJ:GetSpellName( spellID )
	name = GetSpellInfo( spellID );
	if ( name == nil ) then
		HoJ:DebugMsg( "Got nil spell name for id " .. spellID )
	end
	return name;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellRank
-- -----------------------------------------------------------------------------
function HoJ:GetSpellRank( spellID )
	_, rank = GetSpellInfo( spellID )
	return rank;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIcon
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIcon( spellID )
	_, _, icon = GetSpellInfo( spellID )
	return icon;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetSpellIndex
-- this tries to find the index of a spell in the spellbook, which is not the
-- same as the spell ID of a spell. For some reason the spell index is needed
-- to find if a target is in range of a spell, rather than the spell ID.
-- -----------------------------------------------------------------------------
function HoJ:GetSpellIndex( spellInfo )
	-- tabs are the side tabs which are currently "General", "Retribution", "Holy", and "Protection"
	-- there is no way to know how many spells are on a single tab, but it won't be over 48 for paladins in 8.1
	local numSpellTabs = GetNumSpellTabs()
	--HoJ:DebugMsg( "Found " .. numSpellTabs .. " spellbook tabs" );
	local max = numSpellTabs * 96
	for i = 1, max do
		local spellName, spellSubName = GetSpellBookItemName( i, BOOKTYPE_SPELL )
		if ( spellName == nil ) then
			--HoJ:DebugMsg( "No spellbook info found for index " .. i );
			return -1;
	   	else
			-- use spellName and spellSubName here
			-- HoJ:DebugMsg( "Spell #" .. i .. ": '" .. spellName .. "'" )
			if ( spellInfo.name == spellName ) then
				--HoJ:DebugMsg( "Found spellbook index " .. i .. " for spell id " .. spellInfo.name );
				return i
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ_InitSpellInfo
-- -----------------------------------------------------------------------------
function HoJ_InitSpellInfo()
	-- HoJ:DebugMsg( "InitSpellInfo" );
	for key, value in pairs( HoJ.SpellInfo ) do
		-- HoJ:DebugMsg( value.name );
		local name, rank, icon = GetSpellInfo( value.id )
		-- some spell infos use an alternate icon id -- for instance a buff that allows another
		-- spell to be cast (Empyrean Power), we want to show the spell that can be cast on the
		-- rotation frame, not the buff.
		if ( value.iconId ~= nil ) then
			-- HoJ:DebugMsg( "Getting iconId " .. value.iconId .. " for " .. value.name );
			_, _, icon = GetSpellInfo( value.iconId )
		end
		-- For some reason, if we query Crusader Strike while prot, we get a weird texture id
		-- so when we switch roles we go ahead and re-init all of the spell infos.
		-- HoJ:DebugMsg( "  id = " .. value.id );
		-- HoJ:DebugMsg( "  old icon = " .. value.icon .. ", new icon = " .. icon );
		value.name = name
		value.rank = rank
		value.icon = icon
		value.spellIndex = HoJ:GetSpellIndex( value )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIcon
-- -----------------------------------------------------------------------------
function HoJ:AddIcon( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" )
	text = text or ""

	if ( list == nil ) then
		HoJ:Error( "HoJ:AddIcon - Icon list is NIL" )
	end

	if ( not spellInfo.showOnBar ) then
		HoJ:RemoveIcon( list, spellInfo )
		return
	end

	-- HoJ:FTLog( 15, spellInfo.id == 188370, "AddIcon( " .. spellInfo.name .. " )" )
	local si = spellInfo
	si.text = text
	-- table.insert( list, si )
	list[#list + 1] = si
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconExclusive
-- -----------------------------------------------------------------------------
function HoJ:AddIconExclusive( list, spellInfo, text )
	-- HoJ:DebugMsg( "HoJ:AddIcon( " .. spellInfo.name .. " )" )
	text = text or ""

	if ( list == nil ) then
		self:DebugMsg( "HoJ:AddIcon - Icon list is NIL" )
	end

	if ( not spellInfo.showOnBar ) then
		self:RemoveIcon( list, spellInfo )
		return
	end

	for i = 1, #list do
		if ( list[i].name == spellInfo.name ) then
			return
		end
	end

	-- self:DebugMsg( "AddIconExclusive( " .. spellInfo.name .. " )" )
	local si = spellInfo
	si.text = text
	-- table.insert( list, si )
	list[#list + 1] = si
end

-- -----------------------------------------------------------------------------
-- HoJ:AddIconDebug
-- just adds debug text where the cooldown text would be
function HoJ:AddIconDebug( list, spellInfo, debugText )
	HoJ:AddIcon( list, spellInfo, "" )
end

-- -----------------------------------------------------------------------------
-- HoJ:RemoveIcon
-- -----------------------------------------------------------------------------
function HoJ:RemoveIcon( list, spellInfo )
	if ( list == nil ) then
		HoJ:DebugMsg( "HoJ:RemoveIcon - Icon list is NIL" );
	end
	-- HoJ:DebugMsg( "RemoveIcon( " .. spellInfo.name .. " )" );
	for i, v in pairs( list ) do
		if ( v == spellInfo ) then
			table.remove( list, i );
			break;
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:ClearIconList
-- -----------------------------------------------------------------------------
function HoJ:ClearIconList( list )
	for i, v in pairs( list ) do
		list[ i ] = nil;
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SetIconFrame
-- -----------------------------------------------------------------------------
function HoJ:SetIconFrame( curTime, v, spellInfo )
	--local debug = spellInfo.id == 188370
	--self:Log( debug, "SetIconFrame for " .. spellInfo.name .. ", frame " .. v.frame:GetName() .. ", icon = " .. spellInfo.icon )
	if ( v == nil ) then
		HoJ:DebugMsg( "WTF" )
		return -- there are more skills than frames
	end

	v.frame:EnableMouse( true )
	--v.frame:Enable();
	v.frame:Show()
	-- HoJ:DebugMsg( "Enable mouse for " .. v.frame:GetName() );

	v.spellInfo = spellInfo

	v.texture:SetTexture( spellInfo.icon )
	v.texture:Show()

	local hideText = true
	if ( v.text ~= nil ) then
		if ( spellInfo.text ~= nil and spellInfo.text ~= "" ) then
			local fontFile = GameFontHighlightLarge:GetFont();
			v.text:SetFont( fontFile, 12, "THICKOUTLINE" );
			v.text:SetText( spellInfo.text );
			v.text:Show();
			if ( spellInfo.spellType == HoJ.SPELL_SKILL or spellInfo.spellType == HoJ.SPELL_TALENT ) then
				if ( v.texture ~= nil ) then
					v.texture:SetAlpha( HoJ._skillCDAlpha );
				end
			end
		end
	end
	if ( hideText ) then
		v.text:Hide();
		if ( v.texture ~= nil ) then
			v.texture:SetAlpha( HoJ._skillNormalAlpha );
		end
	end

	-- if the frame has a cooldown, set it
	if ( v._cooldownFrame ~= nil ) then
		local hideCooldown = true
		local start, duration, remaining
		start, duration = GetSpellCooldown( spellInfo.id )
		if ( start > 0 and duration > 0 ) then
			v._cooldownFrame:Show()
			v._cooldownFrame:SetCooldown( start, duration )
			hideCooldown = false
		end
		if ( hideCooldown ) then
			v._cooldownFrame:Hide()
		end
	end

	local buffTimeRemaining, buffStart, buffDuration, buffStacks
	local isBuff = spellInfo.spellType == HoJ.SPELL_BUFF or spellInfo.spellType == HoJ.SPELL_TALENT_BUFF
	if ( isBuff ) then
		buffTimeRemaining, buffStart, buffDuration, buffStacks = HoJ:GetBuffTimeRemaining( spellInfo.id )
	end

	if ( v._timeLeftFontString ~= nil ) then
		local hide = true
		if ( isBuff and buffTimeRemaining > 0 and buffDuration > 0 ) then
			local fontFile = GameFontHighlightLarge:GetFont();
			v._timeLeftFontString:SetFont( fontFile, 12, "THICKOUTLINE" );
			v._timeLeftFontString:SetText( string.format( "%.0f", buffTimeRemaining ) );
			v._timeLeftFontString:Show();
			hide = false
		end
		if ( hide ) then
			v._timeLeftFontString:Hide()
		end
	end

	if ( v._stacksFontString ~= nil ) then
		local hide = true
		if ( isBuff and spellInfo._showStacks and buffStacks > 0 ) then
			if ( buffStacks == nil ) then
				HoJ:Error( "buffStacks is nil for " .. spellInfo.name .. " with _showStacks = true!" )
			else
				local fontFile = GameFontHighlightLarge:GetFont();
				v._stacksFontString:SetFont( "Fonts\\ARIALN.ttf", 10, "OUTLINE" );
				v._stacksFontString:SetText( string.format( "%.0f", buffStacks ) );
				v._stacksFontString:Show();
				hide = false
			end
		end
		if ( hide ) then
			v._stacksFontString:Hide()
		end
	end

	if ( v._chargesFontString ~= nil ) then
		local hide = true
		local fontFile = GameFontHighlightSmallOutline:GetFont()
		v._chargesFontString:SetFont( fontFile, 12, "THICKOUTLINE" )
		local currentCharges, maxCharges = GetSpellCharges( spellInfo.id )
		if ( spellInfo.id == 202539 ) then
			HoJ:DebugMsg( "currentCharges = " .. currentCharges .. ", maxCharges = " .. maxCharges )
		end
		if ( currentCharges ~= nil and maxCharges > 1 ) then
			v._chargesFontString:SetText( tostring( currentCharges ) )
			v._chargesFontString:Show()
			hide = false
		end
		if ( hide ) then
			v._chargesFontString:Hide()
		end
	end
end

-- -----------------------------------------------------------------------------
-- SkillCompare
-- -----------------------------------------------------------------------------
function SkillCompare( tbla, tblb )
	local cdra = tbla.cooldownRemaining;
	local cdrb = tblb.cooldownRemaining;

	if ( cdra <= 0.00001 ) then
		if ( cdrb <= 0.00001 ) then
			-- NOTE: a nil warning here means the spellInfo needs to have _priorty added to it
			local hadError = false
			if ( tbla._priority == nil ) then
				HoJ:FrameThrottleError( HoJ.UPDATES_PER_SEC * 5, tbla.name .. " is missing _priority field!" )
				hadError = true
			end
			if ( tblb._priority == nil ) then
				HoJ:FrameThrottleError( HoJ.UPDATES_PER_SEC * 5, tblb.name .. " is missing _priority field!" )
				hadError = true
			end
			if ( hadError ) then
				return true
			end
			return ( tbla._priority > tblb._priority );
		else
			return true;
		end
	elseif ( cdrb <= 0.00001 ) then
		-- tbl[a] must be on cooldown or we'd have gone into the first if
		return false;
	else
		-- both are on cooldown, sort by cooldown remaining time
		return ( cdra < cdrb );
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateIcons
-- -----------------------------------------------------------------------------
local SortedKeyLists = {
	{ 1 },
	{ 1, 2 },
	{ 1, 2, 3 },
	{ 1, 2, 3, 4 },
	{ 1, 2, 3, 4, 5 }
}

function HoJ:UpdateIcons( list, frameList, sortList )
	-- clear all icon frames

	local setIconFrame = HoJ.SetIconFrame
	local curTime = HoJ._curTime
	local frameNum = 0

	local size = #list
	if ( size > 1 and sortList ) then
		local sortIconsStartMs = HoJ.ProfileBegin2( "SortIcons" )

		-- use a table lookup to initialize the sorted keys list to the correct size
		local sortedKeys = SortedKeyLists[size]
		table.sort( sortedKeys, function( a, b ) return SkillCompare( list[a], list[b] ) end )

		HoJ.ProfileEnd( "SortIcons", sortIconsStartMs )

		for i = 1, size do
			setIconFrame( self, curTime, frameList[i], list[ sortedKeys[i] ] )
			frameNum = i
		end
	else
		--local sortIconFramesStartMs = HoJ.ProfileBegin2( "SetIconFramesSorted" )

		-- using #list is slightly faster in this case than ipairs()
		for i = 1, size do
			setIconFrame( self, curTime, frameList[i], list[i] )
			frameNum = i
		end

		-- HoJ.ProfileEnd( "SetIconFramesSorted", sortIconFramesStartMs )
	end

	--local enableMouseStartMs = HoJ.ProfileBegin2( "EnableMouse" )
	for i = (frameNum + 1), #frameList do
		local frame = frameList[ i ].frame
		frame:EnableMouse( false )
		frame:Hide()
	end
	--HoJ.ProfileEnd( "EnableMouse", enableMouseStartMs )
end

-- -----------------------------------------------------------------------------
-- SkillCompare2
-- -----------------------------------------------------------------------------
function SkillCompare2( priorityList, aSpellInfo, bSpellInfo )
	local aPriority = priorityList[aSpellInfo.id];
	local bPriority = priorityList[bSpellInfo.id];

	-- NOTE: a nil warning here means the priority list is missing an entry for this a spell id
	local hadError = false
	if ( aPriority == nil ) then
		HoJ:Error(aSpellInfo.name .. " is missing from priorityList " .. priorityList._name )
		hadError = true
	end
	if ( bPriority == nil ) then
		HoJ:Error(bSpellInfo.name .. " is missing from priorityList "  .. priorityList._name )
		hadError = true
	end
	if ( hadError ) then
		return true
	end
	return ( aPriority > bPriority )
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateIcons2
-- -----------------------------------------------------------------------------
local SortedKeyLists2 = {
	{ 1 },
	{ 1, 2 },
	{ 1, 2, 3 },
	{ 1, 2, 3, 4 },
	{ 1, 2, 3, 4, 5 }
}

function HoJ:UpdateIcons2( list, priorityList, frameList, sortList )
	-- clear all icon frames
	local setIconFrame = HoJ.SetIconFrame
	local curTime = HoJ._curTime
	local frameNum = 0

	local size = #list
	if ( size > 1 and sortList ) then
		local sortIconsStartMs = HoJ.ProfileBegin2( "SortIcons" )

		-- use a table lookup to initialize the sorted keys list to the correct size
		local sortedKeys = SortedKeyLists2[size]
		table.sort( sortedKeys, function( a, b ) return SkillCompare2( priorityList, list[a], list[b] ) end )

		HoJ.ProfileEnd( "SortIcons", sortIconsStartMs )

		for i = 1, size do
			local v = list[sortedKeys[i]]
			setIconFrame( self, curTime, frameList[i], list[ sortedKeys[i] ] )
			frameNum = i
		end
	else
		--local sortIconFramesStartMs = HoJ.ProfileBegin2( "SetIconFramesSorted" )

		-- using #list is slightly faster in this case than ipairs()
		for i = 1, size do
			setIconFrame( self, curTime, frameList[i], list[i] )
			frameNum = i
		end

		-- HoJ.ProfileEnd( "SetIconFramesSorted", sortIconFramesStartMs )
	end

	--local enableMouseStartMs = HoJ.ProfileBegin2( "EnableMouse" )
	for i = (frameNum + 1), #frameList do
		local frame = frameList[ i ].frame
		frame:EnableMouse( false )
		frame:Hide()
	end
	--HoJ.ProfileEnd( "EnableMouse", enableMouseStartMs )
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitBuffForName
-- -----------------------------------------------------------------------------
function HoJ:UnitBuffForName( unit, buffName )
	local i = 1;
	local name, icon, count = UnitAura( unit, i );
	while name do
		if ( name == buffName ) then
			return name, icon, count;
		end
		i = i + 1;
		name, icon, count = UnitBuff( unit, i );
	end
	return name, icon, count;
end

-- -----------------------------------------------------------------------------
-- HoJ:LoopOverActiveBuffs
-- returns number of buffs found
-- -----------------------------------------------------------------------------
function HoJ:LoopOverActiveBuffs( unit, activeBuffs, filter )
	local curTime = self._curTime
	local i = 1
	local numBuffs = 0
	local name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
	while name do
		if ( spellId ~= nil ) then
			--local debug = spellId == 188370

			numBuffs = numBuffs + 1
			-- HoJ:FTLog( 15, debug, filter .. ": Buff " .. name .. " is active, spellId = " .. spellId )
			local activeBuff = activeBuffs[spellId]
			if ( activeBuff ~= nil ) then
				--HoJ:Log( debug, name .. ": duration = " .. tostring( duration ) .. ", expiration = " .. tostring( expirationTime ) )
				activeBuff.isActive = true

				if ( duration ~= nil and expirationTime ~= nil ) then
					if ( duration == 0.0 and expirationTime == 0.0 ) then
						--HoJ:Log( debug, "Check for active AoE buff " .. name )
						local aoeBuff = self._curSpecInfo._aoeBuffTable[spellId]
						if ( aoeBuff ~= nil and aoeBuff._expiration > curTime ) then
							activeBuff.start = aoeBuff._start
							activeBuff.duration = aoeBuff._duration
							activeBuff.timeRemaining = aoeBuff._expiration - curTime
							--HoJ:Log( debug, "Found active AoE buff " .. name )
						end
					else
						-- HoJ:FTLog( 15, debug, "duration = " .. tostring( duration ) .. ", expiration = " .. tostring( expirationTime ) )
						activeBuff.start = expirationTime - duration
						activeBuff.duration = duration
						activeBuff.timeRemaining = expirationTime - curTime
						if ( activeBuff.timeRemaining < 0.0 ) then
							activeBuff.timeRemaining = 0.0
						end
					end
				end

				if ( count ~= nil ) then
					activeBuff.stacks = count
				end
			else
				-- HoJ:FTLog( 15, debug, filter .. ": Buff " .. name .. " (id " .. spellId .. ") not in activeBuffs list" )
			end
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
	end	
	return numBuffs
end

-- -----------------------------------------------------------------------------
-- HoJ:FindActiveBuffs
-- table of flags for each buff, indexed by spell id
--	{
--		isActive,		-- whether or not the buff is active
--		start,			-- time buff started
--		duration,		-- duration of the buff
--		timeRemaining, 	-- how long the buff has before expiration
--		stacks,			-- number of stacks of the buff
--		spellId,		-- spell Id of the buff
--	}
-- -----------------------------------------------------------------------------
function HoJ:FindActiveBuffs( unit, activeBuffs )
    -- set all active skills active flag to false
    if ( activeBuffs == nil ) then
        return 
    end

	for key, buff in pairs( activeBuffs ) do
		buff.isActive = false
		buff.start = 0.0
		buff.duration = 0.0
		buff.timeRemaining = 0.0
		buff.stacks = 0
	end

	-- we can't get back helpful and harmful buffs at the same time so we must loop twice
	HoJ:LoopOverActiveBuffs( unit, activeBuffs, "HELPFUL" )
	HoJ._hasDebuff = HoJ:LoopOverActiveBuffs( unit, activeBuffs, "HARMFUL" ) ~= 0
	--HoJ:DebugMsg( "_hasDebuff = " .. tostring( HoJ._hasDebuff ) )
end

-- -----------------------------------------------------------------------------
-- HoJ:IsBuffActive
-- -----------------------------------------------------------------------------
function HoJ:IsBuffActive( activeBuffList, spellId, predictAhead )
	if ( activeBuffList == nil ) then
        return false
	end

	--local debug = spellId == 188370

	local buff = activeBuffList[spellId]
	if ( buff == nil ) then
		--HoJ:FTLog( 15, debug, "Buff == nil" )
		return false
	end

	-- AoE buffs like consecration have no duration set
	if ( buff.duration > 0.0 ) then
		if ( predictAhead ~= nil ) then
			local isStillActive = buff.timeRemaining - predictAhead > 0.0
			--HoJ:FTLog( 15, debug, string.format( "Buff: TR = %.2f, PA = %.2f", buff.timeRemaining, predictAhead ) ..
			--		", isStillActive = " .. tostring( isStillActive ) )
			return isStillActive
		end
	end

	--HoJ:FTLog( 15, debug, "Buff " .. spellId .. " isActive = " .. tostring( buff.isActive ) )

	return buff.isActive
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffStackCount
-- -----------------------------------------------------------------------------
function HoJ:GetBuffStackCount( activeBuffList, spellId )
	if ( activeBuffList == nil ) then
		return 0
	end

	local buff = activeBuffList[spellId]
	if ( buff == nil ) then
		return 0
	end
	return buff.stacks;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuff
-- -----------------------------------------------------------------------------
function HoJ:GetBuff( activeBuffList, spellId )
	return activeBuffList[spellId]
end

-- -----------------------------------------------------------------------------
-- HoJ:GetUnitBuffTimeRemaining
-- returns:
--	float 	buffTimeRemaining
--	int 	stacks
-- -----------------------------------------------------------------------------
function HoJ:GetUnitBuffTimeRemaining( unit, inSpellId, predictAhead, filter )
	for i = 1, 40 do
		local name, _, count, _, duration, expirationTime, _, _, _, spellId = UnitAura( unit, i, filter )
		if ( name == nil ) then
			return 0.0, 0
		end
		-- self:DebugMsg( "Index " .. i .. " " .. unit .. " buff " .. name .. " with id = " .. spellId )
		if ( spellId == inSpellId ) then
			--self:DebugMsg( "Index " .. i .. " " .. unit .. " buff " .. name .. " with id = " .. spellId )
			return expirationTime - ( HoJ._curTime + predictAhead ), count
		end
	end
	return 0.0, 0
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingWithPrediction
-- gets the cooldown remaining, including GCD if GCD is in progress
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingWithPrediction( spellId, predictAhead )
	local start, duration, enabled = GetSpellCooldown( spellId );

	-- we don't really care about enabled... that just tells us if the spell / buff is active
	if ( start > 0 and duration > 0 ) then
		-- we "look into the future" to just past the time when the current GCD is done,
		-- to see what will be usable then
		local remaining = ( start + duration ) - ( HoJ._curTime + predictAhead );
		if ( remaining < 0.0 ) then
			remaining = 0.0;
		end
		return remaining;
	end
	return 0.0;
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingAfterGCD
-- For predicting what use next, we don't really care about the GCD period, we
-- just want to know if a spell will be usable once the GCD ends.
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingAfterGCD( spellId )
	-- we "look into the future" to just past the time when the current GCD is done,
	-- to see what will be usable then
	return HoJ:GetCDRemainingWithPrediction( spellId, HoJ._gcdRemaining );
end

-- -----------------------------------------------------------------------------
-- HoJ:GetCDRemainingIgnoringGCD
-- normally when we check for a spell cooldown, all spells return the GCD cooldown,
-- i.e., even if the spell itself is not on cooldown, if the GCD is, then it will
-- appear that the spell is on cooldown.
-- -----------------------------------------------------------------------------
function HoJ:GetCDRemainingIgnoringGCD( spellId )
	local start, duration, enabled = GetSpellCooldown( spellId );

	if ( start <= 0 or duration <= 0 ) then
		return 0.0
	end

	local endTime = start + duration
	local remaining = endTime - HoJ._curTime

	if ( self._gcdRemaining ) then
		-- we're in GCD...
		if ( self._gcdRemaining >= remaining ) then
			-- this is just a GCD cooldown, not the actual spell cooldown
			return 0.0
		end
	end
	return remaining
end

-- -----------------------------------------------------------------------------
-- HoJ:GetBuffTimeRemaining
-- returns:
--   time remaining in seconds
--   start time
--   duration in seconds
-- -----------------------------------------------------------------------------
function HoJ:GetBuffTimeRemaining( spellId, predictAhead )
	local buff = self._curSpecInfo._activeBuffList[spellId]
	if ( buff == nil ) then
		return 0.0, 0.0, 0.0
	end

	if ( predictAhead ~= nil ) then
		return buff.timeRemaining - predictAhead, buff.start, buff.duration, buff.stacks
	end
	return buff.timeRemaining, buff.start, buff.duration, buff.stacks
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckProcdBuffs
-- -----------------------------------------------------------------------------
function HoJ:CheckProcdBuffs( activeBuffList, procdBuffList )
	if ( not HoJ._showBuffsFrame ) then
		return;
	end

	if ( activeBuffList == nil ) then
		HoJ:Error( "activeBuffList is nil!" )
	end
	if ( procdBuffList == nil ) then
		HoJ:Error( "procdBuffList is nil!" )
	end

    for i, v in ipairs( procdBuffList ) do
		if ( v._spellInfo ~= nil ) then
			--local debug = v._spellInfo.id == 188370
			--HoJ:Log( debug, "Check procd buff '" .. v._spellInfo.name .. "'" )
			if ( self:IsBuffActive( activeBuffList, v._spellInfo.id ) ) then
				--HoJ:Log( debug, "Found procd buff '" .. v._spellInfo.name .. "'" )
				self:AddIcon( self._buffs, v._spellInfo )
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:SpellInRange_SpellIndex
-- Wraps IsSpellInRange with error checking
--------------------------------------------------------------------------------
function HoJ:SpellInRange_SpellIndex( spellIndex, name, unit, debugName )
	if ( spellIndex < 0 or spellIndex == nil ) then
		HoJ:DebugMsg( "Invalid spellIndex " .. spellIndex .. " for " .. debugName )
		return false
	end
	return IsSpellInRange( spellIndex, name, unit )
end

-- -----------------------------------------------------------------------------
-- HoJ:SpellInRange_SpellName
-- Wraps IsSpellInRange with error checking
--------------------------------------------------------------------------------
function HoJ:SpellInRange_SpellName( spellName, unit, debugName )
	if ( spellName == nil ) then
		HoJ:DebugMsg( "Invalid spell name for " .. debugName )
		return false
	end
	return IsSpellInRange( spellName, unit )
end

-- -----------------------------------------------------------------------------
-- HoJ:InRange
-- -----------------------------------------------------------------------------
function HoJ:InRange( spellInfo, unit, debugName )
	local inRange = false;
--[[
	if ( spellInfo.spellType == HoJ.SPELL_TALENT ) then
		inRange = HoJ:SpellInRange_SpellIndex( spellInfo.spellIndex, "spell", unit, debugName );
	else
		inRange = HoJ:SpellInRange_SpellName( spellInfo.name, unit, debugName );
	end
--]]
	inRange = HoJ:SpellInRange_SpellName( spellInfo.name, unit, debugName );
	
	if ( inRange == nil ) then
		--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = nil" );
		return false;
	elseif ( inRange == 0 ) then
		--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = 0" );
		return false;
	end
	--HoJ:DebugMsg( "InRange( " .. spellInfo.name .. " ) = true" );
	return true;
end

-- -----------------------------------------------------------------------------
-- HoJ:WillOverflowHoPo
-- CLASSFIX - HoPo
-- -----------------------------------------------------------------------------
function HoJ:WillOverflowHoPo( spellInfo, curHoPo, maxHoPo )
	local overflow = ( curHoPo + spellInfo._powerGen ) - maxHoPo;
	-- HoJ:DebugMsg( spellInfo.name .. ": overflow = " .. overflow );
	return overflow;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsOffCooldown
-- -----------------------------------------------------------------------------
function HoJ:IsOffCooldown( spellInfo, predictAhead )
	local cdRemaining = HoJ:GetCDRemainingAfterGCD( spellInfo.id );
	-- HoJ:DebugMsg( "spell " .. spellInfo.name .. ": cdRemaining = " .. cdRemaining .. ", GCD Remaining = " .. HoJ._gcdRemaining );
	if ( cdRemaining == nil ) then
		HoJ:DebugMsg( "cdRemaining == nil!" )
		return false;
	elseif ( predictAhead == nil ) then
		return cdRemaining <= 0.0
	end
	if ( cdRemaining <= predictAhead ) then
		return true;
	end
	return false;
end

-- -----------------------------------------------------------------------------
-- HoJ:IsSkillUsable
-- -----------------------------------------------------------------------------
function HoJ:IsSkillUsable( spellInfo, predictAhead )
	if ( not HoJ._targetInRange[spellInfo.range] ) then
		-- HoJ:DebugMsg( spellInfo.name .. " is not in range" )
		return false;
	end
	if ( not HoJ:IsOffCooldown( spellInfo, predictAhead ) ) then
		return false;
	end
	return true;
end

-- -----------------------------------------------------------------------------
-- ArrayContains
-- -----------------------------------------------------------------------------
function ArrayContains( t, value )
    for i, curValue in ipairs( t ) do
        if ( value == curValue ) then
            return true
        end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:IsTalentSelected
-- -----------------------------------------------------------------------------
function HoJ:IsTalentSelected( spellInfo )
    if ( spellInfo._talentSpecs == nil ) then
        return false
    end

    if ( not ArrayContains( spellInfo._talentSpecs, self._curSpecInfo._spec ) ) then
		if ( HoJ._debugSpellName ~= nil and spellInfo.name == HoJ._debugSpellName ) then
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " does not match spec " .. HoJ._curSpecInfo._spec .. "!" )
		end
		return false
	end

	if ( spellInfo.talentRow == nil or spellInfo.talentCol == nil ) then
		if ( spellInfo.spellType ~= HoJ.SPELL_TALENT ) then
			HoJ:DebugMsg( "IsTalentSelected() called on non-talent " .. spellInfo.name );
		else
			HoJ:DebugMsg( "Talent " .. spellInfo.name .. " missing row and column" );
		end
		return true;
	end

	local _, _, _, selected, available = GetTalentInfo( spellInfo.talentRow, spellInfo.talentCol, 1 );
	if ( HoJ._debugSpellName ~= nil and spellInfo.name == HoJ._debugSpellName ) then
		HoJ:DebugMsg( "Talent " .. spellInfo.name .. ", available = " .. tostring( available ) .. ", selected = " .. tostring( selected ) )
	end
	return ( available and selected );
end

-- -----------------------------------------------------------------------------
-- HoJ:GetPower
-- -----------------------------------------------------------------------------
function HoJ:GetPower( predictAhead )
	if ( predictAhead == nil ) then
		return HoJ._power
	end

	return HoJ._power + ( predictAhead * 1.0 * HoJ._powerPerSec )
end

-- -----------------------------------------------------------------------------
-- HoJ:SelectMitigationSkill
-- expects an array of table items of format:
--	{
--		spellInfo,			-- info about the spell
--		replacedByTalent,	-- optional, can be nil, otherwise specifies the spellInfo of the talent that replaces this spell
--      physicalOnly,		-- optional, can be nil
--		excludedByBuffId,	-- optional, if not nil, then this buff prevents the skill from working (Forbearance)
--	}
-- NOTE: the list MUST be in priority order because the loop short-circuits once a
-- mitigation is found, so the order IS the priority
-- -----------------------------------------------------------------------------
function HoJ:SelectMitigationSkill( mitigationSkills, listName )
	if ( listName ~= nil ) then
		self:DebugMsg( HoJ.STR_GREEN .. "Checking mitigations: " .. listName )
	end

	for i, mitigation in ipairs( mitigationSkills ) do
		-- self:DebugMsg( "Checking mitigation skill " .. mitigation.spellInfo.name )
		if ( mitigation.spellInfo.requiredRaceId == nil or mitigation.spellInfo.requiredRaceId == self._playerRaceId ) then
			if ( not self._protActiveBuffList[mitigation.spellInfo.id].isActive ) then
				if ( ( mitigation.replacedByTalent == nil or not self:IsTalentSelected( mitigation.replacedByTalent ) )
					and ( mitigation.excludedByBuffId == nil or not self:IsBuffActive( self._curSpecInfo._activeBuffList, mitigation.excludedByBuffId ) ) ) then
					local isSelected = mitigation.spellInfo.type ~= HoJ.SPELL_TALENT or self:IsTalentSelected( mitigation.spellInfo )
					if ( not C_SpellBook.IsSpellDisabled( mitigation.spellInfo.id ) ) then
						if ( ( mitigation.physicalOnly ~= nil and HoJ._isTakingPhysicalDamage ) or mitigation.physicalOnly == nil ) then
							if ( isSelected ) then
								if ( self:IsOffCooldown( mitigation.spellInfo, 0.0 ) ) then
									-- self:DebugMsg( HoJ.STR_GREEN .. "Selected mitigation: " .. mitigation.spellInfo.name )
									return mitigation
								end
							end
						end
					end
				end
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:GetGroupType
-- -----------------------------------------------------------------------------
function HoJ:GetGroupType()
	if ( IsInRaid() ) then
		return HoJ.GROUP_RAID
	elseif ( IsInGroup() ) then
		return HoJ.GROUP_PARTY
	end
	return HoJ.GROUP_NONE
end

-- -----------------------------------------------------------------------------
-- HoJ:MaxMemberForGroupType
-- The player doesn't count as a partyN unit id, so if there are 3 party members,
-- only "party1" and "party2" are valid unit ids.
-- If the player is in a raid, however, this is not true. If there are 3 raid
-- members, then "raid1", "raid2", and "raid3" are all valid and one of them will
-- represent the "player".
-- To account for this, we subtract one from the number of group members when
-- the player is in a party, but not when in a raid.
-- -----------------------------------------------------------------------------
function HoJ:MaxMemberForGroupType( groupType )
	if ( groupType == HoJ.GROUP_PARTY ) then
		return GetNumGroupMembers() - 1
	end
	return GetNumGroupMembers()
end

-- -----------------------------------------------------------------------------
-- HoJ:PrefixForGroupType
-- -----------------------------------------------------------------------------
function HoJ:PrefixForGroupType( groupType )
	if ( IsInRaid() ) then
		return "raid"
	elseif ( IsInGroup() ) then
		return "party"
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:UnitIDForName
-- -----------------------------------------------------------------------------
function HoJ:UnitIDForName( name )
	if ( name == HoJ._playerName ) then
		return "player"
	end

	local prefix = HoJ:PrefixForGroupType( HoJ._groupType )
	if ( prefix == nil ) then
		return nil
	end

	local max = HoJ:MaxMemberForGroupType( HoJ:GetGroupType() )
	for i = 1, max do
		local unitID = prefix .. i
		local memberName = UnitName( unitID )
		if ( memberName == name ) then
			return unitID
		end
	end
	return nil
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetBlessing
-- -----------------------------------------------------------------------------
function HoJ:ResetBlessing( blessing )
	blessing.targetName = nil
	blessing.start = 0.0
	blessing.duration = 0.0
	blessing.updateIndex = -1
end

-- -----------------------------------------------------------------------------
-- HoJ:ResetBlessings
-- -----------------------------------------------------------------------------
function HoJ:ResetBlessings( blessings )
    if ( blessings == nil ) then
        return 
    end
	for spellId, blessing in pairs( blessings) do
		--HoJ:DebugMsg( "Resetting " .. blessing.spellInfo.name .. "..." )
		HoJ:ResetBlessing( blessing )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateBlessingFromUnitID
-- returns true if the unitID matches the name, in which case we can definitively
-- know that the unit is the one that was cast on.
-- -----------------------------------------------------------------------------
function HoJ:UpdateBlessingFromUnitID( blessing, unitID )
	--HoJ:DebugMsg( "Checking for " .. blessing.spellInfo.name .. " on " .. unitID )
	if ( unitID == nil ) then
		--HoJ:FrameThrottleMsg( 75, "Error: unitID == nil!" )
		return false
	end

	local unitName = UnitName( unitID )
	if ( unitName == nil ) then
		--HoJ:FrameThrottleMsg( 75, "Error: UnitName() == nil for unit id '" .. unitID .. "'!" )
		return false
	end

	if ( blessing.targetName ~= unitName ) then
		return false
	end

	--HoJ:DebugMsg( "Blessing target is " .. unitID .. "( " .. unitName .. " )" )

	local found = false
	local i = 1
	local name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	while name do
		if ( spellId == blessing.spellInfo.id and unitCaster == "player" ) then
			-- this is the spell we cast
			--HoJ:DebugMsg( "Updating " .. blessing.spellInfo.name .. " from " .. unitName )
			
			-- assume buffs are expired if < 1 second, because there is a bug where Wisdom sometimes sticks at
			-- 0 seconds remaining and causes the active buff to flash on. Even if we do this here, we also need
			-- to be sure that we won't alsojust re-add the blessing in FindBlessingOnUnitID because it wasn't
			-- found on a known target.
			local timeRemaining = expirationTime - self._curTime
			if ( timeRemaining > 0.0 ) then
				found = true
			end
			if ( HoJ._debugBlessings ) then
				HoJ:DebugMsg( "UpdateBlessing: " .. blessing.spellInfo.name .. " duration = " .. duration .. ", expire = " .. expirationTime .. ", remain = " .. timeRemaining )
			end

			local start = expirationTime - duration
			blessing.start = start
			blessing.duration = duration
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	end
	-- not found on the unit it was cast on, so it's expired or been canceled
	if ( not found ) then
		--HoJ:DebugMsg( blessing.spellInfo.name .. " was not found. Resetting." )
		HoJ:ResetBlessing( blessing )
	end

	return true
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessingOnUnitID
-- -----------------------------------------------------------------------------
function HoJ:FindBlessingOnUnitID( blessing, unitID )
	--HoJ:DebugMsg( "Checking for " .. blessing.spellInfo.name .. " on " .. unitID )
	local i = 1
	local name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	while name do
		if ( spellId == blessing.spellInfo.id and unitCaster == "player" ) then
			-- this is the spell we cast
			--HoJ:DebugMsg( "Found " .. blessing.spellInfo.name .. " on " .. unitID )
			local timeRemaining = expirationTime - self._curTime
			if ( HoJ._debugBlessings ) then
				HoJ:DebugMsg( "FindBlessing: " .. blessing.spellInfo.name .. " duration = " .. duration .. ", expire = " .. expirationTime .. ", remain = " .. timeRemaining )
			end			
			if ( timeRemaining <= 0.0 ) then
				-- blessing sometimes remain on the unit beyond their expiration time
				return false
			end
			local start = expirationTime - duration
			blessing.start = start
			blessing.duration = duration
			return true
		end
		i = i + 1;
		name, _, count, _, duration, expirationTime, unitCaster, _, _, spellId = UnitAura( unitID, i, "HELPFUL" )
	end
	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessing
-- -----------------------------------------------------------------------------
function HoJ:UnitIDForUpdateIndex( index, groupType )
	if ( index < 0 ) then
		return "player"
	elseif ( index == 0 ) then
		return "target"
	end

	local max = HoJ:MaxMemberForGroupType( groupType )
	local prefix = HoJ:PrefixForGroupType( groupType )
	if ( prefix == nil ) then
		return nil
	end

	local i = index
	if ( i > max ) then
		i = max
	end

	local unitID = prefix .. i
	return unitID
end

-- -----------------------------------------------------------------------------
-- HoJ:FindBlessing
-- -----------------------------------------------------------------------------
function HoJ:FindBlessing( blessing )
	--HoJ:DebugMsg( "FindBlessing " .. blessing.spellInfo.name )

	local unitID = HoJ:UnitIDForUpdateIndex( blessing.updateIndex, self._groupType, self._numGroupMembers )
	if ( unitID == nil ) then
		return
	end

	if ( HoJ:FindBlessingOnUnitID( blessing, unitID ) ) then
		-- store the name so we can use the direct lookup
		blessing.targetName = UnitName( unitID )
		return
	end

	-- start over from the beginning
	blessing.updateIndex = blessing.updateIndex + 1
	if ( blessing.updateIndex > self._numGroupMembers ) then
		blessing.updateIndex = -1
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:UpdateBlessings
-- For our purposes here a blessing is a buff that can only be on a single target
-- at a time and must be tracked for that target.

-- We can't rely on a blessing being set on a target immediately after cast due
-- to network lag.
-- We can't rely on a blessing being set on the player immediately after they
-- enter the worl due to network lag.
-- We also cannot rely on being able to check the buff on the a player the buff
-- has been cast on if that player is not in the player's party.
-- For this reason, if we don't find a blessing on a target we have to treat it
-- differently based on whether or not the buff target is in the player's party.

-- To deal with all of the above, when a blessing is cast, we record the name
-- of the target it was cast on (not the unitID it was cast on).

-- Each time we check for blessings, if the blessing has a recorded targetName,
-- we cycle through unit IDs ("player",
-- "target" or and "partyN"/"raidN" for our current group) to try to find the
-- one that matches the recorded name.

-- If we don't find the a unit ID with a matching name, we just check for
-- buff expiration using the stored state.

-- If we do find a unit ID with a matching name, we search the unit's auras for
-- the blessing. If we don't find it, we know it's expired or canceled and we
-- can confidently show the buff icon to indicate it has expired.
--
-- If the buff has no recorded name, then we search through auras on the "player",
-- "target", and "partyN" or "raidN" unitIDs to find a matching buff cast by
-- the player. If we find it, we then store the name of that unit. This handles
-- initializing the buff tracking at any time after the player enters the world.
-- If any player buff is on another player that can be inspected via a unit ID
-- then it will start being tracked. This search through auras on all unit IDs
-- could be very expensive to do all at once for a large raid, so this test is
-- amortized and will check one unit ID per update frame (currently @ 15 fps).
-- This means it could take up to 40 / 15 seconds to find a buff after logging
-- in.

-- table of flags for each buf indexed by spell id
--	{
--		targetName,	-- name of the target the spell was cast on
--		start,		-- time buff started
--		duration,	-- duration of the buff
--		updateIndex	-- index of the current party member to check auras for
--	}
-- -----------------------------------------------------------------------------

-- -----------------------------------------------------------------------------
function HoJ:UpdateBlessings( blessings )
	-- HoJ:DebugMsg( "UpdateBlessings: " .. tostring( blessings ) )
	if ( blessings == nil ) then
		return
	end

	local groupType = HoJ:GetGroupType()
	local numGroupMembers = GetNumGroupMembers()
	if ( groupType ~= self._groupType ) then
		-- group changed so restart the buff checks
		--HoJ:DebugMsg( "Resetting update due to group change!" )
		for spellId, blessing in pairs( blessings ) do
			blessing.updateIndex = -1
		end
	elseif ( numGroupMembers ~= self._numGroupMembers ) then
		-- number of group members changed, so restart the buff checks if we're beyond the number of group members
		-- HoJ:DebugMsg( "Resetting update due to group size change!" )
		local max = HoJ:MaxMemberForGroupType( self._groupType )
		for spellId, blessing in pairs( blessings ) do
			if ( blessing.updateIndex >= max ) then
				blessing.updateIndex = -1
			end
		end
	end

	self._groupType = groupType
	self._numGroupMembers = numGroupMembers

	for spellId, blessing in pairs( blessings ) do
		if ( blessing.targetName ~= nil ) then
			--HoJ:DebugMsg( "Blessing " .. blessing.spellInfo.name .. " targets " .. blessing.targetName )
			if ( not HoJ:UpdateBlessingFromUnitID( blessing, "player" ) ) then
				if ( not HoJ:UpdateBlessingFromUnitID( blessing, "target" ) ) then
					if ( HoJ._groupType ~= HoJ.GROUP_NONE ) then

						local unitID = HoJ:UnitIDForName( blessing.targetName )
						if ( unitID ~= nil ) then
							if ( HoJ:UpdateBlessingFromUnitID( blessing, unitID ) ) then
								found = true
								break
							end
						end
					end
				end
			end
			-- check for expiration
			local expireTime = blessing.start + blessing.duration
			local timeRemaining = expireTime - self._curTime
			if ( timeRemaining <= 0.0 ) then
				HoJ:ResetBlessing( blessing )
			end
		else
			HoJ:FindBlessing( blessing )
		end
	end
	HoJ._nextUpdateBlessingFrame = self._updateFrameNum + self._updateBlessingsInterval
end

-- -----------------------------------------------------------------------------
-- Data-driven Skill Selection
-- -----------------------------------------------------------------------------
HoJ.OP_NONE = -999

-- boolean comparison operators
HoJ.OP_LESS_EQUAL = -2
HoJ.OP_LESS = -1
HoJ.OP_EQUAL = 0
HoJ.OP_GREATER = 1
HoJ.OP_GREATER_EQUAL = 2
HoJ.OP_NOT_EQUAL = 3

-- logic operators
HoJ.OP_AND = 1000
HoJ.OP_OR = 1001

-- range operator
HoJ.OP_IN_RANGE_INCLUSIVE = 1002 -- requires a const array { min, max }, true if >= min and <= max, i.e. [ min, max ]
HoJ.OP_IN_RANGE_MIN_EX_MAX_IN = 1003 -- requires a const array { min, max }, true if > min and <= max, i.e. ( min, max ]
HoJ.OP_IN_RANGE_MIN_IN_MAX_EX = 1004 -- requires a const array { min, max }, true if >= min and < max, i.e. [ min, max )
HoJ.OP_IN_RANGE_MIN_EX_MAX_EX = 1005 -- requires a const array { min, max }, true if > min and < max, i.e. ( min, max )

-- TODO: need a way to group operations

-- constant types
HoJ.TYPE_CONSTANT = 0 	-- use the constant specified by aValue or bValue

-- result types
HoJ.TYPE_RESULT = 1     -- variale is the result of a previous evaluation

-- variable types
HoJ.TYPE_POWER 					        = 1000	-- get player's current power level (holy, rage, etc.)
HoJ.TYPE_PLAYER_HEALTH_RATIO			= 1001	-- get the target's health ratio
HoJ.TYPE_TARGET_HEALTH_RATIO			= 1002	-- get the target's health ratio
HoJ.TYPE_PLAYER_BUFF_ACTIVE				= 1003	-- (constant = spellId) true if the buff specified by the const parm is active
HoJ.TYPE_TARGET_RANGE					= 1004	-- get the target's range with const as the range type
HoJ.TYPE_TALENT_SELECTED				= 1005	-- true if the talent specified by the const parm (spellInfo) is selected
HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING	= 1006	-- time remaining for a debuff applied by the player
HoJ.TYPE_PLAYER_BUFF_REMAINING			= 1007	-- time remaining for a buff on the player
HoJ.TYPE_SKILL_CHARGES                  = 1008	-- number of current charges a skill has (for instance, crusader strike has up to 2 charges)
HoJ.TYPE_NUM_MELEE_ATTACKERS			= 1009	-- number of enemies attacking the player with attacks <= 8 yards
HoJ.TYPE_PLAYER_BUFF_STACK_COUNT		= 1010	-- (constant = spellId) number of stacks a buff (on player only) has
HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO		= 1011	-- minor heals will show in rotation if health ratio < this
HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO		= 1012	-- major heals will show in rotation if health ratio < this
HoJ.TYPE_TARGET_IS_BOSS					= 1014	-- maps to HoJ._targetIsBoss
HoJ.TYPE_SKILL_OFF_COOLDOWN				= 1015	-- returns true if the skill is off cooldown
HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED		= 1016	-- returns true if the target can be interrupted
HoJ.TYPE_COMBO_POINTS					= 1017	-- current combo points
HoJ.TYPE_PLAYER_IS_STEALTHED			= 1018	-- true if player is stealthed
HoJ.TYPE_INCOMING_DPS_RATIO				= 1019	-- returns damage per second ratio
HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL	= 1020	-- seconds since the player last took spell damage
HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE		= 1021	-- seconds until target is predicted to die, based on combat session dps
HoJ.TYPE_SPELL_DURATION					= 1022	-- duration a skill or spell's effect lasts, in seconds
HoJ.TYPE_TARGET_IS_TOUGH				= 1023	-- returns HoJ._targetIsTough
HoJ.TYPE_COMBAT_ATTACKS_HIT				= 1024	-- returns number of attacks that have hit during current combat session
HoJ.TYPE_SKILL_POWER_COST				= 1025	-- cost of the specified skill
HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF		= 1026	-- returns true if the target has the specified debuff
HoJ.TYPE_SKILL_COOLDOWN_REMAINING		= 1027	-- returns time remaining on the spell's cooldown, not including GCD
HoJ.TYPE_UNIT_IS_SLOWED					= 1028	-- returns true if the unit is under some slowing effect, determined by checking speed
HoJ.TYPE_NET_INCOMING_DPS				= 1029	-- returns net incoming damage per second
HoJ.TYPE_NET_INCOMING_DPS_RATIO			= 1030	-- returns net incoming damage per second ratio
HoJ.TYPE_NET_INCOMING_DPS_PCT			= 1031	-- returns net incoming damage per second ratio as a percentage
HoJ.TYPE_PLAYER_TIME_TO_LIVE			= 1032	-- returns predicted time before player will die
HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH		= 1033	-- returns predicted time before player will be at 10% health
HoJ.TYPE_NET_INCOMING_DPS_CHR			= 1034	-- returns ( net incoming dps / current health ) ratio
HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL	= 1035	-- returns true if a known major damage spell is being cast by the target

local _varTable = {
	[HoJ.TYPE_RESULT] = function( constant, results )
		if ( results == nil ) then
			HoJ:Error( "Type is a result but results list == nil!" )
			return 0
		elseif ( constant == nil ) then
			HoJ:Error( "Type is a result but constant == nil! constant  must specify a result index!" )
			return 0
		else
			return results[constant]
		end
	end,
	[HoJ.TYPE_POWER] = function( constant, results )
		if ( HoJ._curSpecInfo._autoPowerGen ) then
			-- TODO: this is a hack -- it would be better to remove the power check from skill rules
			-- report as if we have max power. This makes the rotation simply choose the next
			-- best skill and wait on the energy to pool
			return 100
		else
			return HoJ:GetPower( HoJ._lookAheadTime ) 
		end
	end,
	[HoJ.TYPE_PLAYER_HEALTH_RATIO] = function( constant, results ) 
		return HoJ._playerHealthRatio 
	end,
	[HoJ.TYPE_TARGET_HEALTH_RATIO] = function( constant, results ) 
		return HoJ._targetHealthRatio 
	end,
	[HoJ.TYPE_PLAYER_BUFF_ACTIVE] = function( constant, results ) 
		return HoJ:IsBuffActive( HoJ._curSpecInfo._activeBuffList, constant, HoJ._lookAheadTime )
	end,
	[HoJ.TYPE_TARGET_RANGE] = function( constant, results ) 
		return HoJ._targetInRange[constant] 
	end,
	[HoJ.TYPE_TALENT_SELECTED] = function( constant, results ) 
		return HoJ:IsTalentSelected( constant ) 
	end,
	[HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING] = function( constant, results ) 
		return HoJ:GetUnitBuffTimeRemaining( "target", constant, HoJ._lookAheadTime, "PLAYER|HARMFUL" )
	end,
	[HoJ.TYPE_PLAYER_BUFF_REMAINING] = function( constant, results ) 
		return HoJ:GetBuffTimeRemaining( constant, HoJ._lookAheadTime )
	end,
	[HoJ.TYPE_SKILL_CHARGES] = function( constant, results ) 
		local numCharges = GetSpellCharges( constant )
		if ( numCharges == nil ) then
			HoJ:Error( "GetSpellCharges() == nil! Check spell ID in rule!" )
			return 0
		end
		return numCharges 
	end,
	[HoJ.TYPE_NUM_MELEE_ATTACKERS] = function( constant, results ) 
		return HoJ._numMeleeAttackers 
	end,
	[HoJ.TYPE_PLAYER_BUFF_STACK_COUNT] = function( constant, results ) 
		return HoJ:GetBuffStackCount( HoJ._curSpecInfo._activeBuffList, constant ) 
	end,
	[HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO] = function( constant, results ) 
		return HoJ._curSpecInfo._minorHealHealthRatio 
	end,
	[HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO] = function( constant, results ) 
		return HoJ._curSpecInfo._majorHealHealthRatio 
	end,
	[HoJ.TYPE_TARGET_IS_BOSS] = function( constant, results ) 
		--HoJ:DebugMsg( "targetIsBoss = " .. tostring( HoJ._targetIsBoss ) )
		return HoJ._targetIsBoss
	end,
	[HoJ.TYPE_SKILL_OFF_COOLDOWN] = function( constant, results ) 
		return HoJ:IsOffCooldown( constant, HoJ._lookAheadTime )
	end,
	[HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED] = function( constant, results ) 
		return HoJ._targetCanBeInterrupted
	end,
	[HoJ.TYPE_COMBO_POINTS] = function( constant, results )
		return HoJ._comboPoints
	end,
	[HoJ.TYPE_PLAYER_IS_STEALTHED] = function( constant, results )
		return HoJ._isStealthed
	end,
	[HoJ.TYPE_INCOMING_DPS_RATIO] = function( constant, results )
		local dps = HoJ._incomingDpsRatio * constant
		-- HoJ:DebugMsg( "dps = " .. dps )
		return dps
	end,
	[HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL] = function( constant, results )
		local t = ( HoJ._curTime + HoJ._lookAheadTime ) - HoJ._lastSpellDamageTime
		-- HoJ:DebugMsg( "Seconds since damaged by spell = " .. t )
		return t
	end,
	[HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE] = function( constant, results )
		-- HoJ:DebugMsg( "Target TTL = " .. HoJ._combatDpsStats._targetTimeToLive )
		return HoJ._combatDpsStats._targetTimeToLive - HoJ._lookAheadTime
	end,
	[HoJ.TYPE_SPELL_DURATION] = function( constant, results )
		if ( constant == nil or constant._duration == nil ) then
			return 0
		end
		-- HoJ:DebugMsg( constant.name .. "._duration = " .. constant._duration )
		return constant._duration
	end,
	[HoJ.TYPE_TARGET_IS_TOUGH] = function( constant, results )
		return HoJ._targetIsTough
	end,
	[HoJ.TYPE_COMBAT_ATTACKS_HIT] = function( constant, results )
		-- HoJ:DebugMsg( "Combat attacks hit = " .. HoJ._combatDpsStats._attacksHit )
		return HoJ._combatDpsStats._attacksHit
	end,
	[HoJ.TYPE_SKILL_POWER_COST] = function( constant, results )
		return -constant._powerGen
	end,
	[HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF] = function( constant, results ) 
		return HoJ:GetUnitBuffTimeRemaining( "target", constant, HoJ._lookAheadTime, "PLAYER|HARMFUL" ) > 0.0
	end,
	[HoJ.TYPE_SKILL_COOLDOWN_REMAINING] = function( constant, results ) 
		return HoJ:GetCDRemainingAfterGCD( constant )
	end,
	[HoJ.TYPE_UNIT_IS_SLOWED] = function( constant, results ) 
		if ( not HoJ._hasDebuff ) then
			return false
		end
		if ( HoJ._controlLost ) then
			return true
		end
		local speed = GetUnitSpeed( constant )
		return speed > 0 and ( ( speed / BASE_MOVEMENT_SPEED ) < 1.0 )
	end,
	[HoJ.TYPE_NET_INCOMING_DPS] = function( constant, result )
		local dps = HoJ._netIncomingDps * constant
		-- HoJ:DebugMsg( "netIncomingDps = " .. dps )
		return dps
	end,
	[HoJ.TYPE_NET_INCOMING_DPS_RATIO] = function( constant, result )
		local dps = ( HoJ._netIncomingDps * constant ) / HoJ._playerMaxHealth
		-- HoJ:DebugMsg( "netIncomingDpsRatio = " .. dps )
		return dps
	end,
	[HoJ.TYPE_NET_INCOMING_DPS_PCT] = function( constant, result )
		local dps = ( ( HoJ._netIncomingDps * constant ) / HoJ._playerMaxHealth ) * 100.0
		-- HoJ:DebugMsg( "netIncomingDpsRatio = " .. dps )
		return dps
	end,
	[HoJ.TYPE_PLAYER_TIME_TO_LIVE] = function( constant, result )
		return HoJ._playerTimeToLive
	end,
	[HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH] = function( constant, result )
		return HoJ._playerTimeToNearDeath
	end,
	[HoJ.TYPE_NET_INCOMING_DPS_CHR] = function( constant, result )
		local dpschr = ( HoJ._netIncomingDps * constant ) / HoJ._playerHealth
		return dpschr
	end,
	[HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL] = function( constant, result )
		return ( HoJ._incomingMajorDamage ~= nil )
	end,
}

local _opTable = {
	[HoJ.OP_LESS_EQUAL] = function( a, b )
		return a <= b
	end,
	[HoJ.OP_LESS] = function( a, b )
		return a < b
	end,
	[HoJ.OP_EQUAL] = function( a, b )
		--HoJ:DebugMsg( "OP_EQUAL: ( " .. tostring( a ) .. " == " .. tostring( b ) .. " ) = " .. tostring( a == b ) )
		return a == b
	end,
	[HoJ.OP_GREATER] = function( a, b )
		return a > b
	end,
	[HoJ.OP_GREATER_EQUAL] = function( a, b )
		return a >= b
	end,
	[HoJ.OP_NOT_EQUAL] = function( a, b )
		return a ~= b
	end,
	[HoJ.OP_AND] = function( a, b )
		return a and b
	end,
	[HoJ.OP_OR] = function( a, b )
		return a or b
	end,
	[HoJ.OP_IN_RANGE_INCLUSIVE] = function( a, b )
		return a >= b[1] and a <= b[2]
	end,
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_IN] = function( a, b )
		return a > b[1] and a <= b[2]
	end,
	[HoJ.OP_IN_RANGE_MIN_IN_MAX_EX] = function( a, b )
		return a >= b[1] and a < b[2]
	end,
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_EX] = function( a, b )
		return a > b[1] and a < b[2]
	end,
}

local _opNameTable = {
	[HoJ.OP_LESS_EQUAL] = "<=",
	[HoJ.OP_LESS] = "<",
	[HoJ.OP_EQUAL] = "==",
	[HoJ.OP_GREATER] = ">",
	[HoJ.OP_GREATER_EQUAL] = ">=",
	[HoJ.OP_NOT_EQUAL] = "!=",
	[HoJ.OP_AND] = "&&",
	[HoJ.OP_OR] = "||",
	[HoJ.OP_IN_RANGE_INCLUSIVE] = "[]",
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_IN] = "[]",
	[HoJ.OP_IN_RANGE_MIN_IN_MAX_EX] = "()",
	[HoJ.OP_IN_RANGE_MIN_EX_MAX_EX] = "()",
}

function valueToString( value )
	local typeName = type( value )
	if ( typeName == "number" ) then
		return format( "%.2f", value )
	elseif ( typeName == "string" ) then
		return value
	else
		return tostring( value )
	end
end

function valueFn( name, value, constant, results )
	return valueToString( value )
end

function valueScaledFn( name, value, constant, results )
	return "{ " .. valueToString( value ) .. " * " .. format( "%.2f }", constant )
end

function scaledRatioFn( name, value, constant, results )
	return format( "( ( %s * %.2f ) / ? ) { ", name, constant ) .. valueToString( value ) .. " }" 
end

function arrayValueFn( name, value, constant, results )
	return format( "%s[%i]{ ", name, constant ) .. valueToString( value ) .. " }"
end

function resultValueFn( name, value, constant, results )
	return format( "result[%i]:{ ", constant ) .. valueToString( value ) .. " }"
end

function funcValueFn( name, value, constant, results )
	return name .. "( " .. tostring( constant ) .. " ){ " .. valueToString( value ) .. " }"
end

function funcSpellInfoFn( name, value, constant, results )
	return name .. "( " .. constant.name .. " ){ " .. valueToString( value ) .. " }"
end


function spellMemberFn( name, value, constant, results )
	return constant.name .. "." .. name .. "{ " .. value .. " }"
end

local _varNameTable = {
	[HoJ.TYPE_CONSTANT] = {
		_name = "const",
		_fn = valueFn
	},
	[HoJ.TYPE_RESULT] = {
		_name = "result",
		_fn = resultValueFn
	},
	[HoJ.TYPE_POWER] = {
		_name = "_power",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_HEALTH_RATIO] = {
		_name = "_playerHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_TARGET_HEALTH_RATIO] = {
		_name = "_targetHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_BUFF_ACTIVE] = {
		_name = "IsBuffActive(",
		_fn = funcValueFn
	},
	[HoJ.TYPE_TARGET_RANGE] = {
		_name = "InRange",
		_fn = arrayValueFn
	},
	[HoJ.TYPE_TALENT_SELECTED] = {
		_name = "IsTalentSelected",
		_fn = funcSpellInfoFn
	},
	[HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING] = {
		_name = "GetUnitBuffTimeRemaining",
		_fn = funcValueFn
	},
	[HoJ.TYPE_PLAYER_BUFF_REMAINING] = {
		_name = "GetBuffTimeRemaining",
		_fn = varValueFn
	},
	[HoJ.TYPE_SKILL_CHARGES] = {
		_name = "GetSpellCharges",
		_fn = funcValueFn
	},
	[HoJ.TYPE_NUM_MELEE_ATTACKERS] = {
		_name = "_numAttackers",
		_fn = varValue
	},
	[HoJ.TYPE_PLAYER_BUFF_STACK_COUNT] = {
		_name = "GetBuffStackCount",
		_fn = funcValueFn
	},
	[HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO] = {
		_name = "_minorHealHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO] = {
		_name = "_majorHealHealthRatio",
		_fn = valueFn
	},
	[HoJ.TYPE_TARGET_IS_BOSS] = {
		_name = "_targetIsBoss",
		_fn = valueFn
	},
	[HoJ.TYPE_SKILL_OFF_COOLDOWN] = {
		_name = "IsSkillOffCooldown",
		_fn = funcValueFn
	},
	[HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED] = {
		_name = "_targetCanBeInterrupted",
		_fn = valueFn
	},
	[HoJ.TYPE_COMBO_POINTS] = {
		_name = "_comboPoints",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_IS_STEALTHED] = {
		_name = "_isStealthed",
		_fn = valueFn
	},
	[HoJ.TYPE_INCOMING_DPS_RATIO] = {
		_name = "_incomingDpsRatio",
		_fn = valueScaledFn
	},
	[HoJ.TYPE_SECONDS_SINCE_DAMAGED_BY_SPELL] = {
		_name = "SecondsSinceDamageBySpell",
		_fn = funcValueFn
	},
	[HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE] = {
		_name = "_targetTimeToLive",
		_fn = valueFn
	},
	[HoJ.TYPE_SPELL_DURATION] = {
		_name = "_duration",
		_fn = spellMemberFn
	},
	[HoJ.TYPE_TARGET_IS_TOUGH] = {
		_name = "_targetIsTough",
		_fn = valueFn
	},
	[HoJ.TYPE_COMBAT_ATTACKS_HIT] = {
		_name = "_attacksHit",
		_fn = valueFn
	},
	[HoJ.TYPE_SKILL_POWER_COST] = {
		_name = "_powerGen",
		_fn = spellMemberFn
	},
	[HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF] = {
		_name = "GetUnitBuffRemaining",
		_fn = funcValueFn
	},
	[HoJ.TYPE_SKILL_COOLDOWN_REMAINING] = {
		_name = "GetCDRemainingAfterGCD",
		_fn = funcValueFn
	},
	[HoJ.TYPE_UNIT_IS_SLOWED] = {
		_name = "UnitIsSlowed",
		_fn = funcValueFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS] = {
		_name = "_netIncomingDPS",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS_RATIO] = {
		_name = "_netIncomingDPSRatio",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS_PCT] = {
		_name = "_netIncomingDPSPct",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_PLAYER_TIME_TO_LIVE] = {
		_name = "_playerTimeToLive",
		_fn = valueFn
	},
	[HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH] = {
		_name = "_playerTimeToNearDeath",
		_fn = valueFn
	},
	[HoJ.TYPE_NET_INCOMING_DPS_CHR] = {
		_name = "_netIncomingDpsChr",
		_fn = scaledRatioFn
	},
	[HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL] = {
		_name = "_incomingMajorDamageSpell",
		_fn = valueFn
	},
}

-- -----------------------------------------------------------------------------
-- HoJ:BooleanOp
-- compares two values using the specified operation constant
-- returns true or false
-- -----------------------------------------------------------------------------
function HoJ:BooleanOp( op, a, b )
	local func = _opTable[op]
	if ( func ~= nil ) then
		return func( a, b )
	end
	self:Error( "Unknown op: " .. op )
	return false
end

-- -----------------------------------------------------------------------------
-- HoJ:GetValue
-- returns a value based on the type and/or constant
-- -----------------------------------------------------------------------------
function HoJ:GetValue( valueType, constant, results )
	if ( valueType == HoJ.TYPE_CONSTANT ) then
		return constant
	else
		local func = _varTable[valueType]
		if ( func ~= nil ) then
			return func( constant, results )
		end
	end

	self:Error( "Unknown value type: " .. valueType )
	return 0
end

-- -----------------------------------------------------------------------------
-- HoJ:Compare
-- compares two values specified by a type constant and or a value using the
-- specified operation and returns a bool
-- -----------------------------------------------------------------------------
function HoJ:Compare( aType, aConst, bType, bConst, op, results )
	local a = self:GetValue( aType, aConst, results )
	local b = self:GetValue( bType, bConst, results )
	return self:BooleanOp( op, a, b ), a, b
end

-- -----------------------------------------------------------------------------
-- HoJ:DoRotSkillTest
function HoJ:DoRotSkillTest( test, results, debug )
	if ( test == nil ) then
		return true 	-- no test, so always pass
    end
    if ( test.op == nil ) then
        HoJ:DebugMsg( "test has nil operator!" )
        return
	end
	if ( test.aType == nil ) then
		HoJ:DebugMsg( "test.aType = nil in " .. tostring( test._debugName ) )
	end
	if ( test.bType == nil ) then
		HoJ:DebugMsg( "test.bType = nil in " .. tostring( test._debugName ) )
	end
	
	local result, valueA, valueB = self:Compare( test.aType, test.aConst, test.bType, test.bConst, test.op, results )

	if ( debug ) then
		local varATbl = _varNameTable[test.aType]
		local varBTbl = _varNameTable[test.bType]
		HoJ:DebugMsg( varATbl._fn( varATbl._name, valueA, test.aConst, results ) .. 
			" " .. _opNameTable[test.op] .. " " ..
			varBTbl._fn( varBTbl._name, valueB, test.bConst, results ) ..
			" = " .. tostring( result ) )
	end

	return result
end

-- -----------------------------------------------------------------------------
-- GetRuleName
-- -----------------------------------------------------------------------------
function GetRuleName( rule, ruleIndex )
	if ( rule._debugName == nil ) then
		return tostring( ruleIndex )
	end
	return rule._debugName
end

-- -----------------------------------------------------------------------------
-- HoJ:ExecuteRuleTests
-- runs the encoded tests and returns true or false
-- -----------------------------------------------------------------------------
function HoJ:ExecuteRuleTests( ruleSetName, ruleIndex, rule )
    -- if disabled, never return true
	if ( C_SpellBook.IsSpellDisabled( rule.spellInfo.id ) ) then
		return false
    end

	local results = { }
	local numRules = 0
	for i, test in ipairs( rule.tests ) do
		if ( test.op == nil ) then
			HoJ:DebugMsg( "Test " .. i .. " for rule " .. GetRuleName( rule, ruleIndex ) .. 
					" in ruleset '" .. tostring( ruleSetName ) .. "' has a nil operator!" )
			results[i] = false
        else
			local result = self:DoRotSkillTest( test, results, false )
			if ( result == nil ) then
				HoJ:Error( "Rule " .. i .. " in " .. ruleSetName .. " returned nil!")
			end
            results[i] = result
		end
		numRules = numRules + 1
	end

	-- return the last result
	local lastResult = results[#results]
	if ( lastResult == nil ) then
		self:Error( "lastResult = nil in rule set " .. ruleSetName .. "! numRules = " .. numRules )
	elseif ( ( rule._debugSuccess and lastResult == true ) or ( rule._debugFail and lastResult == false ) ) then
----[[		
		local ruleName = GetRuleName( rule, ruleIndex )
		HoJ:Log( true, "Skill '" .. ruleName .. "' results:" )

		-- execute tests again with debug info
		local retestResults = { }
		for i, test in ipairs( rule.tests ) do
			if ( test.op == nil ) then
				HoJ:Log( true, "Test " .. i .. " for " .. ruleName .. " has a nil operator!" )
				retestResults[i] = false
			else
				local result = self:DoRotSkillTest( test, retestResults, true )
				retestResults[i] = result
			end
		end

		for i, r in ipairs(retestResults) do
			HoJ:Log( true, format("  [%i] = %s", i, tostring( r ) ) )
		end
--]]
    end

	return lastResult
end

-- -----------------------------------------------------------------------------
-- HoJ:DumpRotationRules
-- -----------------------------------------------------------------------------
function HoJ:DumpRotationRules( rulesName, rules )
    if ( rules == nil ) then
        HoJ:DebugMsg( "Rules " .. rulesName .. " is nil!" )
        return
    end

    for i, v in ipairs( rules ) do
        if ( v.spellInfo == nil ) then
            HoJ:DebugMsg( "Rule " .. i .. " spellInfo is nil!" )
        else
            HoJ:DebugMsg( "Rule " .. i .. " spellInfo = " .. v.spellInfo.name )
        end
    end
end

-- -----------------------------------------------------------------------------
-- HoJ:CalcFuturePower
-- -----------------------------------------------------------------------------
function HoJ:CalcFuturePower( curTime, duration, nextPowerGainTime, powerGainInterval, curPower, powerStep )

	if ( powerStep <= 0 ) then
		return curPower
	end

	local timeUntilGain = nextPowerGainTime - curTime
	if ( timeUntilGain > duration ) then
		return curPower
	end

	local numIntervals = math.floor( ( duration - timeUntilGain ) / powerGainInterval ) + 1
	local predictedPower = curPower + ( powerStep * numIntervals )
	return predictedPower

end

-- -----------------------------------------------------------------------------
-- HoJ:SelectSkill
-- returns first-best and second-best choices
-- -----------------------------------------------------------------------------
function HoJ:SelectSkill( rulesList )
	--HoJ:DebugMsg( "SelectSkill" )
	if ( rulesList == nil ) then
		-- HoJ:Error( "rulesList is nil!" )
		return
	end

	local isTalentSelected = self.IsTalentSelected
	local isOffCooldown = self.IsOffCooldown
	local executeRuleTests = self.ExecuteRuleTests

	local emptyFn = function( rule, msg )
	end

	local printFn = function( rule, msg )
		if ( rule._debugName ) then
			HoJ:DebugMsg( rule._debugName .. ": " .. msg )
		else
			HoJ:DebugMsg( rule.spellInfo.name .. ": " .. msg )
		end
	end

	local SPELL_TALENT = HoJ.SPELL_TALENT -- local to avoid hash lookups in loop
	local SPELL_TALENT_BUFF = HoJ.SPELL_TALENT_BUFF -- local to avoid hash lookups in loop

	local firstSpell = nil
	local secondSpell = nil

	for i, rule in ipairs( rulesList ) do
		if ( firstSpell == nil or ( firstSpell ~= rule.spellInfo ) ) then
			if ( rule.replacedByTalent == nil or not isTalentSelected( self, rule.replacedByTalent ) ) then
				if ( rule.spellInfo == nil ) then
					if ( rule._debugName ) then
						HoJ:DebugMsg( "Rule " .. rule._debugName .. " in " .. rulesList._name .. " has nil spellInfo!" )
					else
						HoJ:DebugMsg( "Rule " .. i .. " in " .. rulesList._name .. " has nil spellInfo!" )
					end
				elseif ( rule.spellInfo.showOnBar ) then
					if ( ( rule.spellInfo.spellType ~= SPELL_TALENT and rule.spellInfo.spellType ~= SPELL_TALENT_BUFF )
						or isTalentSelected( self, rule.spellInfo ) ) then
						if ( isOffCooldown( self, rule.spellInfo ) ) then
							if ( executeRuleTests( self, rulesList._name, i, rule ) ) then
								if ( firstSpell == nil ) then
									if ( rule.spellInfo == 61304 ) then
										break
									else
										firstSpell = rule.spellInfo
									end
								else
									if ( rule.spellInfo == 61304 ) then
										firstSpell = nil
									else
										secondSpell = rule.spellInfo
									end
									break
								end
							end
						end
					end
				end
			end
		end
	end

	return firstSpell, secondSpell
end

-- -----------------------------------------------------------------------------
-- HoJ:CheckRules
-- -----------------------------------------------------------------------------
function HoJ:CheckRules( skillsList, show, rulesList, requireCombat, requireTargetCanAttack )
	if ( not show or rulesList == nil ) then
		return 
	end

	local combat = ( not requireCombat ) or ( self._inCombat )
	local targetCanAttack = (not requireTargetCanAttack ) or ( self._targetCanAttack )

	if ( not combat and not targetCanAttack ) then
		return 
	end

	local firstSpellInfo, secondSpellInfo = HoJ:SelectSkill( rulesList )
	if ( firstSpellInfo ~= nil ) then
		self:AddIconExclusive( skillsList, firstSpellInfo )
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:PredictPower_AutoRegenerated
-- -----------------------------------------------------------------------------
function HoJ:PredictPower_AutoRegenerated( curTime, curPower, spellInfo, cooldownRemaining )
	local timeUntilRegen = HoJ._nextPowerGainTime - curTime
	local power = HoJ:CalcFuturePower( curTime, timeUntilRegen, HoJ._nextPowerGainTime, 
		HoJ._powerGainInterval, curPower, HoJ._powerGainStep )

	return power
end

-- -----------------------------------------------------------------------------
-- HoJ:PredictPower_SkillGenerated
-- -----------------------------------------------------------------------------
function HoJ:PredictPower_SkillGenerated( curTime, curPower, spellInfo, cooldownRemaining )
	local power = curPower + HoJ._powerPerSec * cooldownRemaining

	return power
end

-- -----------------------------------------------------------------------------
-- HoJ:ShouldPredictSkill
-- -----------------------------------------------------------------------------
function HoJ:ShouldPredictSkill( curTime, curSpecAutoGeneratesPower, spellInfo, debug )
	local shouldPredict = false
	-- only predict a skill if the spec can auto-generate it's power
	-- OR
	-- if the skill generates power or uses no power
	if ( curSpecAutoGeneratesPower == true or spellInfo._powerGen >= 0 ) then
		shouldPredict = true
	end
	return shouldPredict
end

-- -----------------------------------------------------------------------------
-- HoJ:NextSkillOffCooldown
-- returns 
-- nextSkill - skill that will come off cooldown first
-- nextNextSkill - the skill that will come off of cooldown second
-- -----------------------------------------------------------------------------
function HoJ:NextSkillOffCooldown( rotationSelectionList )
	local nextSkill = nil
	local nextNextSkill = nil
	local isTalentSelected = self.IsTalentSelected
	local executeRuleTests = self.ExecuteRuleTests
	local getSpellCooldown = GetSpellCooldown
	local shouldPredictSkill = self.ShouldPredictSkill

	local predictPower = self.PredictPower_SkillGenerated
	if ( HoJ._curSpecInfo._autoPowerGen == true ) then
		predictPower = self.PredictPower_AutoRegenerated
	end

	local SPELL_TALENT = HoJ.SPELL_TALENT -- local to avoid hash lookups in loop
	local SPELL_TALENT_BUFF = HoJ.SPELL_TALENT_BUFF -- local to avoid hash lookups in loop
	local curTime = HoJ._curTime
	local power = HoJ._power
	local gcdRemaining = HoJ._gcdRemaining


	local shortestTime = 999999.0
	local secondShortestTime = 999999.0
	for i, v in ipairs( rotationSelectionList ) do
		if ( shouldPredictSkill( self, curTime, power, v.spellInfo ) ) then
			if ( v.replacedByTalent == nil or not isTalentSelected( self, v.replacedByTalent ) ) then
				if ( ( v.spellInfo.spellType ~= SPELL_TALENT and v.spellInfo.spellType ~= SPELL_TALENT_BUFF )
					or isTalentSelected( self, v.spellInfo ) ) then

					local start, duration, enabled = getSpellCooldown( v.spellInfo.id )
					local cooldownRemaining = gcdRemaining
					if ( start > 0 and duration > 0 ) then
						-- spell is on cooldown (or cooldown is GCD)
						cooldownRemaining = ( start + duration ) - curTime
					end

					HoJ._power = predictPower( self, curTime, power, v.spellInfo, cooldownRemaining )

					if ( executeRuleTests( self, rotationSelectionList._name, i, v ) ) then
						if ( cooldownRemaining < shortestTime ) then
							if ( nextSkill ~= nil ) then
								secondShortestTime = shortestTime
								nextNextSkill = nextSkill
							end
							nextSkill = v.spellInfo
							shortestTime = cooldownRemaining
						elseif ( cooldownRemaining < secondShortestTime ) then
							nextNextSkill = v.spellInfo
							secondShortestTime = cooldownRemaining
						end
					end
				end
			end
		end
	end

	-- reset power
	HoJ._power = power

	return nextSkill, nextNextSkill
end

-- -----------------------------------------------------------------------------
-- HoJ:IsBossUnit
-- returns:
-- - (bool) true if enemy boss
-- - (bool) true if friendly boss
-- -----------------------------------------------------------------------------
function HoJ:IsBossUnit( unit, targetClassification )
	if ( targetClassification == "worldboss" ) then
		return true, false
	end

	local isFriendlyBoss = UnitIsFriend( "player", unit ) and UnitLevel( unit ) == -1
	local isInstanceBoss = false

	for i = 1, 5 do
	    local bossGUID = UnitGUID( "boss" .. i )

	    if UnitGUID(unit) == bossGUID then
	        isInstanceBoss = true
	        break
	    end
	end
	return isInstanceBoss, isFriendlyBoss
end

local INTERACT_INSPECT = 1 -- 28 yards -- this appears to == 30 yards
local INTERACT_TRADE = 2 -- 11.11 yards according to WoWWiki, but exactly equals duel distance now, at ~9.9 yards
local INTERACT_DUEL = 3 -- 9.9 yards according to WoWWiki
local INTERACT_FOLLOW = 4 -- 28 yards -- this is < 30 yards, ~28 yards

local rangeToInteract = {
	[HoJ.RANGE_NONE] = {
		_type = 0,
		_name = "NONE"
	},
	[HoJ.RANGE_SELF] = {
		_type = 0,
		_name = "SELF"
	},
	[HoJ.RANGE_MELEE] = {
		_type = INTERACT_DUEL,
		_name = "MELEE"
	},
	[HoJ.RANGE_8_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "8 yards"
	},
	[HoJ.RANGE_10_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "10 yards"
	},
	[HoJ.RANGE_12_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "12 yards"
	},
	[HoJ.RANGE_15_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "15 yards"
	},
	[HoJ.RANGE_20_YARDS] = {
		_type = INTERACT_DUEL,
		_name = "20 yards"
	},
	[HoJ.RANGE_25_YARDS] = {
		_type = INTERACT_FOLLOW,
		_name = "25 yards"
	},
	[HoJ.RANGE_30_YARDS] = {
		_type = INTERACT_INSPECT,
		_name = "30 yards"
	},
	[HoJ.RANGE_40_YARDS] = {
		_type = INTERACT_INSPECT,
		_name = "40 yards"
	},
	[HoJ.RANGE_100_YARDS] = {
		_type = INTERACT_INSPECT,
		_name = "100 yards"
	},
}

-- -----------------------------------------------------------------------------
-- HoJ:InRangeOfSpells
-- -----------------------------------------------------------------------------
function HoJ:InRangeOfSpells( rangeEnum, t, target, debugName, debug )
	-- spell info may be a table array if multiple spells can be used
	local inRange = false
	local subTable = t[rangeEnum]
	if ( subTable == nil ) then
		local canInteract = false
		local tbl = rangeToInteract[rangeEnum]
		if ( tbl._type < 1 ) then
			inRange = true
		else
			inRange = CheckInteractDistance( target, tbl._type )
		end
		if ( debug ) then
			HoJ:FrameThrottleMsg( 5, "CheckInteractDistance(" .. tbl._type .. ") = " .. tostring( inRange ) .. " for " .. tbl._name )
		end
	else 
		local spellInfos = subTable._spellInfos
		if ( spellInfos ~= nil ) then
			for i, spell in ipairs( spellInfos ) do
				if ( HoJ:InRange( spell, target, debugName ) ) then
					inRange = true
					break
				end
			end
		else
			inRange = HoJ:InRange( subTable._spellInfo, target, debugName )
		end
	end
	if ( debug ) then
		HoJ:FrameThrottleMsg( 5, "InRangeOfSpells(" .. debugName .. ") = " .. tostring( inRange ) )
	end
	return inRange
end

-- -----------------------------------------------------------------------------
-- HoJ:FindRanges
-- -----------------------------------------------------------------------------
function HoJ:FindRanges( target, specInfo )
	-- target range finding
	HoJ._targetInRange[HoJ.RANGE_NONE] = true;
	HoJ._targetInRange[HoJ.RANGE_SELF] = true;
    HoJ._targetInRange[HoJ.RANGE_MELEE] = false;
    HoJ._targetInRange[HoJ.RANGE_8_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_10_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_12_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_15_YARDS] = false;
    HoJ._targetInRange[HoJ.RANGE_20_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_25_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_30_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_40_YARDS] = false;
	HoJ._targetInRange[HoJ.RANGE_100_YARDS] = false;

	local t = nil
	if ( ( not UnitExists( target ) ) or UnitIsDead( target ) ) then
		return
	elseif ( UnitCanAttack( "player", target ) ) then
		t = specInfo._enemyRangeSpells
	elseif ( UnitCanAssist( "player", target ) ) then
		t = specInfo._friendlyRangeSpells
	else
		-- HoJ:DebugMsg( "Cannot find range to target!" )
		return
	end

	local debug = HoJ._showRange

    HoJ._targetInRange[HoJ.RANGE_MELEE] 		= HoJ:InRangeOfSpells( HoJ.RANGE_MELEE,    t, target, "Melee", debug )
    HoJ._targetInRange[HoJ.RANGE_8_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_8_YARDS,  t, target, "8", debug )
	HoJ._targetInRange[HoJ.RANGE_10_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_10_YARDS, t, target, "10", debug )
	HoJ._targetInRange[HoJ.RANGE_12_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_12_YARDS, t, target, "12", debug )
	HoJ._targetInRange[HoJ.RANGE_15_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_15_YARDS, t, target, "15", debug )
    HoJ._targetInRange[HoJ.RANGE_20_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_20_YARDS, t, target, "20", debug )
    HoJ._targetInRange[HoJ.RANGE_25_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_25_YARDS, t, target, "25", debug )
	HoJ._targetInRange[HoJ.RANGE_30_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_30_YARDS, t, target, "30", debug )
	HoJ._targetInRange[HoJ.RANGE_40_YARDS] 		= HoJ:InRangeOfSpells( HoJ.RANGE_40_YARDS, t, target, "40", debug )
	HoJ._targetInRange[HoJ.RANGE_100_YARDS] 	= HoJ:InRangeOfSpells( HoJ.RANGE_100_YARDS, t, target, "100", debug )

    if ( debug ) then
        local foundRange = false
        for i, v in ipairs( HoJ._targetInRange ) do
            if ( i >= HoJ.RANGE_MELEE ) then
                if ( v == true ) then                  
                    HoJ:FrameThrottleMsg( 5, "Target range = " .. HoJ._rangeNames[i] )
                    foundRange = true
                    break
                end
            end
        end
        if ( not foundRange ) then
            HoJ:FrameThrottleMsg( 5, "Target range = max" )
        end
    end
end
