-- -----------------------------------------------------------------------------
-- -----------------------------------------------------------------------------
--  Hammer of Justas
--
-- Copyright (C) 2010-2020
-- By Jonathan E. Wright
--
-- FILE: HoJ_Warrior.lua
-- DESC: Warrior-specific functions.

-- BUGS:
-- - another warrior's battleshout buff doesn't make buff bar go away
-- TODO:
-- X rallying cry as mitigation
-- X dragonRoar as mitigation
-- X victory rush almost never shows right now because health must be below 60%
--   X with impending victory this seemed ok because you can use it more often
--   X maybe need a health ratio on healing spells to make it part of the rules
-- X recklessness option to show in rotation
-- X recklessness (add as "dpsCooldowns", also convert avenging wrath)
-- X battleshout (as a blessing)
--
-- NOTES:
-- * if target is near to dying and the warrior needs to heal, Impending Victory could be 
--   shown as a cooldown as a reminder to use it before the target dies (since it will 
--   reset when the target dies)

local addonName, HoJ = ...

-- DEFAULT_CHAT_FRAME:AddMessage( "HoJ_Warrior.lua" )

-- execute 5308
-- sudden death 280721
-- bloodthirst 23881
-- enrage 184361
-- raging blow 85288
-- whirlwind 190411
local warriorSpellInfo = HoJ._warriorSpellInfos

-- -----------------------------------------------------------------------------
-- HoJ:Warrior_SelectSingleOrMultitarget
-- returns
--   rotationRules
--   interruptRules
--   mitigationRules
--   dpsCooldownRules
-- -----------------------------------------------------------------------------
function HoJ:Warrior_SelectSingleOrMultitarget( curSpecInfo, numAttackers, threshold )
    if ( self._numMeleeAttackers > threshold ) then
        return curSpecInfo._rotationRules._multiTarget, 
                curSpecInfo._interruptRules._multiTarget,
                curSpecInfo._mitigationRules._multiTarget,
                curSpecInfo._dpsCooldownRules._multiTarget
    else
        return curSpecInfo._rotationRules._singleTarget, 
                curSpecInfo._interruptRules._singleTarget,
                curSpecInfo._mitigationRules._singleTarget,
                curSpecInfo._dpsCooldownRules._singleTarget
    end
end


-- -----------------------------------------------------------------------------
-- HoJ:Warrior_CheckBlessings
-- CLASSFIX - paladin specific
-- -----------------------------------------------------------------------------
function HoJ:Warrior_CheckBlessings( activeBlessingsList )
	if ( not HoJ._showBuffsFrame ) then
		return
	end

	-- this has to be called every time we check blessings to update non-party member blessings
	if ( self._updateFrameNum >= self._nextUpdateBlessingFrame ) then
		HoJ:UpdateBlessings( activeBlessingsList )
	end

	--HoJ:DebugMsg( "Warrior_CheckBlessings )

	-- dont show any blessing buffs for a second after entering world
	if ( self._updateFrameNum < self._enteredWorldUpdateFrameNum + HoJ.UPDATES_PER_SEC ) then
		return
	end


	if ( HoJ._level >= HoJ.SpellInfo.battleShout.minLevel ) then
		-- handle party sync disabling this buff
		if ( not C_SpellBook.IsSpellDisabled( HoJ.SpellInfo.battleShout.id ) ) then
			local battleShout = activeBlessingsList[HoJ.SpellInfo.battleShout.id]
			if ( battleShout.start <= 0.0 ) then
				--HoJ:DebugMsg( "Adding Blessing of Kings icon" )
				HoJ:AddIcon( HoJ._buffs, HoJ.SpellInfo.battleShout )
			end
		end
	end
end

-- -----------------------------------------------------------------------------
-- HoJ:Warrior_Update
-- -----------------------------------------------------------------------------
function HoJ:Warrior_Update( curSpecInfo, updateCooldowns, updateBuffs, updateRotation )
    -- HoJ:DebugMsg( "Updating spec " .. specInfo._name );
    local rotationRules, interruptRules, mitigationRules, dpsCooldownRules = curSpecInfo._selectSingleOrMultiTargetFn( 
        HoJ, curSpecInfo,
        self._numMeleeAttackers, 1 )

    -- update buffs
    if ( updateBuffs ) then
        local startBuffsMs = HoJ.ProfileBegin2( "Buffs" )
        
        HoJ:ClearIconList( HoJ._buffs );

        HoJ:CheckProcdBuffs( curSpecInfo._activeBuffList, curSpecInfo._procdBuffList );
        HoJ:Warrior_CheckBlessings( curSpecInfo._activeBlessings );

        HoJ.ProfileFunction( "UpdateIcons:Buffs2:", HoJ.UpdateIcons2, self, HoJ._buffs, 
                curSpecInfo._buffSortPriorities, HoJ._buffFrames, true )

        HoJ.ProfileEnd( "Buffs", startBuffsMs )
    end

    -- update cooldowns
    if ( updateCooldowns ) then
        local startCooldownsMs = HoJ.ProfileBegin2( "Cooldowns" )

        HoJ.ProfileFunction( "ClearIconList", HoJ.ClearIconList, self, HoJ._cooldowns )

        HoJ.ProfileFunction( "CheckRules(dpsCooldowns)", HoJ.CheckRules, self, self._cooldowns,
                self._showCooldownsFrame, dpsCooldownRules, true, true )
        HoJ.ProfileFunction( "CheckRules(interrupts)", HoJ.CheckRules, self, self._cooldowns,
                self._showCooldownsFrame, interruptRules, true, true )
        HoJ.ProfileFunction( "CheckRules(mitigations)", HoJ.CheckRules, self, self._cooldowns, 
                self._showCooldownsFrame, mitigationRules, true, true )

        HoJ.ProfileFunction( "UpdateIcons:Cooldowns:", HoJ.UpdateIcons2, self, self._cooldowns, 
                curSpecInfo._cooldownSortPriorities, self._cooldownFrames, true )

        HoJ.ProfileEnd( "Cooldowns", startCooldownsMs )
    end

    -- update rotation
    if ( updateRotation ) then
        local startRotationMs = HoJ.ProfileBegin2( "Rotation" )

        HoJ:ClearIconList( HoJ._rotationSkills )

        HoJ:CheckDPSRotation( HoJ._rotationSkills, rotationRules )

        HoJ.ProfileFunction( "UpdateIcons:Rotation:", HoJ.UpdateIcons, self, HoJ._rotationSkills, HoJ._rotationFrames, false );

        HoJ.ProfileEnd( "Rotation", startRotationMs )
    end    
end

local armsMeleeAoESpells = { 

}
local furyMeleeAoESpells = {

}

local protMeleeAoESpells = {

}

local lowHealthMitigationList = { }
local moderateMitigationList = { }
local emergencyMitigationList = { }


--------------------------------------------------------------------------------
-- ARMS
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/arms-warrior-pve-dps-rotation-cooldowns-abilities
local armsRotationRules = {
    _name = "armsRotationRules",
    {
        -- impending victory if below health ratio
        spellInfo = warriorSpellInfo.impendingVictory,
        -- _debugName = "impendingVictory",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.rend,
        -- _debugName = "rend",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.rend, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.rend.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.colossusSmashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.warbreakerDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.rendDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.skullSplitter,
        -- _debugName = "skullSplitter",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 60, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.skullSplitter.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
--[[
    {
        spellInfo = warriorSpellInfo.avatar,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.avatar, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
--]]
    {
        spellInfo = warriorSpellInfo.ravager,
        -- _debugName = "ravager",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.ravager, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ravager.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.warbreaker,
        -- _debugName = "warbreaker",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.warbreaker, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.warbreaker.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.colossusSmash,
        -- _debugName = "colossusSmash",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.warbreaker, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.colossusSmash.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },            
        }
    },
    {
        spellInfo = warriorSpellInfo.deadlyCalm,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.deadlyCalm, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.armsExecute,
        -- _debugName = "armsExecute",
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.armsMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.armsSuddenDeathBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.armsExecute.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.overpower,
        -- _debugName = "overpower",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.overpower.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.mortalStrike,
        -- _debugName = "mortalStrike_to_refresh_deepWounds",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.mortalStrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.mortalStrike, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.deepWoundsDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- bladestorm during colossus smash
        spellInfo = warriorSpellInfo.armsBladestorm,
        -- _debugName = "armsBladestorm",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.armsBladestorm.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.armsBladestorm.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.ravager, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        -- storm bolt if it's talented and within range
        spellInfo = warriorSpellInfo.stormBolt,
        -- _debugName = "stormBolt",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.stormBolt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.mortalStrike,
        -- _debugName = "mortalStrike",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.mortalStrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.mortalStrike, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- for multi-target only
--[[
    {
        spellInfo = warriorSpellInfo.armsWhirlwind,
        -- _debugName = "armsWhirlwind",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.armsWhirlwind, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.fervorOfBattle, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.armsWhirlwind.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
--]]
    {  
        spellInfo = warriorSpellInfo.charge,
        -- _debugName = "charge",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.charge.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        -- _debugName = "heroicLeap",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicLeap.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- heroic throw if within range
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        -- _debugName = "heroicThrow",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicThrow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }                
        }
    },
    {
        spellInfo = warriorSpellInfo.slam,
        -- _debugName = "slam",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.slam, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.slam.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local armsMultiTargetRotationRules = {
    _name = "armsMultiTargetRotationRules",
    {
        -- impending victory if below health ratio
        spellInfo = warriorSpellInfo.impendingVictory,
        -- _debugName = "impendingVictory",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.sweepingStrikes,
        -- _debugName = "sweepingStrikes",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = warriorSpellInfo.sweepingStrikes.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.sweepingStrikes.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.armsBladestorm.id, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.sweepingStrikes, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.ravager.id, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.sweepingStrikes, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.rend,
        -- _debugName = "rend",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.rend, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.rend.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.colossusSmashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.warbreakerDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.rendDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 4.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 6, bType = HoJ.TYPE_RESULT, bConst = 7, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.skullSplitter,
        -- _debugName = "skullSplitter",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 60, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.skullSplitter.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.armsBladestorm.id, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.sweepingStrikes, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_SKILL_COOLDOWN_REMAINING, aConst = warriorSpellInfo.ravager.id, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.sweepingStrikes, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
--[[
    {
        spellInfo = warriorSpellInfo.avatar,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.avatar, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
--]]
    {
        spellInfo = warriorSpellInfo.ravager,
        -- _debugName = "ravager",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.ravager, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ravager.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.warbreaker,
        -- _debugName = "warbreaker",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.warbreaker, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.warbreaker.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.colossusSmash,
        -- _debugName = "colossusSmash",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.warbreaker, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.colossusSmash.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },            
        }
    },
    {
        -- bladestorm during colossus smash
        spellInfo = warriorSpellInfo.armsBladestorm,
        -- _debugName = "armsBladestorm",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.ravager, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.armsBladestorm.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.colossusSmashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HAS_PLAYER_DEBUFF, aConst = warriorSpellInfo.warbreakerDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.deadlyCalm,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.deadlyCalm, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.armsExecute,
        -- _debugName = "armsExecute",
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.armsMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.armsSuddenDeathBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.armsExecute.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.overpower,
        -- _debugName = "overpower",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.overpower.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.mortalStrike,
        -- _debugName = "mortalStrike_to_refresh_deepWounds",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.mortalStrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.mortalStrike, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_PLAYER_DEBUFF_REMAINING, aConst = warriorSpellInfo.deepWoundsDebuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
--[[
    {
        -- storm bolt if it's talented and within range
        spellInfo = warriorSpellInfo.stormBolt,
        -- _debugName = "stormBolt",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.stormBolt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
==]]
    {
        spellInfo = warriorSpellInfo.mortalStrike,
        -- _debugName = "mortalStrike",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.mortalStrike.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.mortalStrike, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {  
        spellInfo = warriorSpellInfo.charge,
        -- _debugName = "charge",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.charge.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        -- _debugName = "heroicLeap",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicLeap.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- heroic throw if within range
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        -- _debugName = "heroicThrow",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicThrow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }                
        }
    },
    {
        spellInfo = warriorSpellInfo.slam,
        -- _debugName = "slam_during_sweepingStrikes_without_fervorOfBattle",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.slam, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.slam.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.fervorOfBattle, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.sweepingStrikes.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.armsWhirlwind,
        -- _debugName = "armsWhirlwind",
        tests = {
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.armsWhirlwind, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.fervorOfBattle, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.armsWhirlwind.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
}

local armsInterruptRules = {
    _name = "armsInterruptRules",
    {
        spellInfo = warriorSpellInfo.pummel,
        -- _debugName = "rebuke",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.pummel.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        -- _debugName = "intimidatingShout",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.intimidatingShout.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local armsMitigationRules = {
    _name = "armsMitigationRules",
    -- emergency mitigations
    -- these are selected if:
    --   - the amount of damage predicted over the next 5 seconds exceeds the player's current health
    --   - predicted death time is < 10 seconds
    -- NOTE: non-instant healing is not considered a mitigation
    {
        spellInfo = warriorSpellInfo.dieByTheSword,
        -- _debugName = "enragedRegeneration",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.stoneform,
        -- _debugName = "stoneform",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.rallyingCry,
        -- _debugName = "rallyingCry",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        -- _debugName = "heroicLeap",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        -- _debugName = "piercingHowl",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.berserkerRage,
        -- _debugName = "berserkerRage",
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.avatar,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local armsDpsCooldownRules = {
    {
        spellInfo = warriorSpellInfo.avatar,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.avatar, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.recklessness.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local armsCooldownSortPriorities = {
    _name = "armsCooldownSortPriorities",
    [warriorSpellInfo.avatar.id] = 8,
    [warriorSpellInfo.dieByTheSword.id] = 7,
    [warriorSpellInfo.stoneform.id] = 6,
    [warriorSpellInfo.pummel.id] = 5,
    [warriorSpellInfo.intimidatingShout.id] = 4,
    [warriorSpellInfo.heroicLeap.id] = 3,
    [warriorSpellInfo.rallyingCry.id] = 2,
    [warriorSpellInfo.berserkerRage.id] = 1,
}

local armsActiveBuffList = { 
    [warriorSpellInfo.avatar.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.overpower.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.armsSuddenDeathBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.victorious.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.bladestorm.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.sweepingStrikes.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
}

-- this is a list of buffs we want to check for showing on the buffs bar
local armsProcdBuffList = {
    { _spellInfo = warriorSpellInfo.victorious },
}

local armsActiveBlessingList = {
    [warriorSpellInfo.battleShout.id] = {
        spellInfo = warriorSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local armsBuffSortPriorities = {
    _name = "armsBuffSortPriorities",
    [warriorSpellInfo.victorious.id] = 7,
    [warriorSpellInfo.avatar.id] = 6,
    [warriorSpellInfo.armsSuddenDeathBuff.id] = 5,
    [warriorSpellInfo.overpower.id] = 4,
    [warriorSpellInfo.bladestorm.id] = 3,
    [warriorSpellInfo.sweepingStrikes.id] = 2,
    [warriorSpellInfo.battleShout.id] = 1,
}

--------------------------------------------------------------------------------
-- FURY
--------------------------------------------------------------------------------

-- https://www.icy-veins.com/wow/fury-warrior-pve-dps-rotation-cooldowns-abilities
local furyRotationRules = {
    _name = "furyRotationRules",
    {
        spellInfo = warriorSpellInfo.impendingVictory,
        -- _debugName = "impendingVictory",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.victoryRush,
        -- _debugName = "victoryRush",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.victoryRush.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.furiousSlash,
        -- _debugName = "furiousSlash",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = warriorSpellInfo.furiousSlashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.furiousSlash.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.rampage,
        -- _debugName = "rampage",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 90, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    --[[
    {
        spellInfo = warriorSpellInfo.recklessness, -- if not on cooldown
        -- _debugName = "recklessness",
        tests = {
            -- TODO: allow this to be removed from the rotation and used only as a cooldown?
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    --]]
    {
        spellInfo = warriorSpellInfo.siegebreaker,
        -- _debugName = "siegebreaker",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.siegebreaker.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.execute,
        -- _debugName = "execute",
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.furyMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furySuddenDeathBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.execute.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        -- _debugName = "bloodthirst_no_enrage",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        -- _debugName = "ragingBlow_2_charges",
        tests = {
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = warriorSpellInfo.ragingBlow.id, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- bloodthirst only if below minor heal health ratio
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        -- _debugName = "bloodthirst_low_health",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        -- _debugName = "ragingBlow_off_cooldown",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    -- bloodthirst as a filler if it's available
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        -- _debugName = "bloodthirst_off_cooldown",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.dragonRoar,
        -- _debugName = "dragonRoar_during_enrage",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.dragonRoar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        -- furious slash if ranging blow and bloodthirst are on cooldown
        spellInfo = warriorSpellInfo.furiousSlash, 
        --_debugName = "furiousSlash_filler",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.furiousSlash.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_STACK_COUNT, aConst = warriorSpellInfo.furiousSlashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.charge,
        -- _debugName = "charge",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.charge.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- prefer whirlwind as the first filler, if it is in range
    {
        spellInfo = warriorSpellInfo.whirlwind,
        -- _debugName = "whirlwind",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    -- storm bolt if it's talented and within range
    {
        spellInfo = warriorSpellInfo.stormBolt,
        -- _debugName = "stormBolt",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.stormBolt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        -- _debugName = "heroicLeap",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicLeap.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- heroic throw if within range
    {
        spellInfo = warriorSpellInfo.heroicThrow,
        -- _debugName = "heroicThrow",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicThrow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }                
        }
    },
    -- whirlwind for the heck of it, maybe another target is in range
    {
        spellInfo = warriorSpellInfo.whirlwind,
        -- _debugName = "whirlwind",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.whirlwind.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

-- https://www.icy-veins.com/wow/fury-warrior-pve-dps-rotation-cooldowns-abilities
-- note that furious slash is not used in the multi-target rotation based on the icy-veins recommendation
local furyMultiTargetRotationRules = {
    _name = "furMultiTargetRotationRules",
    {
        spellInfo = warriorSpellInfo.impendingVictory,
        -- _debugName = "impendingVictory",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.impendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },        
    {
        spellInfo = warriorSpellInfo.victoryRush,
        -- _debugName = "rampvictoryRush",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.impendingVictory, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.victoryRush.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.furiousSlash,
        -- _debugName = "furiousSlash",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_REMAINING, aConst = warriorSpellInfo.furiousSlashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 2.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.furiousSlash.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.whirlwind,
        -- _debugName = "whirlwind",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.whirlwindBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.siegebreaker,
        -- _debugName = "rampage",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.siegebreaker.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.rampage,
        -- _debugName = "rampage",
        tests = {
            { aType = HoJ.TYPE_POWER, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 75, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        spellInfo = warriorSpellInfo.bladestorm,
        -- _debugName = "rampage",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bladestorm.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL }
        }
    },
    {
        spellInfo = warriorSpellInfo.dragonRoar,
        -- _debugName = "rampage",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.dragonRoar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL }
        }
    },
    -- NOTE: we don't included recklessness in the rotation -- it's a cooldown
    {
        spellInfo = warriorSpellInfo.execute,
        -- _debugName = "execute",
        -- if ( inRange and ( ( hasMassacreTalent and targetHealth < 0.35 ) or suddenDeathBuffed or targetHealth < 0.2 ) )
        tests = {
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.furyMassacre, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.35, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furySuddenDeathBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 0.2, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.execute.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 7, bType = HoJ.TYPE_RESULT, bConst = 8, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        -- _debugName = "bloodthirst_no_enrage",
        tests = {
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.furyEnrageBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        -- _debugName = "ragingBlow",
        tests = {
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = warriorSpellInfo.ragingBlow.id, bType = HoJ.TYPE_CONSTANT, bConst = 2, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    -- bloodthirst only if below minor heal health ratio
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        -- _debugName = "bloodthirst_low_health",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MINOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.ragingBlow,
        -- _debugName = "ragingBlow_off_cooldown",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.ragingBlow.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    -- bloodthirst as a filler if it's available
    {
        spellInfo = warriorSpellInfo.bloodthirst,
        -- _debugName = "bloodthirst_off_cooldown",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.bloodthirst.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.charge,
        -- _debugName = "charge",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.charge.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        -- _debugName = "heroicLeap",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.heroicLeap.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = HoJ.RANGE_8_YARDS, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND }
        }
    },
    {
        -- furious slash if ranging blow and bloodthirst are on cooldown, and we have less than three stacks
        spellInfo = warriorSpellInfo.furiousSlash, 
        -- _debugName = "furiousSlash_filler",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.furiousSlash.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_STACK_COUNT, aConst = warriorSpellInfo.furiousSlashBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = 3, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    -- prefer whirlwind as a filler in multi-target situations, even if our target is out of range
    {
        spellInfo = warriorSpellInfo.whirlwind,
        -- _debugName = "whirlwind",
        tests = {
            { aType = HoJ.TYPE_CONSTANT, aConst = true, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local furyInterruptRules = {
    _name = "furyInterruptRules",
    {
        spellInfo = warriorSpellInfo.pummel,
        -- _debugName = "pummel",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.pummel.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.stormBolt,
        -- _debugName = "stormBolt",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.stormBolt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        -- _debugName = "intimidatingShout",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.intimidatingShout.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local furyMitigationRules = {
    _name = "furyMitigationRules",
    -- emergency mitigations
    -- these are selected if:
    --   - the amount of damage predicted over the next 5 seconds exceeds the player's current health
    --   - predicted death time is < 10 seconds
    -- NOTE: non-instant healing is not considered a mitigation
    {
        spellInfo = warriorSpellInfo.enragedRegeneration,
        -- _debugName = "enragedRegeneration",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.stoneform,
        -- _debugName = "stoneform",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.rallyingCry,
        -- _debugName = "rallyingCry",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.heroicLeap,
        -- _debugName = "heroicLeap",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        -- _debugName = "piercingHowl",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.dragonRoar,
        -- _debugName = "dragonRoar",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.piercingHowl,
        -- _debugName = "piercingHowl",
        tests = {
            { aType = HoJ.TYPE_INCOMING_DPS_RATIO, aConst = 5.0, bType = HoJ.TYPE_PLAYER_HEALTH_RATIO, bConst = 0 , op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_OR },
        }
    },
    {
        spellInfo = warriorSpellInfo.berserkerRage,
        -- _debugName = "berserkerRage",
        tests = {
            { aType = HoJ.TYPE_UNIT_IS_SLOWED, aConst = "player", bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local furyDpsCooldownRules = {
    _name = "furyDpsCooldownRules",
    {
        spellInfo = warriorSpellInfo.recklessness,
        -- _debugName = "recklessness",
        tests = {
            { aType = HoJ.TYPE_COMBAT_TARGET_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_SPELL_DURATION, bConst = warriorSpellInfo.recklessness, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_COMBAT_ATTACKS_HIT, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_IS_TOUGH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.recklessness.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 5, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local furyCooldownSortPriorities = {
    _name = "furyCooldownSortPriorities",
    [warriorSpellInfo.pummel.id] = 11,
    [warriorSpellInfo.stormBolt.id] = 10,
    [warriorSpellInfo.enragedRegeneration.id] = 9,
    [warriorSpellInfo.stoneform.id] = 8,
    [warriorSpellInfo.recklessness.id] = 7,
    [warriorSpellInfo.rallyingCry.id] = 6,
    [warriorSpellInfo.heroicLeap.id] = 5,
    [warriorSpellInfo.intimidatingShout.id] = 4,
    [warriorSpellInfo.dragonRoar.id] = 2,
    [warriorSpellInfo.piercingHowl.id] = 3,
    [warriorSpellInfo.berserkerRage.id] = 1,
}

local furyActiveBuffList = { 
    [warriorSpellInfo.furyEnrageBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.furySuddenDeathBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.whirlwindBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.furiousSlashBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.victorious.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
}

-- this is the list of buffs that we want to show on the buffs bar
-- in general we just show buffs that proc randomly here, or buffs with a stack count that needs tracking
local furyProcdBuffList = {
    { _spellInfo = warriorSpellInfo.furyEnrageBuff },
    { _spellInfo = warriorSpellInfo.furySuddenDeathBuff },
    { _spellInfo = warriorSpellInfo.whirlwindBuff },
    { _spellInfo = warriorSpellInfo.furiousSlashBuff },
    { _spellInfo = warriorSpellInfo.victorious },
}

local furyActiveBlessingList = {
    [warriorSpellInfo.battleShout.id] = {
        spellInfo = warriorSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    }
}

local furyBuffSortPriorities = {
    _name = "furyBuffSortPriorities",
    [warriorSpellInfo.victorious.id] = 6,
    [warriorSpellInfo.furyEnrageBuff.id] = 5,
    [warriorSpellInfo.furySuddenDeathBuff.id] = 4,
    [warriorSpellInfo.whirlwindBuff.id] = 3,
    [warriorSpellInfo.furiousSlashBuff.id] = 2,
    [warriorSpellInfo.battleShout.id] = 1,
}

--------------------------------------------------------------------------------
-- PROTECTION
--------------------------------------------------------------------------------
local protRotationRules = {
    _name = "protRotationRules",
    -- 1) avatar on cooldown
    -- 2) demoralizing shout on cooldown (only with booming voice)
    -- 3) shield slam on cooldown
    -- 4) thunder clap on cooldown
    -- 5) revenge if free
    -- 6) devastate
    {
        spellInfo = warriorSpellInfo.victoryRush,
        -- _debugName = "victoryRush",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.protImpendingVictory, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.victoryRush.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.impendingVictory,
        -- _debugName = "impendingVictory",
        tests = {
            { aType = HoJ.TYPE_PLAYER_HEALTH_RATIO, aConst = 0, bType = HoJ.TYPE_MAJOR_HEAL_HEALTH_RATIO, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.protImpendingVictory.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.protAvatar,
        -- _debugName = "avatar",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.avatar.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.demoralizingShoutWithBoomingVoice,
        -- _debugName = "demoralizingShout_with_boomingVoice",
        -- _debugFail = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.demoralizingShoutWithBoomingVoice.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TALENT_SELECTED, aConst = warriorSpellInfo.boomingVoice, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldSlam,
        -- _debugName = "shieldSlam",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shieldSlam.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.thunderClap,
        -- _debugName = "thunderClap",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.thunderClap.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
    {
        spellInfo = warriorSpellInfo.revenge,
        -- _debugName = "revenge_if_free",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.revenge.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.revengeBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.devastate,
        -- _debugName = "devastate",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.devastate.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
        }
    },
}

local protMitigationRules = {
    _name = "protMitigationRules",
    {
        spellInfo = warriorSpellInfo.lastStand,
        -- _debugName = "lastStand",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 10.0, op = HoJ.OP_LESS },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldWall,
        -- _debugName = "shieldWall",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 20.0, op = HoJ.OP_LESS },
        }
    },
    {
        spellInfo = warriorSpellInfo.ignorePain,
        -- _debugName = "ignorePain",
        tests = {
            { aType = HoJ.TYPE_PLAYER_TIME_TO_LIVE, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 20.0, op = HoJ.OP_LESS },
            { aType = HoJ.TYPE_SKILL_POWER_COST, aConst = warriorSpellInfo.ignorePain, bType = HoJ.TYPE_POWER, bConst = 0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shieldBlock,
        -- _debugName = "shieldBlock",
        tests = {
            { aType = HoJ.TYPE_SKILL_CHARGES, aConst = warriorSpellInfo.shieldBlock.id, bType = HoJ.TYPE_CONSTANT, bConst = 1, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_PLAYER_BUFF_ACTIVE, aConst = warriorSpellInfo.shieldBlockBuff.id, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_PCT, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.25, op = HoJ.OP_GREATER_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shockwave,
        -- _debugName = "shockwave",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shockwave.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_NET_INCOMING_DPS_PCT, aConst = 1.0, bType = HoJ.TYPE_CONSTANT, bConst = 0.25, op = HoJ.OP_GREATER },
            { aType = HoJ.TYPE_PLAYER_TIME_TO_NEAR_DEATH, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = 20.0, op = HoJ.OP_LESS_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 4, bType = HoJ.TYPE_RESULT, bConst = 5, op = HoJ.OP_OR },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 6, op = HoJ.OP_AND },
        }
    },
}

local protInterruptRules = {
    {
        spellInfo = warriorSpellInfo.pummel,
        -- _debugName = "pummel",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.pummel.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.shockwave,
        -- _debugName = "shockwave",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.shockwave.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_TARGET_CAN_BE_INTERRUPTED, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.protStormBolt,
        -- _debugName = "protStormBolt",
        -- _debugFail = true,
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.protStormBolt.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
            { aType = HoJ.TYPE_INCOMING_MAJOR_DAMAGE_SPELL, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 3, bType = HoJ.TYPE_RESULT, bConst = 4, op = HoJ.OP_AND },
        }
    },
    {
        spellInfo = warriorSpellInfo.intimidatingShout,
        -- _debugName = "intimidatingShout",
        tests = {
            { aType = HoJ.TYPE_TARGET_RANGE, aConst = warriorSpellInfo.intimidatingShout.range, bType = HoJ.TYPE_CONSTANT, bConst = true, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_TARGET_IS_BOSS, aConst = 0, bType = HoJ.TYPE_CONSTANT, bConst = false, op = HoJ.OP_EQUAL },
            { aType = HoJ.TYPE_RESULT, aConst = 1, bType = HoJ.TYPE_RESULT, bConst = 2, op = HoJ.OP_AND },
        }
    },
}

local protCooldownSortPriorities = {
    _name = "protCooldownSortPriorities",
    [warriorSpellInfo.shieldWall.id] = 9,
    [warriorSpellInfo.lastStand.id] = 8,
    [warriorSpellInfo.ignorePain.id] = 7,
    [warriorSpellInfo.spellReflection.id] = 6,
    [warriorSpellInfo.shieldBlock.id] = 5,
    [warriorSpellInfo.pummel.id] = 4,
    [warriorSpellInfo.shockwave.id] = 3,
    [warriorSpellInfo.protStormBolt.id] = 2,
    [warriorSpellInfo.intimidatingShout.id] = 1,
}

local protActiveBuffList = { 
    [warriorSpellInfo.revengeBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.victorious.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
    [warriorSpellInfo.shieldBlockBuff.id] = { isActive = false, start = 0, duration = 0, timeRemaining = 0, stacks = 0 },
}

local protProcdBuffList = {
    { _spellInfo = warriorSpellInfo.revengeBuff },
    { _spellInfo = warriorSpellInfo.victorious },
    { _spellInfo = warriorSpellInfo.shieldBlockBuff },
}

local protActiveBlessingList = {
    [warriorSpellInfo.battleShout.id] = {
        spellInfo = warriorSpellInfo.battleShout,
        targetName = nil,
        start = 0.0,
        duration = 0.0,
        updateIndex = -1
    },
}

local protBuffSortPriorities = {
    _name = "protBuffSortPriorities",
    [warriorSpellInfo.shieldBlockBuff.id] = 4,
    [warriorSpellInfo.revengeBuff.id] = 3,
    [warriorSpellInfo.victorious.id] = 2,
    [warriorSpellInfo.battleShout.id] = 1,
}

HoJ[HoJ.CLASS_NAME_WARRIOR] = {
    _spellInfos = HoJ._warriorSpellInfos,
    _className = "WARRIOR",
    _defaultSpecId = HoJ.SPEC_WARRIOR_ARMS,
    _powerType = 1,
    [HoJ.SPEC_NAME_WARRIOR_ARMS] = {
        -- properties
        _name = HoJ._text["Arms"],
        _spec = HoJ.SPEC_WARRIOR_ARMS,
        _role = HoJ.ROLE_DPS,
        _enabled = true,
        _multiTargetRange = 8,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 20, -- testing shows around 10 pps
        _autoPowerGen = true,
        _powerGainStep = 25,
        _powerMax = 100,
        -- functions
        _updateFn = HoJ.Warrior_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Warrior_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        -- tables
        _meleeAoESpells = armsMeleeAoESpells,
        _activeBuffList = armsActiveBuffList,
        _procdBuffList = armsProcdBuffList,
        _activeBlessings = armsActiveBlessingList,
        _rotationRules = { _singleTarget = armsRotationRules, _multiTarget = armsMultiTargetRotationRules },
        _interruptRules = { _singleTarget = armsInterruptRules, _multiTarget = armsInterruptRules },
        _mitigationRules = { _singleTarget = armsMitigationRules, _multiTarget = armsMitigationRules },
        _dpsCooldownRules = { _singleTarget = armsDpsCooldownRules, _multiTarget = armsDpsCooldownRules },
        _cooldownSortPriorities = armsCooldownSortPriorities,
        _buffSortPriorities = armsBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = warriorSpellInfo.mortalStrike
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.charge
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
        },
    },
    [HoJ.SPEC_NAME_WARRIOR_FURY] = {
        -- properties
        _name = HoJ._text["Fury"],
        _spec = HoJ.SPEC_WARRIOR_FURY,
        _role = HoJ.ROLE_DPS,
        _enabled = true,    -- addon enabled for this spec?
        _multiTargetRange = 8,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.6,
        _defaultPowerPerSec = 20, -- testing with my 114 shows about 8 pps
        _autoPowerGen = false,
        _powerGainStep = 0,
        _powerMax = 100,
        -- functions
        _updateFn = HoJ.Warrior_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Warrior_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        -- tables
        _meleeAoESpells = furyMeleeAoESpells,
        _activeBuffList = furyActiveBuffList,
        _procdBuffList = furyProcdBuffList,
        _activeBlessings = furyActiveBlessingList,
        _rotationRules = { _singleTarget = furyRotationRules, _multiTarget = furyMultiTargetRotationRules },
        _interruptRules = { _singleTarget = furyInterruptRules, _multiTarget = furyInterruptRules },
        _mitigationRules = { _singleTarget = furyMitigationRules, _multiTarget = furyMitigationRules },
        _dpsCooldownRules = { _singleTarget = furyDpsCooldownRules, _multiTarget = furyDpsCooldownRules },
        _cooldownSortPriorities = furyCooldownSortPriorities,
        _buffSortPriorities = furyBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = nil,
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = warriorSpellInfo.bloodthirst
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.charge
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
        },          
    },
    [HoJ.SPEC_NAME_WARRIOR_PROTECTION] = {
        -- properties
        _name = HoJ._text["Protection"],
        _spec = HoJ.SPEC_WARRIOR_PROTECTION,
        _role = HoJ.ROLE_TANK,            
        _enabled = true,    -- enabled for this spec?
        _multiTargetRange = 8,
        _minorHealHealthRatio = 0.9,
        _majorHealHealthRatio = 0.99,
        _defaultPowerPerSec = 30,
        _autoPowerGen = false,
        _powerGainStep = 0,
        _powerMax = 100,
        -- functions
        _updateFn = HoJ.Warrior_Update,
        _updateRenderFn = nil,
        _resetFn = nil,
        _selectSingleOrMultiTargetFn = HoJ.Warrior_SelectSingleOrMultitarget,
        _nextSkillSelectFn = HoJ.NextSkillOffCooldown,
        _checkBlessingsFn = nil,
        -- tables
        _meleeAoESpells = protMeleeAoESpells,
        _activeBuffList = protActiveBuffList,
        _procdBuffList = protProcdBuffList,
        _activeBlessings = protActiveBlessingList,
        _rotationRules = { _singleTarget = protRotationRules, _multiTarget = protRotationRules },
        _interruptRules = { _singleTarget = protInterruptRules, _multiTarget = protInterruptRules },
        _mitigationRules = { _singleTarget = protMitigationRules, _multiTarget = protMitigationRules },
        _dpsCooldownRules = { _singleTarget = nil, _multiTarget = nil },
        _cooldownSortPriorities = protCooldownSortPriorities,
        _buffSortPriorities = protBuffSortPriorities,
        _friendlyRangeSpells = {
            [HoJ.RANGE_MELEE] = nil,
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = nil,
            [HoJ.RANGE_12_YARDS] = nil,
            [HoJ.RANGE_15_YARDS] = nil,
            [HoJ.RANGE_20_YARDS] = nil,
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.intercept
            },
            [HoJ.RANGE_30_YARDS] = nil,
            [HoJ.RANGE_40_YARDS] = nil,
        },
        _enemyRangeSpells = {
            [HoJ.RANGE_MELEE] = {
                _spellInfo = warriorSpellInfo.pummel
            },
            [HoJ.RANGE_8_YARDS] = nil,
            [HoJ.RANGE_10_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_12_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_15_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_20_YARDS] = {
                _spellInfo = warriorSpellInfo.intimidatingShout,
            },
            [HoJ.RANGE_25_YARDS] = {
                _spellInfo = warriorSpellInfo.intercept
            },
            [HoJ.RANGE_30_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
            [HoJ.RANGE_40_YARDS] = {
                _spellInfo = warriorSpellInfo.taunt
            },
        },
    }
}
    
-- -----------------------------------------------------------------------------
-- HoJ:LoadWarrior
-- -----------------------------------------------------------------------------
function HoJ:LoadWarrior()
    if ( HoJ._classId ~= HoJ.CLASS_ID_WARRIOR ) then
        HoJ.Error( "LoadWarrior called but _classId is not warrior!" )
    end

    HoJ.SpellInfo = HoJ._warriorSpellInfos

    ----------------------------------------------------------------------------
    -- hide unused UI
    for i, v in ipairs( HoJ._HoPo ) do
        --HoJ:DebugMsg( "Setting alpha " .. i );
        v:SetAlpha( 0.0 );
        HoJ._HoPoGlow[i]:SetAlpha( 0.0 )
    end		
    -- hide tv frame
    HoJFrame.HPMaxed:SetAlpha( 0.0 );
    HoJFrame.HPMaxedGlow:SetAlpha( 0.0 );    
end

